/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef ClipFlags_h
#define ClipFlags_h


#include "Primitives.h"




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;


/**
 * stores if a point is inside or outside each of six canonical clipping planes.
 * <br/><br/>
 *
 * assumes subject point is in clip space.<br/>
 * clip space is the unit cube, after a homogeneous perspective transform, but
 * before the perspective divide.<br/><br/>
 *
 * reference:
 * <cite>ch13, 'jim blinns corner, a trip down the graphics pipeline' - blinn
 * (morgan kaufmann 1996)</cite>
 */
class ClipFlags
{
/// standard object services -----------------------------------------------------------------------
public:
	         ClipFlags();
	explicit ClipFlags(  const Vector4f&  );
private:
	explicit ClipFlags( dword );
public:

	        ~ClipFlags();
	         ClipFlags( const ClipFlags& );
	ClipFlags& operator=( const ClipFlags& );


/// commands ---------------------------------------------------------------------------------------
	        void  setForPoint( const Vector4f& );

	        void  setFromBoundaryCoords( const Vector3f& lower,
	                                     const Vector3f& upper );
	        void  setFromBoundaryCoords( const float XLoHiYLoHiZLoHi[6] );

	        void  zero();

	        ClipFlags& operator|=( const ClipFlags& );
	        ClipFlags& operator&=( const ClipFlags& );


/// queries ----------------------------------------------------------------------------------------
	        bool  isClipped()                                                                 const;
	        bool  isUnclipped()                                                               const;
	        bool  isCrossing( const ClipFlags& )                                              const;

	enum EPlane { XLower = 32, XUpper = 16, YLower = 8, YUpper = 4, ZLower = 2, ZUpper = 1 };
	        bool  isClipped( EPlane )                                                         const;

	        ClipFlags operator|( const ClipFlags& )                                           const;
	        ClipFlags operator&( const ClipFlags& )                                           const;


/// constants --------------------------------------------------------------------------------------
	static  const ClipFlags& NONE();
	static  const ClipFlags& ALL();


/// fields -----------------------------------------------------------------------------------------
private:
	/// plane:  XLo XHi YLo YHi ZLo ZHi
	/// bit:    5   4   3   2   1   0
	dword xXyYzZ_m;
};








/// inlines ///

/// standard object services -----------------------------------------------------------------------
inline
ClipFlags::ClipFlags()
 :	xXyYzZ_m( 0 )
{
}


inline
ClipFlags::ClipFlags
(
	const Vector4f& point
)
{
	setForPoint( point );
}


inline
ClipFlags::ClipFlags
(
	const dword xXyYzZ
)
 :	xXyYzZ_m( xXyYzZ )
{
}


inline
ClipFlags::~ClipFlags()
{
}


inline
ClipFlags::ClipFlags
(
	const ClipFlags& other
)
 :	xXyYzZ_m( other.xXyYzZ_m )
{
}


inline
ClipFlags& ClipFlags::operator=
(
	const ClipFlags& other
)
{
	xXyYzZ_m = other.xXyYzZ_m;

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
inline
void ClipFlags::zero()
{
	xXyYzZ_m = 0;
}


inline
ClipFlags& ClipFlags::operator|=
(
	const ClipFlags& other
)
{
	xXyYzZ_m |= other.xXyYzZ_m;

	return *this;
}


inline
ClipFlags& ClipFlags::operator&=
(
	const ClipFlags& other
)
{
	xXyYzZ_m &= other.xXyYzZ_m;

	return *this;
}




/// queries ----------------------------------------------------------------------------------------
inline
bool ClipFlags::isClipped() const
{
	return xXyYzZ_m != 0;
}


inline
bool ClipFlags::isUnclipped() const
{
	return xXyYzZ_m == 0;
}


inline
bool ClipFlags::isCrossing
(
	const ClipFlags& other
) const
{
	return bool(xXyYzZ_m ^ other.xXyYzZ_m);
}


inline
bool ClipFlags::isClipped
(
	const EPlane plane
) const
{
	return (xXyYzZ_m & plane) != 0;
}


inline
ClipFlags ClipFlags::operator|
(
	const ClipFlags& other
) const
{
	return ClipFlags( xXyYzZ_m | other.xXyYzZ_m );
}


inline
ClipFlags ClipFlags::operator&
(
	const ClipFlags& other
) const
{
	return ClipFlags( xXyYzZ_m & other.xXyYzZ_m );
}


}}




#endif//ClipFlags_h
