/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Filters_h
#define Filters_h


#include "Primitives.h"




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;


/**
 * a box radial 2d filter kernel function.<br/><br/>
 *
 * supply the kernel radius, and the query point distance.<br/><br/>
 */
float boxFilter( float kernelRadius,
                 float pointDistance );

/**
 * a cone radial 2d filter kernel function.<br/><br/>
 *
 * supply the kernel radius, and the query point distance.<br/><br/>
 */
float coneFilter( float kernelRadius,
                  float pointDistance );


/**
 * a gaussian radial 2d filter kernel function.<br/><br/>
 *
 * supply the kernel radius, and the query point distance.<br/><br/>
 *
 * reference: <cite>p144-p146 'convienient anti aliasing filters that minimize
 * bumpy sampling' - pavicic (graphics gems 1, AP 1990)</cite>
 */
float pavicicFilter( float kernelRadius,
                     float pointDistance );

/**
 * a gaussian radial 2d filter kernel function.<br/><br/>
 *
 * supply the kernel radius squared, and the query point distance squared.
 * <br/><br/>
 *
 * reference: <cite>p144-p146 'convienient anti aliasing filters that minimize
 * bumpy sampling' - pavicic (graphics gems 1, AP 1990)</cite>
 */
float pavicicFilter2( float kernelRadiusSquared,
                      float pointDistanceSquared );


/**
 * a gaussian radial 2d filter kernel function.<br/><br/>
 *
 * supply the kernel radius, and the query point distance.<br/><br/>
 *
 * reference: <cite>ch10 'principles of digital image synthesis'
 * - glassner (morgan kaufmann 1995)</cite>
 *
 * reference: <cite>'stochastic sampling in computer graphics'
 * - cook [acm transactions on graphics, jan 1986]</cite>
 */
float cookFilter( float kernelRadius,
                  float pointDistance );


/**
 * a gaussian radial 2d filter kernel function.<br/><br/>
 *
 * supply the kernel radius squared, and the query point distance squared.
 * <br/><br/>
 *
 * reference: <cite>ch10 'principles of digital image synthesis'
 * - glassner (morgan kaufmann 1995)</cite>
 *
 * reference: <cite>'stochastic sampling in computer graphics'
 * - cook [acm transactions on graphics, jan 1986]</cite>
 */
float cookFilter2( float kernelRadiusSquared,
                   float pointDistanceSquared );


/**
 * a quadratic spline radial 2d filter kernel function.<br/><br/>
 *
 * supply the kernel radius, and the query point distance.<br/><br/>
 *
 * reference: <cite>ch10 'principles of digital image synthesis'
 * - glassner (morgan kaufmann 1995)</cite>
 *
 * reference: <cite>p101-p104 'an optimal filter for image reconstruction'
 * - max (graphics gems 2, AP 1991)</cite>
 */
float maxFilter( float kernelRadius,
                 float pointDistance );


/**
 * a cubic spline radial 2d filter kernel function.<br/><br/>
 *
 * supply the kernel radius, and the query point distance.<br/><br/>
 *
 * has negative tail.<br/><br/>
 *
 * reference: <cite>ch10 'principles of digital image synthesis'
 * - glassner (morgan kaufmann 1995)</cite>
 *
 * reference: <cite>'reconstruction filters in computer graphics'
 * - mitchell, netravali [siggraph conference proceedings 1988]</cite>
 */
float mitchellNetravaliFilter( float kernelRadius,
                               float pointDistance );




/**
 * test for all filters.<br/><br/>
 *
 * returns true for success.
 */
bool test_Filters( bool isLogging );


}}




#endif//Filters_h
