/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Matrix4f_h
#define Matrix4f_h


#include "Primitives.h"
#include "Vector4f.h"
#include <iosfwd>

#include "hxa7241_utility.h"




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;
	using hxa7241::utility::SerializerOut;
	using hxa7241::utility::SerializerIn;


/**
 * a 4d matrix of floats.<br/><br/>
 *
 * usable for general 3d transforms, and perspective projection.<br/><br/>
 *
 * inversion code adapted from mgf, from graphics gems.
 */
class Matrix4f
{
/// standard object services -----------------------------------------------------------------------
public:
	         Matrix4f();
protected:
	explicit Matrix4f( bool );
public:
	         Matrix4f( bool           isColumns,
	                   const Vector4f parts[4] );
	         Matrix4f( float, float, float, float,
	                   float, float, float, float,
	                   float, float, float, float,
	                   float, float, float, float );
	explicit Matrix4f( const Matrix3f& );

	enum ETranslateOrScale { TRANSLATE, SCALE };
	enum ERotationAxis     { ROTATE_X, ROTATE_Y, ROTATE_Z };
	         Matrix4f( ETranslateOrScale,
	                   const Vector3f& );
	         Matrix4f( ERotationAxis,
	                   float degrees );
	         Matrix4f( float viewAngleDegrees,
	                   float zNear,
	                   float zFar );

	        ~Matrix4f();
	         Matrix4f( const Matrix4f& );
	Matrix4f& operator=( const Matrix4f& );

	friend  std::ostream& operator<<( std::ostream&, const Matrix4f& );
	friend  std::istream& operator>>( std::istream&,       Matrix4f& );


/// commands ---------------------------------------------------------------------------------------
	        void      setColumns( const Vector4f columns[4] );
	        void      setRows   ( const Vector4f rows[4] );
	        void      setParts  ( bool           isColumns,
	                              const Vector4f parts[4] );
	        void      setMatrix( const Matrix3f& );

	        void      setToTranslate( const Vector3f& );
	        void      setToScale( const Vector3f& );
	        void      setToRotateX( float degrees );
	        void      setToRotateY( float degrees );
	        void      setToRotateZ( float degrees );
	        void      setToRotateXYZ( float degreesX, float degreesY, float degreesZ );
	        void      setToPerspective( float viewAngleDegrees,
	                                    float zNear,
	                                    float zFar );

	        /**
	         * this = param * this.
	         */
	        void      multiplyEqPre ( const Matrix4f& );
	        /**
	         * this = this * param.
	         */
	        void      multiplyEqPost( const Matrix4f& );

	        void      transpose();
	        bool      invert();


/// queries ----------------------------------------------------------------------------------------
	        void      getColumns( Vector4f columns[4] )                                       const;
	        void      getRows   ( Vector4f rows[4] )                                          const;
	        void      getParts  ( bool     isColumns,
	                              Vector4f parts[4] )                                         const;

	        void      multiply( const Vector4f&,
	                            Vector4f& )                                                   const;

	        void      multiply( const Vector3f&,
	                            Vector4f& )                                                   const;
	        void      multiply( const Vector3f&,
	                            Vector3f& )                                                   const;
	        void      multiply( const UnitVector3f&,
	                            UnitVector3f& )                                               const;

	        bool      operator==( const Matrix4f& )                                           const;
	        bool      operator!=( const Matrix4f& )                                           const;


/// constants --------------------------------------------------------------------------------------
	static const Matrix4f& ZERO();
	static const Matrix4f& IDENTITY();


/// serialization ----------------------------------------------------------------------------------
	friend  SerializerOut& operator<<( SerializerOut&, const Matrix4f& );
	friend  SerializerIn&  operator>>( SerializerIn&,        Matrix4f& );
	static const char* CLASS_NAME();
	static const char* CLASS_DESCRIPTION();


/// implementation ---------------------------------------------------------------------------------
protected:
	        void      setColumn( int             index,
	                             const Vector4f& column );
	        void      setRow   ( int             index,
	                             const Vector4f& row );
	        void      setPart  ( bool            isColumn,
	                             int             index,
	                             const Vector4f& part );

	        void      setToRotate( float degrees,
	                               int   axis );

	        /**
	         * this = first * second.
	         */
	        void      multiplyEq( const Matrix4f&,
	                              const Matrix4f& );

	        void      getColumn( int       index,
	                             Vector4f& column )                                           const;
	        void      getRow   ( int       index,
	                             Vector4f& row )                                              const;
	        void      getPart  ( bool      isColumn,
	                             int       index,
	                             Vector4f& part )                                             const;

	static  bool      invert( float matrix[4][4] );
	static  void      adjoint( float matrix[4][4] );
	static  double    determinant4x4( const float matrix[4][4] );
	static  double    determinant3x3( double a1, double a2, double a3,
	                                  double b1, double b2, double b3,
	                                  double c1, double c2, double c3 );
	static  double    determinant2x2( double a, double b, double c, double d );


/// fields -----------------------------------------------------------------------------------------
private:
	float elements_m[4][4];
};








/// serialization ----------------------------------------------------------------------------------
SerializerOut& operator<<( SerializerOut&, const Matrix4f& );
SerializerIn&  operator>>( SerializerIn&,        Matrix4f& );


}}




#endif//Matrix4f_h
