/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "OctreeAuxiliary.h"   /// own header is included last


using namespace hxa7241::graphics;




/// OctreeBound ////////////////////////////////////////////////////////////////////////////////////


/// standard object services -----------------------------------------------------------------------
OctreeBound::OctreeBound
(
	const Vector3f& positionOfLowerCorner,
	const float     size
)
 :	positionOfLowerCorner_m( positionOfLowerCorner ),
	size_m                 ( size > 0.0f ? size : 0.0f ),
	center_m               ( (Vector3f::HALF() * size_m) += positionOfLowerCorner_m ),
	circumSphereRadius_m   ( (Vector3f::HALF() * size_m).length() )
{
}


OctreeBound::OctreeBound
(
	const OctreeBound& parentCellBound,
	const dword        subCellIndex
)
 :	positionOfLowerCorner_m( parentCellBound.positionOfLowerCorner_m.getX() + (float( subCellIndex       & 1) * (parentCellBound.size_m * 0.5f)),
	                         parentCellBound.positionOfLowerCorner_m.getY() + (float((subCellIndex >> 1) & 1) * (parentCellBound.size_m * 0.5f)),
	                         parentCellBound.positionOfLowerCorner_m.getZ() + (float((subCellIndex >> 2) & 1) * (parentCellBound.size_m * 0.5f)) ),
	size_m                 ( parentCellBound.size_m * 0.5f ),
	center_m               ( positionOfLowerCorner_m.getX() + (size_m * 0.5f),
	                         positionOfLowerCorner_m.getY() + (size_m * 0.5f),
	                         positionOfLowerCorner_m.getZ() + (size_m * 0.5f) ),
	circumSphereRadius_m   ( parentCellBound.circumSphereRadius_m * 0.5f )
{
}


OctreeBound::~OctreeBound()
{
}


OctreeBound::OctreeBound
(
	const OctreeBound& other
)
 :	positionOfLowerCorner_m( other.positionOfLowerCorner_m ),
	size_m                 ( other.size_m ),
	center_m               ( other.center_m ),
	circumSphereRadius_m   ( other.circumSphereRadius_m )
{
}


OctreeBound& OctreeBound::operator=
(
	const OctreeBound& other
)
{
	if( &other != this )
	{
		positionOfLowerCorner_m = other.positionOfLowerCorner_m;
		size_m                  = other.size_m;
		center_m                = other.center_m;
		circumSphereRadius_m    = other.circumSphereRadius_m;
	}

	return *this;
}








/// OctreeData /////////////////////////////////////////////////////////////////////////////////////


/// standard object services -----------------------------------------------------------------------
OctreeData::OctreeData
(
	const Vector3f& positionOfLowerCorner,
	const float     size,
	const dword     maxElementsPerCell,
	const dword     maxLevels
)
 :	bound_m             ( positionOfLowerCorner, size ),
	level_m             ( 0 ),
	maxElementsPerCell_m( maxElementsPerCell > 0 ? maxElementsPerCell : 1 ),
	maxLevel_m          ( maxLevels          > 0 ? maxLevels - 1      : 0 )
{
}


OctreeData::OctreeData
(
	const OctreeData& parentCellData,
	const dword       subCellIndex
)
 :	bound_m             ( parentCellData.bound_m, subCellIndex ),
	level_m             ( parentCellData.level_m + 1 ),
	maxElementsPerCell_m( parentCellData.maxElementsPerCell_m ),
	maxLevel_m          ( parentCellData.maxLevel_m )
{
}


OctreeData::~OctreeData()
{
}


OctreeData::OctreeData
(
	const OctreeData& other
)
 :	bound_m             ( other.bound_m ),
	level_m             ( other.level_m ),
	maxElementsPerCell_m( other.maxElementsPerCell_m ),
	maxLevel_m          ( other.maxLevel_m )
{
}


OctreeData& OctreeData::operator=
(
	const OctreeData& other
)
{
	if( &other != this )
	{
		bound_m              = other.bound_m;
		level_m              = other.level_m;
		maxElementsPerCell_m = other.maxElementsPerCell_m;
		maxLevel_m           = other.maxLevel_m;
	}

	return *this;
}




/// queries ----------------------------------------------------------------------------------------
bool OctreeData::isSubdivide
(
	const dword elementCount
) const
{
	return (elementCount > maxElementsPerCell_m) & (level_m < maxLevel_m);
}
