/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Polar_h
#define Polar_h


#include "Primitives.h"




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;


/**
 * transforms a point on the [0,1] square, to a point on the unit disk centered
 * at the origin.<br/><br/>
 *
 * inverse of diskToSquare.<br/><br/>
 *
 * reference:
 * <cite>'a low distortion map between disk and square' - shirley, chiu
 * (journal of graphics tools, v2n3p45 1997)</cite>
 *
 * @see diskToSquare
 */
void squareToDisk( float squrX01, float squrY01, float& Xminus1plus1, float& Yminus1plus1 );

/**
 * transforms a point on the unit disk centered at the origin, to a point on
 * the [0,1] square.<br/><br/>
 *
 * inverse of squareToDisk.<br/><br/>
 *
 * reference:
 * <cite>'a low distortion map between disk and square' - shirley, chiu
 * (journal of graphics tools, v2n3p45 1997)</cite>
 *
 * @see squareToDisk
 */
void diskToSquare( const float Xminus1plus1, const float Yminus1plus1, float& squrX01, float& squrY01 );




/**
 * transforms a point on the [0,1] square, to a point on the unit hemisphere
 * surface centered at the origin on the positive z side.
 */
void squareToHemisphere( const float squrX01, const float squrY01, float& Xminus1plus1, float& Yminus1plus1, float& Z01 );

/**
 * transforms a point on the unit hemisphere surface centered at the origin on the
 * positive z side, to a point on the [0,1] square.<br/><br/>
 *
 * inverse of squareToHemisphere.
 */
void hemisphereToSquare( float Xminus1plus1, float Yminus1plus1, float Z01, float& squrX01, float& squrY01 );

/**
 * transforms a point on the [0,1] square, to a point on the unit sphere surface
 * centered at the origin.
 */
//void squareToSphere( float squrX01, float squrY01, float& Xminus1plus1, float& Yminus1plus1, float& Zminus1plus1 );




/**
 * converts rectangular coords to polar coords.
 */
void rectangularToPolar( const float x, const float y, float& angle, float& radius );

/**
 * converts polar coords to rectangular coords.
 */
void polarToRectangular( const float angle, const float radius, float& x, float& y );

/**
 * calculates the sine and cosine of an angle, maybe faster than the std library.
 */
void sinAndcos( const float angle, float& sin, float& cos );

/**
 * converts degrees to radians.
 */
inline
float degreesToRadians( const float deg )
{
	const float PI_OVER_180 = float(3.1415926535897932384626433832795 / 180.0);
	return deg * PI_OVER_180;
}

/**
 * converts radians to degrees.
 */
inline
float radiansToDegrees( const float rad )
{
	const float _180_OVER_PI = float(180.0 / 3.1415926535897932384626433832795);
	return rad * _180_OVER_PI;
}




bool test_Polar( bool isLogging );


}}




#endif//Polar_h
