/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef UnitVector3f_h
#define UnitVector3f_h


#include "Primitives.h"
#include "Vector3f.h"
#include <iosfwd>

#include "hxa7241_utility.h"




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;
	using hxa7241::utility::SerializerOut;
	using hxa7241::utility::SerializerIn;


/**
 * a unitized 3d vector of floats.<br/><br/>
 *
 * provides some reflection, refraction, and orthogonalize functionality.
 * <br/><br/>
 *
 * the reflection and refraction methods assume both vectors are pointing away
 * from the transition point.
 */
class UnitVector3f
{
/// standard object services -----------------------------------------------------------------------
public:
	         UnitVector3f();
	         UnitVector3f( float x, float y, float z );
	explicit UnitVector3f( const Vector3f& );

	        ~UnitVector3f();
	         UnitVector3f( const UnitVector3f& );
	         UnitVector3f& operator=( const UnitVector3f& );
	         UnitVector3f& operator=( const Vector3f& );
private:
	         UnitVector3f( const Vector3f&, bool dummy );

	friend  std::ostream& operator<<( std::ostream&,  const UnitVector3f& );
	friend  std::istream& operator>>( std::istream&,        UnitVector3f& );


/// access -----------------------------------------------------------------------------------------
public:
	        UnitVector3f&   setXYZ( float x, float y, float z );
	        UnitVector3f&   setXYZ( const Vector3f& );

	        const Vector3f& getVector()                                                      const;
	        float           getX()                                                           const;
	        float           getY()                                                           const;
	        float           getZ()                                                           const;


/// arithmetic -------------------------------------------------------------------------------------
	        float           dot( const UnitVector3f& )                                       const;
	        UnitVector3f    cross  ( const UnitVector3f& )                                   const;
	        UnitVector3f&   crossEq( const UnitVector3f& );

	        UnitVector3f    orthogonalize  ( const UnitVector3f& )                           const;
	        UnitVector3f&   orthogonalizeEq( const UnitVector3f& );

			/// (both vectors pointing away)
	        UnitVector3f    reflect  ( const UnitVector3f& surfaceNormal )                   const;
	        UnitVector3f&   reflectEq( const UnitVector3f& surfaceNormal );
	        UnitVector3f&   reflectAlongEq( const UnitVector3f& surfaceNormal );

	        UnitVector3f    refract  ( const UnitVector3f& surfaceNormal,
	                                   float               relativeRefractiveIndex )         const;
	        UnitVector3f&   refractEq( const UnitVector3f& surfaceNormal,
	                                   float               relativeRefractiveIndex );
	        bool            isInTirZone( const UnitVector3f& surfaceNormal,
	                                     float               relativeRefractiveIndex )       const;

	        UnitVector3f    operator-()                                                      const;
	        UnitVector3f&   negate();


/// logical ---------------------------------------------------------------------------------------
	        bool            operator==( const UnitVector3f& )                                const;
	        bool            operator!=( const UnitVector3f& )                                const;
	        bool            isZero()                                                         const;


/// constants --------------------------------------------------------------------------------------
	static const UnitVector3f& ZERO();
	static const UnitVector3f& X();
	static const UnitVector3f& Y();
	static const UnitVector3f& Z();


/// serialization ----------------------------------------------------------------------------------
	friend  SerializerOut& operator<<( SerializerOut&, const UnitVector3f& );
	friend  SerializerIn&  operator>>( SerializerIn&,        UnitVector3f& );
	static const char* CLASS_NAME();
	static const char* CLASS_DESCRIPTION();


/// implementation ---------------------------------------------------------------------------------
private:
	        void  unitize();


/// fields -----------------------------------------------------------------------------------------
private:
	Vector3f vector_m;
};








/// serialization ----------------------------------------------------------------------------------
SerializerOut& operator<<( SerializerOut&, const UnitVector3f& );
SerializerIn&  operator>>( SerializerIn&,        UnitVector3f& );

std::ostream& operator<<( std::ostream&, const UnitVector3f& );
std::istream& operator>>( std::istream&,       UnitVector3f& );




/// INLINES ///


/// standard object services -----------------------------------------------------------------------
inline
UnitVector3f::UnitVector3f( const Vector3f& v, bool )
 :	vector_m( v )
{
}


inline
UnitVector3f::~UnitVector3f()
{
}




/// access -----------------------------------------------------------------------------------------
inline
const Vector3f& UnitVector3f::getVector() const
{
	return vector_m;
}


inline
float UnitVector3f::getX() const
{
	return vector_m.x_m;
}


inline
float UnitVector3f::getY() const
{
	return vector_m.y_m;
}


inline
float UnitVector3f::getZ() const
{
	return vector_m.z_m;
}




/// arithmetic -------------------------------------------------------------------------------------
inline
float UnitVector3f::dot( const UnitVector3f& v ) const
{
	return vector_m.dot( v.vector_m );
}


inline
UnitVector3f& UnitVector3f::negate()
{
	vector_m.negate();

	return *this;
}




/// logical ---------------------------------------------------------------------------------------
inline
bool UnitVector3f::operator==( const UnitVector3f& v ) const
{
	return vector_m.operator==( v.vector_m );
}


inline
bool UnitVector3f::operator!=( const UnitVector3f& v ) const
{
	return vector_m.operator!=( v.vector_m );
}




/// serialization ----------------------------------------------------------------------------------
inline
SerializerOut& hxa7241::graphics::operator<<( SerializerOut& out, const UnitVector3f& obj )
{
	return operator<<( out, obj.vector_m );
}


inline
SerializerIn& hxa7241::graphics::operator>>( SerializerIn& in, UnitVector3f& obj )
{
	operator>>( in, obj.vector_m );
	obj.unitize();

	return in;
}




}}




#endif//UnitVector3f_h
