/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Vector2_h
#define Vector2_h


#include "Primitives.h"
//#include <iosfwd>
#include "Serialization.h"
#include "hxa7241math.h"
#include <limits>
#include <iostream>




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;
	using hxa7241::utility::SerializerOut;
	using hxa7241::utility::SerializerIn;


/**
 * a 2d vector.
 *
 * @see Vector3f
 */
template<class TYPE>
class Vector2
{
/// standard object services -----------------------------------------------------------------------
public:
	                 Vector2();
	                 Vector2( TYPE x, TYPE y );
	explicit         Vector2( const TYPE xy[2] );

	                ~Vector2();
	                 Vector2( const Vector2& );
	        Vector2& operator=( const Vector2& );


/// access -----------------------------------------------------------------------------------------
	        Vector2&  setXY( TYPE x, TYPE y );
	        Vector2&  setXY( const TYPE xy[2] );

	        void      getXY( TYPE& x, TYPE& y )                const;
	        void      getXY( TYPE xy[2] )                      const;

	        TYPE      getX()                                   const;
	        TYPE      getY()                                   const;


/// arithmetic -------------------------------------------------------------------------------------
	        TYPE     dot( const Vector2& )                     const;
	        TYPE     distance( const Vector2& )                const;
	        TYPE     length()                                  const;

	        TYPE     total()                                   const;
	        TYPE     average()                                 const;
	        TYPE     largest()                                 const;

	        Vector2  unitize()                                 const;
	        Vector2& unitizeEq();

	        Vector2  cross( const Vector2& )                   const;
	        Vector2& crossEq( const Vector2& );

	        Vector2  operator-()                               const;
	        Vector2& negate();

	        Vector2  abs()                                     const;
	        Vector2& absEq();

	        Vector2  operator+ ( const Vector2& )              const;
	        Vector2& operator+=( const Vector2& );
	        Vector2  operator- ( const Vector2& )              const;
	        Vector2& operator-=( const Vector2& );

	        Vector2  operator* ( TYPE )                        const;
	        Vector2& operator*=( TYPE );
	        Vector2  operator/ ( TYPE )                        const;
	        Vector2& operator/=( TYPE );

	        Vector2  operator* ( const Vector2& )              const;
	        Vector2& operator*=( const Vector2& );
	        Vector2  operator/ ( const Vector2& )              const;
	        Vector2& operator/=( const Vector2& );

	friend  Vector2<TYPE> operator*( TYPE, const Vector2<TYPE>& );
	friend  Vector2<TYPE> operator/( TYPE, const Vector2<TYPE>& );


/// logical ---------------------------------------------------------------------------------------
	        bool      operator==( const Vector2& )             const;
	        bool      operator!=( const Vector2& )             const;
	        bool      isZero()                                 const;


/// clamps -----------------------------------------------------------------------------------------
	        Vector2& clampToMinOf( const Vector2& );
	        Vector2& clampToMaxOf( const Vector2& );
	        Vector2& clampBetween( const Vector2& min, const Vector2& max );
	        /**
	         * 0 to almost 1, ie: [0,1).
	         */
	        Vector2& clamp01();
	        /**
	         * 0 to almost 1, ie: [0,1).
	         */
	        void     clamp01( Vector2& result )                const;


/// constants --------------------------------------------------------------------------------------
	static const Vector2& ZERO();
	static const Vector2& ONE();
	static const Vector2& EPSILON();
	static const Vector2& ONE_MINUS_EPSILON();
	static const Vector2& ALMOST_ONE();
	static const Vector2& MIN();
	static const Vector2& MAX();
	static const Vector2& X();
	static const Vector2& Y();


/// serialization ----------------------------------------------------------------------------------
	static const char* CLASS_NAME();
	static const char* CLASS_DESCRIPTION();


/// fields -----------------------------------------------------------------------------------------
public:
	TYPE x_m;
	TYPE y_m;
};








/// serialization ----------------------------------------------------------------------------------
template<class TYPE>
SerializerOut& operator<<( SerializerOut&, const Vector2<TYPE>& );
template<class TYPE>
SerializerIn&  operator>>( SerializerIn&,        Vector2<TYPE>& );

template<class TYPE>
std::ostream& operator<<( std::ostream&, const Vector2<TYPE>& );
template<class TYPE>
std::istream& operator>>( std::istream&,       Vector2<TYPE>& );




///
template<class TYPE>
Vector2<TYPE> operator*( TYPE, const Vector2<TYPE>& );
template<class TYPE>
Vector2<TYPE> operator/( TYPE, const Vector2<TYPE>& );




/// ALIASES ///

typedef Vector2<dword> Vector2i;
typedef Vector2<float> Vector2f;




/// INLINES ///

/// standard object services -----------------------------------------------------------------------
template<class TYPE>
inline
Vector2<TYPE>::Vector2()
 :	x_m(),
	y_m()
{
}


template<class TYPE>
inline
Vector2<TYPE>::~Vector2()
{
}




/// access -----------------------------------------------------------------------------------------
template<class TYPE>
inline
TYPE Vector2<TYPE>::getX() const
{
	return x_m;
}


template<class TYPE>
inline
TYPE Vector2<TYPE>::getY() const
{
	return y_m;
}








/// TEMPLATES ///

/// standard object services -----------------------------------------------------------------------
template<class TYPE>
Vector2<TYPE>::Vector2( const TYPE x, const TYPE y )
 :	x_m( x ),
	y_m( y )
{
}


template<class TYPE>
Vector2<TYPE>::Vector2( const TYPE xy[2] )
 :	x_m( xy[0] ),
	y_m( xy[1] )
{
}


template<class TYPE>
Vector2<TYPE>::Vector2( const Vector2& v )
 :	x_m( v.x_m ),
	y_m( v.y_m )
{
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::operator=( const Vector2<TYPE>& v )
{
	x_m = v.x_m;
	y_m = v.y_m;

	return *this;
}




/// streaming --------------------------------------------------------------------------------------
template<class TYPE>
std::ostream& hxa7241::graphics::operator<<( std::ostream& out, const Vector2<TYPE>& obj )
{
	return out << "(" << obj.x_m << ", " << obj.y_m << ")";
}


template<class TYPE>
std::istream& hxa7241::graphics::operator>>( std::istream& in, Vector2<TYPE>& obj )
{
	char c;
	return in >> c >> obj.x_m >> c >> obj.y_m >> c;
}




/// access -----------------------------------------------------------------------------------------
template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::setXY( const TYPE x, const TYPE y )
{
	x_m = x;
	y_m = y;

	return *this;
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::setXY( const TYPE xy[2] )
{
	x_m = xy[0];
	y_m = xy[1];

	return *this;
}


template<class TYPE>
void Vector2<TYPE>::getXY( TYPE& x, TYPE& y ) const
{
	x = x_m;
	y = y_m;
}


template<class TYPE>
void Vector2<TYPE>::getXY( TYPE xy[2] ) const
{
	xy[0] = x_m;
	xy[1] = y_m;
}




/// arithmetic -------------------------------------------------------------------------------------
template<class TYPE>
TYPE Vector2<TYPE>::dot( const Vector2<TYPE>& v ) const
{
	return TYPE( (x_m * v.x_m) +
	             (y_m * v.y_m) );
}


template<class TYPE>
TYPE Vector2<TYPE>::distance( const Vector2<TYPE>& v ) const
{
	const TYPE xDif = TYPE(x_m - v.x_m);
	const TYPE yDif = TYPE(y_m - v.y_m);

	return TYPE(::sqrt( (xDif * xDif) +
	                    (yDif * yDif) ));
}


template<class TYPE>
TYPE Vector2<TYPE>::length() const
{
	return TYPE(::sqrt( (x_m * x_m) +
	                    (y_m * y_m) ));
}


template<class TYPE>
TYPE Vector2<TYPE>::total() const
{
	return TYPE(x_m + y_m);
}


template<class TYPE>
TYPE Vector2<TYPE>::average() const
{
	return TYPE( (x_m + y_m) / TYPE(2) );
}


template<class TYPE>
TYPE Vector2<TYPE>::largest() const
{
	return x_m >= y_m ? x_m : y_m;
}


template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::unitize() const
{
	const TYPE length = TYPE(::sqrt( (x_m * x_m) +
	                                 (y_m * y_m) ));

	if( length != TYPE(0) )
	{
		return Vector2<TYPE>( TYPE(x_m / length),
							  TYPE(y_m / length) );
	}
	else
	{
		return Vector2<TYPE>( TYPE(0),
		                      TYPE(0) );
	}
}


template<>
Vector2<float> Vector2<float>::unitize() const
{
	const float length = float(::sqrt( (x_m * x_m) +
	                                   (y_m * y_m) ));
	const float oneOverLength = length != 0.0f ? 1.0f / length : 0.0f;

	return Vector2<float>( x_m * oneOverLength,
	                       y_m * oneOverLength );
}


template<>
Vector2<double> Vector2<double>::unitize() const
{
	const double length = double(::sqrt( (x_m * x_m) +
	                                     (y_m * y_m) ));
	const double oneOverLength = length != 0.0 ? 1.0 / length : 0.0;

	return Vector2<double>( x_m * oneOverLength,
	                        y_m * oneOverLength );
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::unitizeEq()
{
	const TYPE length = TYPE(::sqrt( (x_m * x_m) +
	                                 (y_m * y_m) ));

	if( length != TYPE(0) )
	{
		x_m /= length;
		y_m /= length;
	}
	else
	{
		x_m = TYPE(0);
		y_m = TYPE(0);
	}

	return *this;
}


template<>
Vector2<float>& Vector2<float>::unitizeEq()
{
	const float length = float(::sqrt( (x_m * x_m) +
	                                   (y_m * y_m) ));
	const float oneOverLength = length != 0.0f ? 1.0f / length : 0.0f;

	x_m *= oneOverLength;
	y_m *= oneOverLength;

	return *this;
}


template<>
Vector2<double>& Vector2<double>::unitizeEq()
{
	const double length = double(::sqrt( (x_m * x_m) +
	                                     (y_m * y_m) ));
	const double oneOverLength = length != 0.0 ? 1.0 / length : 0.0;

	x_m *= oneOverLength;
	y_m *= oneOverLength;

	return *this;
}


template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::cross( const Vector2<TYPE>& v ) const
{
	return Vector2<TYPE>( TYPE((y_m * v.x_m) - (x_m * v.y_m)),
	                      TYPE((x_m * v.y_m) - (y_m * v.x_m)) );
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::crossEq( const Vector2<TYPE>& v )
{
	const TYPE x = TYPE((y_m * v.x_m) - (x_m * v.y_m));
	const TYPE y = TYPE((x_m * v.y_m) - (y_m * v.x_m));

	x_m = x;
	y_m = y;

	return *this;
}


template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::operator-() const
{
	return Vector2<TYPE>( TYPE(-x_m), TYPE(-y_m) );
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::negate()
{
	x_m = TYPE(-x_m);
	y_m = TYPE(-y_m);

	return *this;
}


template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::abs() const
{
	return Vector2<TYPE>( *this ).absEq();
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::absEq()
{
	if( x_m < TYPE(0) )
	{
		x_m = TYPE(-x_m);
	}
	if( y_m < TYPE(0) )
	{
		y_m = TYPE(-y_m);
	}

	return *this;
}




/// plus minus -------------------------------------------------------------------------------------
template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::operator+( const Vector2<TYPE>& v ) const
{
	return Vector2<TYPE>( TYPE(x_m + v.x_m),
	                      TYPE(y_m + v.y_m) );
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::operator+=( const Vector2<TYPE>& v )
{
	x_m += v.x_m;
	y_m += v.y_m;

	return *this;
}


template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::operator-( const Vector2<TYPE>& v ) const
{
	return Vector2<TYPE>( TYPE(x_m - v.x_m),
	                      TYPE(y_m - v.y_m) );
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::operator-=( const Vector2<TYPE>& v )
{
	x_m -= v.x_m;
	y_m -= v.y_m;

	return *this;
}




/// mult div ---------------------------------------------------------------------------------------
template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::operator*( const TYPE f ) const
{
	return Vector2<TYPE>( TYPE(x_m * f),
	                      TYPE(y_m * f) );
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::operator*=( const TYPE f )
{
	x_m *= f;
	y_m *= f;

	return *this;
}


template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::operator/( const TYPE f ) const
{
	return Vector2<TYPE>( TYPE(x_m / f),
	                      TYPE(y_m / f) );
}


template<>
Vector2<float> Vector2<float>::operator/( const float f ) const
{
	const float oneOverF = 1.0f / f;

	return Vector2<float>( x_m * oneOverF,
	                       y_m * oneOverF );
}


template<>
Vector2<double> Vector2<double>::operator/( const double f ) const
{
	const double oneOverF = 1.0 / f;

	return Vector2<double>( x_m * oneOverF,
	                        y_m * oneOverF );
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::operator/=( const TYPE f )
{
	x_m /= f;
	y_m /= f;

	return *this;
}


template<>
Vector2<float>& Vector2<float>::operator/=( const float f )
{
	const float oneOverF = 1.0f / f;

	x_m *= oneOverF;
	y_m *= oneOverF;

	return *this;
}


template<>
Vector2<double>& Vector2<double>::operator/=( const double f )
{
	const double oneOverF = 1.0 / f;

	x_m *= oneOverF;
	y_m *= oneOverF;

	return *this;
}


template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::operator*( const Vector2<TYPE>& v ) const
{
	return Vector2<TYPE>( TYPE(x_m * v.x_m),
	                      TYPE(y_m * v.y_m) );
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::operator*=( const Vector2<TYPE>& v )
{
	x_m *= v.x_m;
	y_m *= v.y_m;

	return *this;
}


template<class TYPE>
Vector2<TYPE> Vector2<TYPE>::operator/( const Vector2<TYPE>& v ) const
{
	return Vector2<TYPE>( TYPE(x_m / v.x_m),
	                      TYPE(y_m / v.y_m) );
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::operator/=( const Vector2<TYPE>& v )
{
	x_m /= v.x_m;
	y_m /= v.y_m;

	return *this;
}




/// friends ----------------------------------------------------------------------------------------
template<class TYPE>
Vector2<TYPE> hxa7241::graphics::operator*( const TYPE f, const Vector2<TYPE>& v )
{
	return Vector2<TYPE>( TYPE(f * v.x_m),
	                      TYPE(f * v.y_m) );
}


template<class TYPE>
Vector2<TYPE> hxa7241::graphics::operator/( const TYPE f, const Vector2<TYPE>& v )
{
	return Vector2<TYPE>( TYPE(f / v.x_m),
	                      TYPE(f / v.y_m) );
}




/// logical ---------------------------------------------------------------------------------------
template<class TYPE>
bool Vector2<TYPE>::operator==( const Vector2<TYPE>& v ) const
{
	return (x_m == v.x_m) &
	       (y_m == v.y_m);
}


template<class TYPE>
bool Vector2<TYPE>::operator!=( const Vector2<TYPE>& v ) const
{
	return (x_m != v.x_m) |
	       (y_m != v.y_m);
}


template<class TYPE>
bool Vector2<TYPE>::isZero() const
{
	return bool( (x_m == TYPE(0)) &
	             (y_m == TYPE(0)) );
}




/// clamps -----------------------------------------------------------------------------------------
template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::clampToMinOf( const Vector2<TYPE>& min )
{
	if( x_m < min.x_m )
	{
		x_m = min.x_m;
	}
	if( y_m < min.y_m )
	{
		y_m = min.y_m;
	}

	return *this;
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::clampToMaxOf( const Vector2<TYPE>& max )
{
	if( x_m > max.x_m )
	{
		x_m = max.x_m;
	}
	if( y_m > max.y_m )
	{
		y_m = max.y_m;
	}

	return *this;
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::clampBetween( const Vector2<TYPE>& min, const Vector2<TYPE>& max )
{
	if( x_m > max.x_m )
		x_m = max.x_m;
	else
	if( x_m < min.x_m )
		x_m = min.x_m;

	if( y_m > max.y_m )
		y_m = max.y_m;
	else
	if( y_m < min.y_m )
		y_m = min.y_m;

	return *this;
}


template<class TYPE>
Vector2<TYPE>& Vector2<TYPE>::clamp01()   /// [0,1)  0 to almost 1
{
	if( x_m >= TYPE(1) )
		x_m  = TYPE_ALMOST_ONE;
	else
	if( x_m < TYPE(0) )
		x_m = TYPE(0);

	if( y_m >= TYPE(1) )
		y_m  = TYPE_ALMOST_ONE;
	else
	if( y_m < TYPE(0) )
		y_m = TYPE(0);

	return *this;
}


template<class TYPE>
void Vector2<TYPE>::clamp01( Vector2<TYPE>& result ) const   /// [0,1)  0 to almost 1
{
	if( x_m >= TYPE(1) )
		result.x_m = TYPE_ALMOST_ONE;
	else
	if( x_m < TYPE(0) )
		result.x_m = TYPE(0);
	else
		result.x_m = x_m;

	if( y_m >= TYPE(1) )
		result.y_m = TYPE_ALMOST_ONE;
	else
	if( y_m < TYPE(0) )
		result.y_m = TYPE(0);
	else
		result.y_m = y_m;

//	result = *this;
//	result.clamp01();
}




/// constants --------------------------------------------------------------------------------------
template<class TYPE>
const Vector2<TYPE>& Vector2<TYPE>::ZERO()
{
	static const Vector2<TYPE> k( TYPE(0), TYPE(0) );
	return k;
}


template<class TYPE>
const Vector2<TYPE>& Vector2<TYPE>::ONE()
{
	static const Vector2<TYPE> k( TYPE(1), TYPE(1) );
	return k;
}


template<class TYPE>
const Vector2<TYPE>& Vector2<TYPE>::EPSILON()
{
	static const Vector2<TYPE> k( std::numeric_limits<TYPE>::epsilon(), std::numeric_limits<TYPE>::epsilon() );
	return k;
}


template<class TYPE>
const Vector2<TYPE>& Vector2<TYPE>::ONE_MINUS_EPSILON()
{
	static const Vector2<TYPE> k( TYPE(1) - std::numeric_limits<TYPE>::epsilon(), TYPE(1) - std::numeric_limits<TYPE>::epsilon() );
	return k;
}


template<class TYPE>
const Vector2<TYPE>& Vector2<TYPE>::ALMOST_ONE()
{
	static const Vector2<TYPE> k( TYPE(1) - std::numeric_limits<TYPE>::epsilon(), TYPE(1) - std::numeric_limits<TYPE>::epsilon() );
	return k;
}


template<class TYPE>
const Vector2<TYPE>& Vector2<TYPE>::MIN()
{
	static const Vector2<TYPE> k( std::numeric_limits<TYPE>::min(), std::numeric_limits<TYPE>::min() );
	return k;
}


template<class TYPE>
const Vector2<TYPE>& Vector2<TYPE>::MAX()
{
	static const Vector2<TYPE> k( std::numeric_limits<TYPE>::max(), std::numeric_limits<TYPE>::max() );
	return k;
}


template<class TYPE>
const Vector2<TYPE>& Vector2<TYPE>::X()
{
	static const Vector2<TYPE> k( TYPE(1), TYPE(0) );
	return k;
}


template<class TYPE>
const Vector2<TYPE>& Vector2<TYPE>::Y()
{
	static const Vector2<TYPE> k( TYPE(0), TYPE(1) );
	return k;
}




/// serialization ----------------------------------------------------------------------------------
template<class TYPE>
SerializerOut& hxa7241::graphics::operator<<( SerializerOut& out, const Vector2<TYPE>& obj )
{
	out.write( "x", obj.x_m );
	out.write( "y", obj.y_m );

	return out;
}


template<class TYPE>
SerializerIn& hxa7241::graphics::operator>>( SerializerIn& in, Vector2<TYPE>& obj )
{
	in.read( "x", obj.x_m );
	in.read( "y", obj.y_m );

	return in;
}


template<class TYPE>
const char* Vector2<TYPE>::CLASS_NAME()
{
	static const char CLASS_NAME[] = "hxa7241::graphics::Vector2";
	return CLASS_NAME;
}


template<class TYPE>
const char* Vector2<TYPE>::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a vector of 2 types";
	return CLASS_DESCRIPTION;
}


}}




#endif//Vector2_h
