/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef XyzE_h
#define XyzE_h


#include "Primitives.h"

#include "hxa7241_utility.h"




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;
	using hxa7241::utility::SerializerOut;
	using hxa7241::utility::SerializerIn;


/**
 * a greg ward realpixel, for any float triplet.<br/><br/>
 *
 * stores 3 related floats into one dword, by sharing an 8 bit exponent, and
 * using 8 bits per mantissa.<br/><br/>
 *
 * doesnt handle negative values - clamps to min of zero.<br/><br/>
 *
 * reference:
 * <cite>real pixels, 'graphics gems II' (AP 91)</cite>
 */
class XyzE
{
/// standard object services -----------------------------------------------------------------------
public:
	                 XyzE();
	                 XyzE( float x, float y, float z );
	explicit         XyzE( const float xyz[3] );
	explicit         XyzE( dword xyze );

	                ~XyzE();
	                 XyzE( const XyzE& );
	        XyzE&    operator=( const XyzE& );


/// commands ---------------------------------------------------------------------------------------
public:
	        void     set( float x, float y, float z );
	        void     set( const float xyz[3] );

	        void     set( dword xyze );


/// queries ----------------------------------------------------------------------------------------
public:
	        void     get( float& x, float& y, float& z ) const;
	        float*   get( float xyz[3] )                 const;

	        dword    get()                               const;

	        bool     operator==( const XyzE& )           const;
	        bool     operator!=( const XyzE& )           const;


/// serialization ----------------------------------------------------------------------------------
	friend  SerializerOut& operator<<( SerializerOut&, const XyzE& );
	friend  SerializerIn&  operator>>( SerializerIn&,        XyzE& );
	static const char* CLASS_NAME();
	static const char* CLASS_DESCRIPTION();


/// statics ----------------------------------------------------------------------------------------
public:
	static  dword    toXyzE( float x, float y, float z );
	static  dword    toXyzE( const float xyz[3] );

	static 	void     fromXyzE( dword xyze, float& x, float& y, float& z );
	static  float*   fromXyzE( dword xyze, float xyz[3] );


/// fields -----------------------------------------------------------------------------------------
private:
	dword xyze_m;   /// 0xRRGGBBEE
};




/// serialization ----------------------------------------------------------------------------------
SerializerOut& operator<<( SerializerOut&, const XyzE& );
SerializerIn&  operator>>( SerializerIn&,        XyzE& );








/// INLINES ///


/// standard object services -----------------------------------------------------------------------
inline
XyzE::XyzE()
 :	xyze_m()
{
}


inline
XyzE::XyzE( const float x, const float y, const float z )
 :	xyze_m( toXyzE( x, y, z ) )
{
}


inline
XyzE::XyzE( const float xyz[3] )
 :	xyze_m( toXyzE( xyz ) )
{
}


inline
XyzE::XyzE( const dword xyze )
 :	xyze_m( xyze )
{
}


inline
XyzE::~XyzE()
{
}


inline
XyzE::XyzE( const XyzE& other )
 :	xyze_m( other.xyze_m )
{
}


inline
XyzE&  XyzE::operator=( const XyzE& other )
{
	xyze_m = other.xyze_m;

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
inline
void XyzE::set( const float x, const float y, const float z )
{
	xyze_m = toXyzE( x, y, z );
}


inline
void XyzE::set( const float xyz[3] )
{
	xyze_m = toXyzE( xyz );
}


inline
void XyzE::set( const dword xyze )
{
	xyze_m = xyze;
}




/// queries ----------------------------------------------------------------------------------------
inline
void XyzE::get( float& x, float& y, float& z ) const
{
	fromXyzE( xyze_m, x, y, z );
}


inline
float* XyzE::get( float xyz[3] ) const
{
	return fromXyzE( xyze_m, xyz );
}


inline
dword XyzE::get() const
{
	return xyze_m;
}


inline
bool XyzE::operator==( const XyzE& other ) const
{
	return xyze_m == other.xyze_m;
}


inline
bool XyzE::operator!=( const XyzE& other ) const
{
	return xyze_m != other.xyze_m;
}


}}




#endif//XyzE_h
