/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "TextSerializerOut.h"
#include <sstream>

#include "RenderingOptions.h"


using namespace perceptuum2::control;




/// standard object services -----------------------------------------------------------------------
RenderingOptions::RenderingOptions()
{
	setToDefault();
}


RenderingOptions::~RenderingOptions()
{
}




/// commands ---------------------------------------------------------------------------------------
void RenderingOptions::setToDefault()
{
	immutableOptions_m.setToDefault();

	secondsBetweenSaves_m             = 240;
	tonemapAdaptationLuminance_m      = 30.0f;

	emitterPathNumber_m               = 1;
	photonMapSize_m                   = 50000;

	lightsourceEmittanceThreshold_m   = 10.0f;

	emitterPathLengthLimit_m          = 16;
	emitterPathEndThreshold_m         = 0.0f;
	eyePathLengthLimit_m              = 16;
	eyePathEndThreshold_m             = 0.0f;

	displayCharacteristics_m.setToDefault();

	isIlluminationLogging_m           = false;
	isRasterizerLogging_m             = false;
}


void RenderingOptions::setImmutableOptions( const RenderingOptionsImmutable& roi )
{
	immutableOptions_m = roi;
}




/// queries ----------------------------------------------------------------------------------------
void RenderingOptions::getSummary( string& s ) const
{
	std::ostringstream oss;

	hxa7241::utility::TextSerializerOut::serializeOut( *this, oss );

	s = oss.str();
}




const RenderingOptionsImmutable& RenderingOptions::getImmutableOptions() const
{
	return immutableOptions_m;
}


const Projection& RenderingOptions::getProjection() const
{
	return immutableOptions_m.getProjection();
}


dword RenderingOptions::getImageWidth() const
{
	return immutableOptions_m.getImageWidth();
}


dword RenderingOptions::getImageHeight() const
{
	return immutableOptions_m.getImageHeight();
}


CieXyz& RenderingOptions::getDefaultObjectColor( CieXyz& c ) const
{
	return immutableOptions_m.getDefaultObjectColor( c );
}




dword RenderingOptions::getPersistPeriod() const
{
	return secondsBetweenSaves_m;
}


float RenderingOptions::getTonemapping() const
{
	return tonemapAdaptationLuminance_m;
}




dword RenderingOptions::getEmitterPathNumber() const
{
	return emitterPathNumber_m;
}


dword RenderingOptions::getPhotonMapSize() const
{
	return photonMapSize_m;
}




float RenderingOptions::getLightsourceEmittanceThreshold() const
{
	return lightsourceEmittanceThreshold_m;
}




dword RenderingOptions::getEmitterPathLengthLimit() const
{
	return emitterPathLengthLimit_m;
}


float RenderingOptions::getEmitterPathEndThreshold() const
{
	return emitterPathEndThreshold_m;
}


dword RenderingOptions::getEyePathLengthLimit() const
{
	return eyePathLengthLimit_m;
}


float RenderingOptions::getEyePathEndThreshold() const
{
	return eyePathEndThreshold_m;
}




void RenderingOptions::getDisplayCharacteristics( DisplayMeasures& m ) const
{
	m = displayCharacteristics_m;
}




bool RenderingOptions::getIsIlluminationLogging() const
{
	return isIlluminationLogging_m;
}


bool RenderingOptions::getIsRasterizerLogging() const
{
	return isRasterizerLogging_m;
}




/// implementation ---------------------------------------------------------------------------------
void RenderingOptions::enforceConstraints()
{
	secondsBetweenSaves_m        = secondsBetweenSaves_m        >= 0    ? secondsBetweenSaves_m        : 0;
	tonemapAdaptationLuminance_m = tonemapAdaptationLuminance_m >  0.0f ? tonemapAdaptationLuminance_m : 0.0001f;


	emitterPathNumber_m = emitterPathNumber_m >= 0 ? emitterPathNumber_m : 0;
	photonMapSize_m     = photonMapSize_m     >= 0 ? photonMapSize_m     : 0;


	lightsourceEmittanceThreshold_m = lightsourceEmittanceThreshold_m >= 0.0f ? lightsourceEmittanceThreshold_m : 0.0f;


	emitterPathLengthLimit_m = emitterPathLengthLimit_m >= 0 ? emitterPathLengthLimit_m : 0;
	eyePathLengthLimit_m     = eyePathLengthLimit_m     >  0 ? eyePathLengthLimit_m     : 1;

	emitterPathEndThreshold_m = emitterPathEndThreshold_m >= 0.0f ? emitterPathEndThreshold_m : 0.0f;
	emitterPathEndThreshold_m = emitterPathEndThreshold_m <  1.0f ? emitterPathEndThreshold_m : 1.0f;

	eyePathEndThreshold_m = eyePathEndThreshold_m >= 0.0f ? eyePathEndThreshold_m : 0.0f;
	eyePathEndThreshold_m = eyePathEndThreshold_m <  1.0f ? eyePathEndThreshold_m : 1.0f;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::control::operator<<( SerializerOut& out, const RenderingOptions& obj )
{
	out.write( "immutableOptions",              obj.immutableOptions_m, "options that are fixed across different sessions of the same render" );

	out.write( "secondsBetweenSaves",           obj.secondsBetweenSaves_m );
	out.write( "tonemapAdaptationLuminance",    obj.tonemapAdaptationLuminance_m, "cd/m2" );

	out.write( "emitterPathNumber",             obj.emitterPathNumber_m );
	out.write( "photonMapSize",                 obj.photonMapSize_m );

	out.write( "lightsourceEmittanceThreshold", obj.lightsourceEmittanceThreshold_m, "watts" );

	out.write( "emitterPathLengthLimit",        obj.emitterPathLengthLimit_m );
	out.write( "emitterPathEndThreshold",       obj.emitterPathEndThreshold_m );
	out.write( "eyePathLengthLimit",            obj.eyePathLengthLimit_m );
	out.write( "eyePathEndThreshold",           obj.eyePathEndThreshold_m );

	out.write( "displayCharacteristics",        obj.displayCharacteristics_m );

	out.write( "isIlluminationLogging",         obj.isIlluminationLogging_m );
	out.write( "isRasterizerLogging",           obj.isRasterizerLogging_m );

	return out;
}


SerializerIn& perceptuum2::control::operator>>( SerializerIn& in, RenderingOptions& obj )
{
	in.read( "immutableOptions",              obj.immutableOptions_m );

	in.read( "secondsBetweenSaves",           obj.secondsBetweenSaves_m );
	in.read( "tonemapAdaptationLuminance",    obj.tonemapAdaptationLuminance_m );

	in.read( "emitterPathNumber",             obj.emitterPathNumber_m );
	in.read( "photonMapSize",                 obj.photonMapSize_m );

	in.read( "lightsourceEmittanceThreshold", obj.lightsourceEmittanceThreshold_m );

	in.read( "emitterPathLengthLimit",        obj.emitterPathLengthLimit_m );
	in.read( "emitterPathEndThreshold",       obj.emitterPathEndThreshold_m );
	in.read( "eyePathLengthLimit",            obj.eyePathLengthLimit_m );
	in.read( "eyePathEndThreshold",           obj.eyePathEndThreshold_m );

	in.read( "displayCharacteristics",        obj.displayCharacteristics_m );

	in.read( "isIlluminationLogging",         obj.isIlluminationLogging_m );
	in.read( "isRasterizerLogging",           obj.isRasterizerLogging_m );

	obj.enforceConstraints();

	return in;
}


const char* RenderingOptions::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::control::RenderingOptions";
	return CLASS_NAME;
}


const char* RenderingOptions::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "all options for controlling a rendering";
	return CLASS_DESCRIPTION;
}
