/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "ImageCieXyz.h"
#include "Exception.h"
#include "Serialization.h"
#include "FpToInt.h"
#include <sstream>

#include "RenderingState.h"


using namespace perceptuum2::control;




/// standard object services -----------------------------------------------------------------------
RenderingState::RenderingState()
{
}


RenderingState::~RenderingState()
{
}


RenderingState::RenderingState( const RenderingState& other )
 :	options_m          ( other.options_m ),
	imageAccumulation_m( other.imageAccumulation_m ),
	samplingControl_m  ( other.samplingControl_m )
{
}


RenderingState& RenderingState::operator=( const RenderingState& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void RenderingState::setOptions
(
	const RenderingOptionsImmutable& options
)
{
	options_m = options;
}


void RenderingState::setImage
(
	ImageCieXyz* pImage
)
{
	if( pImage == 0 )
	{
		throw hxa7241::utility::Exception( "null pointer to image given", CLASS_NAME(), "setImage" );
	}

	options_m.setImageSize( pImage->getWidth(), pImage->getHeight() );
	imageAccumulation_m.setImage( pImage );
}


ImageAccumulation& RenderingState::getImageAccumulation()
{
	return imageAccumulation_m;
}


SamplingControl& RenderingState::getSamplingControl()
{
	return samplingControl_m;
}




/// queries ----------------------------------------------------------------------------------------
void RenderingState::getSummary( string& s ) const
{
	std::ostringstream oss;

	oss << '\n' << "sampling statistics" << '\n' << '\n' << '\n';

	oss << "pixel count:  "   << imageAccumulation_m.getPixelCount()   << '\n';
	oss << "max deviation:  " << imageAccumulation_m.getMaxDeviation() << '\n' << '\n';

	const dword sampleCount     = imageAccumulation_m.getSamplesCount();
	const float samplesRequired = imageAccumulation_m.getSamplesRequiredCount();
	const dword percentDone     = hxa7241::utility::fpToIntTrunc( imageAccumulation_m.getCompletedRatio() * 100.0f );
	oss << "sample count:  "     << sampleCount     << '\n';
	oss << "samples required:  " << samplesRequired << '\n';
	oss << "percentage done:  "  << percentDone     << " %" << '\n';

	oss << '\n';

	s = oss.str();
}


const RenderingOptionsImmutable& RenderingState::getImmutableOptions() const
{
	return options_m;
}


dword RenderingState::getImageWidth() const
{
	return options_m.getImageWidth();
}


dword RenderingState::getImageHeight() const
{
	return options_m.getImageHeight();
}




/// implementation ---------------------------------------------------------------------------------
void RenderingState::assign
(
	const RenderingState& other
)
{
	options_m           = other.options_m;
	imageAccumulation_m = other.imageAccumulation_m;
	samplingControl_m   = other.samplingControl_m;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::control::operator<<( SerializerOut& out, const RenderingState& obj )
{
	out.write( "options",           obj.options_m );
	out.write( "imageAccumulation", obj.imageAccumulation_m );
	out.write( "samplingControl",   obj.samplingControl_m );

	return out;
}


SerializerIn& perceptuum2::control::operator>>( SerializerIn& in, RenderingState& obj )
{
	in.read( "options",           obj.options_m );
	in.read( "imageAccumulation", obj.imageAccumulation_m );
	in.read( "samplingControl",   obj.samplingControl_m );

	return in;
}


const char* RenderingState::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::control::RenderingState";
	return CLASS_NAME;
}


const char* RenderingState::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "persistable intermediate state for a rendering, to allow continuation across separate executions";
	return CLASS_DESCRIPTION;
}
