/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "FpToInt.h"
#include "TimePoint.h"
#include <sstream>
#include <time>

#include "RenderingStatus.h"


using namespace perceptuum2::control;




/// statics
const char RenderingStatus::STATUS_LINE_START[]                = "rendering";
const char RenderingStatus::STATUS_LINE_START2[]               = "rendered";
const char RenderingStatus::STATUS_LINE_FRAME[]                = "frame";
const char RenderingStatus::STATUS_LINE_LINE[]                 = "line";
const char RenderingStatus::STATUS_LINE_FRAMECOUNT[]           = "frames";
const char RenderingStatus::STATUS_LINE_IN[]                   = "in";
const char RenderingStatus::STATUS_LINE_DURATION[]             = "taken";
const char RenderingStatus::STATUS_LINE_SECONDS[]              = "seconds";
const char RenderingStatus::STATUS_LINE_FINISHED[]             = "finished";
const char RenderingStatus::STATUS_LINE_UNFINISHED[]           = "unfinished";
const char RenderingStatus::STATUS_LINE_STOPPED[]              = "stopped";
const char RenderingStatus::STATUS_LINE_WORKING[]              = "working...";
const char RenderingStatus::STATUS_LINE_FRAMETIME[]            = "average frame time";
const char RenderingStatus::STATUS_LINE_ESTIMATED_TIME_TOTAL[] = "estimated total time";
const char RenderingStatus::STATUS_LINE_ESTIMATED_TIME_LEFT[]  = "estimated remaining time";
const char RenderingStatus::STATUS_LINE_REMAINING[]            = "left";
const char RenderingStatus::STATUS_LINE_PERCENT_DONE[]         = "percent done";




/// standard object services -----------------------------------------------------------------------
RenderingStatus::RenderingStatus()
 :	isWorking_m          ( false ),
	isFinished_m         ( false ),
	workDurationTotal_m  ( 0 ),
	framesCompleted_m    ( 0 ),
	fractionCompleted_m  (),
	sessionStartTime_m   ( 0 ),
	workDurationSession_m( 0 )
{
}


RenderingStatus::~RenderingStatus()
{
}


RenderingStatus::RenderingStatus( const RenderingStatus& other )
{
	RenderingStatus::assign( other );
}


RenderingStatus& RenderingStatus::operator=( const RenderingStatus& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void RenderingStatus::start()
{
	isWorking_m           = true;
	isFinished_m          = false;

	sessionStartTime_m    = dword(std::time(0));
	workDurationSession_m = 0;
}


void RenderingStatus::stop()
{
	isWorking_m = false;
}


void RenderingStatus::setFinished()
{
	stop();

	isFinished_m = true;
}


void RenderingStatus::setCompletedFrames( const dword framesCompleted )
{
	framesCompleted_m = framesCompleted >= 0 ? framesCompleted : 0;

	/// update durations
	workDurationTotal_m   -= workDurationSession_m;
	workDurationSession_m  = dword(std::time(0)) - sessionStartTime_m;
	workDurationTotal_m   += workDurationSession_m;
}


void RenderingStatus::setCompletedProportion( const float fraction )
{
	fractionCompleted_m = fraction;
}




/// queries ----------------------------------------------------------------------------------------
void RenderingStatus::getSummary( string& s ) const
{
	std::ostringstream oss;

	oss << '\n';

	oss << STATUS_LINE_START2 << '\n';
	oss << '\t' << framesCompleted_m << ' ' << STATUS_LINE_FRAMECOUNT << '\n';
	oss << '\t' << STATUS_LINE_IN << ' ' << hxa7241::utility::TimePoint::getStringDuration( workDurationTotal_m ) << '\n';

	oss << '\n' << STATUS_LINE_FRAMETIME << ": " << getMeanFrameTime() << ' ' << STATUS_LINE_SECONDS << '\n';

	{
		const dword percentDone = hxa7241::utility::fpToIntTrunc(fractionCompleted_m * 100.0f);
		dword completeDuration;
		dword remainingDuration;
		getEstimatedDurations( completeDuration, remainingDuration );

		oss << '\n' << STATUS_LINE_PERCENT_DONE << ": " << percentDone << " %" << '\n';
		oss << STATUS_LINE_ESTIMATED_TIME_TOTAL << ": " << hxa7241::utility::TimePoint::getStringDuration( completeDuration ) << '\n';
		if( !isFinished_m )
		{
			oss << STATUS_LINE_ESTIMATED_TIME_LEFT << ": " << hxa7241::utility::TimePoint::getStringDuration( remainingDuration ) << '\n';
		}
	}

	if( isWorking_m )
	{
		oss << '\n' << STATUS_LINE_WORKING << '\n';
	}
	else
	{
		oss << '\n' << (isFinished_m ? STATUS_LINE_FINISHED : STATUS_LINE_STOPPED) << '\n';
	}

	oss << '\n';

	s = oss.str();
}


void RenderingStatus::getSummaryLine( string& s ) const
{
	std::ostringstream oss;

	if( isWorking_m )
	{
		dword completeDuration;
		dword remainingDuration;
		getEstimatedDurations( completeDuration, remainingDuration );

		const dword currentFrame = (framesCompleted_m + 1);

		oss << STATUS_LINE_START << ": " << STATUS_LINE_FRAME << ' ' << currentFrame << ";  " <<
		       STATUS_LINE_DURATION << ' ' << hxa7241::utility::TimePoint::getStringDuration( workDurationTotal_m );

		if( remainingDuration > 0 )
		{
			const dword percentDone = hxa7241::utility::fpToIntTrunc(fractionCompleted_m * 100.0f);
		    oss << " (" << percentDone << "% done, " << hxa7241::utility::TimePoint::getStringDuration( remainingDuration ) <<
		           ' ' << STATUS_LINE_REMAINING << ")";
		}
	}
	else
	{
		oss << STATUS_LINE_START2 << ": " << framesCompleted_m << ' ' << STATUS_LINE_FRAMECOUNT <<
		       ' ' << STATUS_LINE_IN << ' ' << hxa7241::utility::TimePoint::getStringDuration( workDurationTotal_m );

		oss << ";  ";
		if( isFinished_m )
		{
			oss << STATUS_LINE_FINISHED;
		}
		else
		{
			oss << STATUS_LINE_STOPPED;
		}
	}

	s = oss.str();
}


dword RenderingStatus::getFramesCompleted() const
{
	return framesCompleted_m;
}


dword RenderingStatus::getDuration() const
{
	return workDurationTotal_m;
}


void RenderingStatus::getEstimatedDurations
(
	dword& complete,
	dword& remaining
) const
{
	if( fractionCompleted_m > 0.0f )
	{
		complete  = hxa7241::utility::fpToIntTrunc( ((float(workDurationTotal_m) + 0.5f) / fractionCompleted_m) + FLOAT_ALMOST_ONE );
		remaining = complete - workDurationTotal_m;
	}
	else
	{
		complete  = 0;
		remaining = 0;
	}
}


float RenderingStatus::getMeanFrameTime() const
{
	float mean = 0.0f;
	if( framesCompleted_m > 0 )
	{
		mean = float(workDurationTotal_m) / float(framesCompleted_m);
	}

	return mean;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::control::operator<<( SerializerOut& out, const RenderingStatus& obj )
{
	out.write( "isWorking",         obj.isWorking_m );
	out.write( "isFinished",        obj.isFinished_m );

	out.write( "workDurationTotal", obj.workDurationTotal_m,        "total rendering duration in seconds" );
	out.write( "framesCompleted",   obj.framesCompleted_m,          "number of frames completed" );
	out.write( "fractionCompleted", float(obj.fractionCompleted_m), "fraction completed" );

	/// write some supplementary info
	{
		const float frameTimeMean = obj.getMeanFrameTime();
		dword estimatedTotalDuration;
		dword estimatedRemainingDuration;
		obj.getEstimatedDurations( estimatedTotalDuration, estimatedRemainingDuration );

		out.write( "frameTimeMean",              frameTimeMean,              "average time taken per frame" );
		out.write( "estimatedTotalDuration",     estimatedTotalDuration,     "estimated total rendering duration" );
		out.write( "estimatedRemainingDuration", estimatedRemainingDuration, "estimated rendering duration remaining" );
	}

	out.write( "sessionStartTime",    obj.sessionStartTime_m,    "latest session start time" );
	out.write( "workDurationSession", obj.workDurationSession_m, "latest session rendering duration in seconds" );

	return out;
}


SerializerIn& perceptuum2::control::operator>>( SerializerIn& in, RenderingStatus& obj )
{
	/// read in fields
	in.read( "isWorking",           obj.isWorking_m );
	in.read( "isFinished",          obj.isFinished_m );

	in.read( "workDurationTotal",   obj.workDurationTotal_m );
	in.read( "framesCompleted",     obj.framesCompleted_m );
	float fractionCompleted = 0.0f;
	in.read( "fractionCompleted",   fractionCompleted );
	obj.fractionCompleted_m = fractionCompleted;

	/// read through discarded items
	{
		dword discarded;
		float discardedF;

		in.read( "frameTimeMean",              discardedF );
		in.read( "estimatedTotalDuration",     discarded );
		in.read( "estimatedRemainingDuration", discarded );

		in.read( "sessionStartTime",    discarded );
		in.read( "workDurationSession", discarded );
	}

	/// zero non-deserialized fields
	obj.sessionStartTime_m    = 0;
	obj.workDurationSession_m = 0;


	/// enforce constraints
	obj.workDurationTotal_m = obj.workDurationTotal_m >= 0 ? obj.workDurationTotal_m : 0;
	obj.framesCompleted_m   = obj.framesCompleted_m   >= 0 ? obj.framesCompleted_m   : 0;


	return in;
}


const char* RenderingStatus::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::control::RenderingStatus";
	return CLASS_NAME;
}


const char* RenderingStatus::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "various user/programmer oriented stats, measures and indications of the rendering";
	return CLASS_DESCRIPTION;
}




/// implementation ---------------------------------------------------------------------------------
void RenderingStatus::assign
(
	const RenderingStatus& other
)
{
	if( &other != this )
	{
		isWorking_m           = other.isWorking_m;
		isFinished_m          = other.isFinished_m;

		workDurationTotal_m   = other.workDurationTotal_m;
		framesCompleted_m     = other.framesCompleted_m;
		fractionCompleted_m   = other.fractionCompleted_m;

		sessionStartTime_m    = other.sessionStartTime_m;
		workDurationSession_m = other.workDurationSession_m;
	}
}
