/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "ImageConverter.h"
#include "ImageCieXyz.h"
#include "ImageRgba.h"
#include "Logger.h"
#include "Exception.h"
#include "png.h"
#include <iostream>
#include <setjmp>

#include "PngReader.h"


using namespace perceptuum2::imaging;




/// statics
const char PngReader::FORMAT_NAME[]         = "png";
const char PngReader::EXCEPTION_MESSAGE_1[] = "failed to initialize libpng";
const char PngReader::EXCEPTION_MESSAGE_2[] = "stream read failure";
const char PngReader::EXCEPTION_MESSAGE_3[] = "no image sheet to read into";




/// libpng callbacks
void __cdecl readPngData( png_structp pPngObj, png_bytep pData, png_size_t length );
//void __cdecl userTransform( png_structp pPngObj, png_row_infop pRowInfo, png_bytep pData );
void __cdecl attendToPngError( png_structp pPngObj, png_const_charp pErrorMsg );
void __cdecl attendToPngWarning( png_structp pPngObj, png_const_charp pWarningMsg );




/// queries ----------------------------------------------------------------------------------------
void PngReader::getName
(
	string& s
) const
{
	s = FORMAT_NAME;
}


bool PngReader::read
(
	istream&     in,
	const float  adaptLuminance,
	ImageCieXyz& imageXyz
) const
{
	ImageRgba      imageRgba;
	ImageConverter converter;

	const bool didRead = read( in, imageRgba, converter );
	if( didRead )
	{
		converter.set( adaptLuminance );
		converter.convertRGBtoXYZ( imageRgba, false, (adaptLuminance != 0.0f), imageXyz );
	}

	return didRead;
}


bool PngReader::read
(
	istream&        in,
	ImageRgba&      imageRgba,
	ImageConverter& converter
) const
{
	dword    conversionFlags = 0;
	CieXyz   whiteColor;
	Vector3f chromaticities[3];
	Vector3f gamma;

	const bool didRead = readPng( in, 0, &imageRgba, conversionFlags, whiteColor, chromaticities, gamma );
	if( didRead )
	{
		if( conversionFlags & SRGB )
		{
			converter.setSRGB();
		}
		else
		{
			if( conversionFlags & COLOR )
			{
				converter.set( whiteColor, chromaticities );
			}
			if( conversionFlags & GAMMA )
			{
				converter.set( gamma );
			}
		}
	}

	return didRead;
}


bool PngReader::read
(
	istream&                          in,
	Sheet<hxa7241::primitives::byte>& image8Bit
) const
{
	/// unused
	dword    conversionFlags = 0;
	CieXyz   whiteColor;
	Vector3f chromaticities[3];
	Vector3f gamma;

	return readPng( in, &image8Bit, 0, conversionFlags, whiteColor, chromaticities, gamma );
}


bool PngReader::readPng
(
	istream&   in,
	Sheet<hxa7241::primitives::byte>* pImage8Bit,
	ImageRgba*                        pImageRgba,
	dword&     conversionFlags,
	CieXyz&    whiteColor,
	Vector3f   chromaticitiesXyz[3],
	Vector3f&  gammaXyz
) const
{
	static const char METHOD_NAME[] = "readPng";


	/// check preconditions
	if( (pImage8Bit == 0) & (pImageRgba == 0) )
	{
		throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_3, CLASS_NAME(), METHOD_NAME );
	}


	/// for info see: www.libpng.org/pub/png

	/// save stream start position
	const std::istream::pos_type streamStartPos = in.tellg();

	/// check for png signature
	const dword SIGNATURE_LENGTH = 8;
	bool isPng = false;
	{
		char signature[SIGNATURE_LENGTH];
		in.read( signature, SIGNATURE_LENGTH );
		if( in.fail() )
		{
			throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_2, CLASS_NAME(), METHOD_NAME );
		}
		in.seekg( streamStartPos );
		isPng = !::png_sig_cmp( reinterpret_cast<png_bytep>(signature), 0, SIGNATURE_LENGTH );
	}


	if( isPng )
	{
		png_structp pPngObj  = 0;
		png_infop   pPngInfo = 0;
		string      pngErrorMsg;

		try
		{
			/// do png reading initialization
			{
				pPngObj = ::png_create_read_struct( PNG_LIBPNG_VER_STRING, &pngErrorMsg, attendToPngError, attendToPngWarning );
				if( pPngObj == 0 )
				{
					throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_1, CLASS_NAME(), METHOD_NAME );
				}

				pPngInfo = ::png_create_info_struct( pPngObj );
				if( pPngInfo == 0 )
				{
					throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_1, CLASS_NAME(), METHOD_NAME );
				}

				/// do non c++ exception handling
				if( std::setjmp(pPngObj->jmpbuf) )
				{
					throw hxa7241::utility::Exception( pngErrorMsg.c_str(), CLASS_NAME(), METHOD_NAME );
				}

				/// set stream read callback
				::png_set_read_fn( pPngObj, &in, readPngData );
			}


			/// if in hadnt been repositioned, then this would be needed:
			//::png_set_sig_bytes( pPngObj, SIGNATURE_LENGTH );


			/// read info
			::png_read_info( pPngObj, pPngInfo );


			/// read color characteristics
			const bool isSrgb = bool( ::png_get_valid( pPngObj, pPngInfo, PNG_INFO_sRGB ) );
			conversionFlags = isSrgb ? SRGB : 0;
			if( !isSrgb )
			{
				double chromaticities[4][2];
				if( ::png_get_cHRM( pPngObj, pPngInfo, &(chromaticities[0][0]), &(chromaticities[0][1]), &(chromaticities[1][0]), &(chromaticities[1][1]),
				                                       &(chromaticities[2][0]), &(chromaticities[2][1]), &(chromaticities[3][0]), &(chromaticities[3][1]) ) )
				{
					whiteColor.setXYZ( chromaticities[0][0], chromaticities[0][1], 1.0-(chromaticities[0][0] + chromaticities[0][1]) );
					chromaticitiesXyz[0].setXYZ( chromaticities[1][0], chromaticities[1][1], 1.0-(chromaticities[1][0] + chromaticities[1][1]) );
					chromaticitiesXyz[1].setXYZ( chromaticities[2][0], chromaticities[2][1], 1.0-(chromaticities[2][0] + chromaticities[2][1]) );
					chromaticitiesXyz[2].setXYZ( chromaticities[3][0], chromaticities[3][1], 1.0-(chromaticities[3][0] + chromaticities[3][1]) );
					conversionFlags |= COLOR;
				}

				double gamma;
				if( ::png_get_gAMA( pPngObj, pPngInfo, &gamma ) )
				{
					gammaXyz.setXYZ( gamma, gamma, gamma );
					gammaXyz.clampToMinOf( Vector3f( 0.1f, 0.1f, 0.1f ) );
					gammaXyz = 1.0f / gammaXyz;

					conversionFlags |= GAMMA;
				}
			}


			/// set transformations
			/// - to convert any image into either 8bit grayscale, or 32bit rgba
			{
				const png_byte colorType = ::png_get_color_type( pPngObj, pPngInfo );
				const png_byte bitDepth  = ::png_get_bit_depth( pPngObj, pPngInfo );

				/// set for filling byte sheet
				if( pImage8Bit != 0 )
				{
					/// make grayscale image at least 8 bits per pixel
					if( (colorType == PNG_COLOR_TYPE_GRAY) | (colorType == PNG_COLOR_TYPE_GRAY_ALPHA) )
					{
						if( bitDepth < 8 )
						{
							::png_set_gray_1_2_4_to_8( pPngObj );
						}
					}

					/// composite onto background
					if( bool(colorType & PNG_COLOR_MASK_ALPHA) || bool(::png_get_valid( pPngObj, pPngInfo, PNG_INFO_tRNS )) )
					{
						//png_color_16  my_background = { byte(1), 1, 1, 1, 1 };   /// throws pow exception if zeroes are used
						png_color_16  my_background = { byte(0), 0, 0, 0, 0 };
						png_color_16p image_background;
						if( bool(::png_get_bKGD( pPngObj, pPngInfo, &image_background )) )
						{
							const dword n = dword(colorType == PNG_COLOR_TYPE_PALETTE);
							::png_set_background( pPngObj, image_background, PNG_BACKGROUND_GAMMA_FILE, n, 1.0 );
						}
						else
						{
							::png_set_background( pPngObj, &my_background, PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0 );
						}
					}

					/// change color to grayscale
					if( (colorType == PNG_COLOR_TYPE_PALETTE) |
						(colorType == PNG_COLOR_TYPE_RGB) | (colorType == PNG_COLOR_TYPE_RGB_ALPHA) )
					{
						if( colorType == PNG_COLOR_TYPE_PALETTE )
						{
							::png_set_palette_to_rgb( pPngObj );
						}

						::png_set_rgb_to_gray_fixed( pPngObj, 1, -1, -1 );
					}

					/// compensate for image gamma
					{
						double gamma = 1.0;
						::png_get_gAMA( pPngObj, pPngInfo, &gamma );
						if( conversionFlags & SRGB )
						{
							/// for sRGB, libpng seems to apply gamma^2, so set to sqrt(0.45455)
							gamma = 0.67419986246324208624649067643643;
						}
						::png_set_gamma( pPngObj, 1.0, gamma );
					}

					/// make 16 bit image 8 bit per pixel
					if( bitDepth == 16 )
					{
						::png_set_strip_16( pPngObj );
					}
				}
				/// set for filling rgba sheet
				else if( pImageRgba != 0 )
				{
					/// make grayscale image rgb
					if( (colorType == PNG_COLOR_TYPE_GRAY) | (colorType == PNG_COLOR_TYPE_GRAY_ALPHA) )
					{
						if( bitDepth < 8 )
						{
							::png_set_gray_1_2_4_to_8( pPngObj );
						}
						::png_set_gray_to_rgb( pPngObj );
					}
					/// make palette image rgb
					else if( colorType == PNG_COLOR_TYPE_PALETTE )
					{
						::png_set_palette_to_rgb( pPngObj );
					}

					/// make rgb image rgba
					if( !bool(colorType & PNG_COLOR_MASK_ALPHA) )
					{
						/// use tRNS as alpha value
						if( ::png_get_valid( pPngObj, pPngInfo, PNG_INFO_tRNS ) )
						{
							::png_set_tRNS_to_alpha( pPngObj );
						}
						/// make alpha opaque
						else
						{
							::png_set_filler( pPngObj, 0xFFFF, PNG_FILLER_AFTER );
						}
					}

					/// make 16 bit image 8 bit per pixel
					if( bitDepth == 16 )
					{
						::png_set_strip_16( pPngObj );
					}
				}

				::png_read_update_info( pPngObj, pPngInfo );
			}


			/// read pixels
			{
				const dword width  = ::png_get_image_width( pPngObj, pPngInfo );
				const dword height = ::png_get_image_height( pPngObj, pPngInfo );

				std::vector<png_bytep> rowPointers( height );

				///	read bytes
				if( pImage8Bit != 0 )
				{
					pImage8Bit->setSize( width, height );

					for( int i = height;  i-- > 0; )
					{
						rowPointers[i] = reinterpret_cast<png_bytep>( pImage8Bit->getRow( i ) );
					}
				}
				/// read rgbas
				else if( pImageRgba != 0 )
				{
					pImageRgba->set( width, height, ImageRgba::ABGR );

					for( int i = height;  i-- > 0; )
					{
						rowPointers[i] = reinterpret_cast<png_bytep>( pImageRgba->getRow( i ) );
					}
				}

				::png_read_image( pPngObj, &(rowPointers[0]) );
			}


			/// finish reading
			::png_read_end( pPngObj, 0 );


			/// destruct
			::png_destroy_read_struct( &pPngObj, &pPngInfo, 0 );
		}
		catch( ... )
		{
			/// clear up
			if( pPngObj != 0 )
			{
				::png_destroy_read_struct( &pPngObj, &pPngInfo, 0 );
			}
			in.seekg( streamStartPos );

			throw;
		}
	}

	return isPng;
}


const char* PngReader::CLASS_NAME()
{
	static const char CLASS_NAME[] = "PngReader";
	return CLASS_NAME;
}




/// libpng callbacks -------------------------------------------------------------------------------
void __cdecl readPngData
(
	png_structp pPngObj,
	png_bytep   pData,
	png_size_t  length
)
{
	istream* pIstream = reinterpret_cast<istream*>( ::png_get_io_ptr( pPngObj ) );

	pIstream->read( reinterpret_cast<char*>(pData), length );
	if( pIstream->fail() )
	{
		throw hxa7241::utility::Exception( PngReader::EXCEPTION_MESSAGE_2, PngReader::CLASS_NAME(), "readPngData" );
	}
}


/*void __cdecl userTransform
(
	png_structp   pPngObj,
	png_row_infop pRowInfo,
	png_bytep     pData
)
{
	//contents of row_info:
	//	png_uint_32 width      width of row
	//	png_uint_32 rowbytes   number of bytes in row
	//	png_byte color_type    color type of pixels
	//	png_byte bit_depth     bit depth of samples
	//	png_byte channels      number of channels (1-4)
	//	png_byte pixel_depth   bits per pixel (depth*channels)


	if( (pPngObj != 0) & (pRowInfo != 0) )
	{
		using hxa7241::utility::Logger;
		Logger() << "\t" << "width " << pRowInfo->width << ",  rowbytes " << pRowInfo->rowbytes <<
		            ",  color_type " << dword(pRowInfo->color_type) << ",  bit_depth " << dword(pRowInfo->bit_depth) <<
		            ",  channels " << dword(pRowInfo->channels) << ",  pixel_depth " << dword(pRowInfo->pixel_depth);
	}
}*/


void __cdecl attendToPngError
(
	png_structp     pPngObj,
	png_const_charp pErrorMsg
)
{
	string* pString = reinterpret_cast<string*>( ::png_get_error_ptr(pPngObj) );
	*pString = pErrorMsg;

	std::longjmp( pPngObj->jmpbuf, 1 );
}


void __cdecl attendToPngWarning
(
	png_structp,
	png_const_charp pWarningMsg
)
{
	using hxa7241::utility::Logger;
	Logger() << Logger::EXC1 << PngReader::CLASS_NAME() << " : " << pWarningMsg;
}
