/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Vector3f.h"
#include "hxa7241math.h"

#include "ToneMapperWard94.h"


using namespace perceptuum2::imaging;




/// statics
const float ToneMapperWard94::MIN_LUMINANCE                     =  0.0001f;

const float ToneMapperWard94::DEFAULT_MIN_DISPLAY_LUMINANCE     =  2.5f;   /// like a crt
const float ToneMapperWard94::DEFAULT_MAX_DISPLAY_LUMINANCE     = 85.0f;   /// like a crt
const float ToneMapperWard94::DEFAULT_ADAPT_LUMINANCE_FOR_SCENE = 30.0f;   /// indoors

const float ToneMapperWard94::GAMMA = 1.0f/2.2f;




/// standard object services -----------------------------------------------------------------------
ToneMapperWard94::ToneMapperWard94()
 :	//maxDisplayLuminance_m(),
	adaptLuminanceForScene_m( DEFAULT_ADAPT_LUMINANCE_FOR_SCENE )
	//scaleFactor_m(),
	//oneOverScaleFactor_m()
{
	set( DEFAULT_MIN_DISPLAY_LUMINANCE, DEFAULT_MAX_DISPLAY_LUMINANCE );
}


ToneMapperWard94::ToneMapperWard94
(
	const float minDisplayLuminance,
	const float maxDisplayLuminance
)
 :	//maxDisplayLuminance_m(),
	adaptLuminanceForScene_m( DEFAULT_ADAPT_LUMINANCE_FOR_SCENE )
	//scaleFactor_m(),
	//oneOverScaleFactor_m()
{
	set( minDisplayLuminance, maxDisplayLuminance );
}


ToneMapperWard94::~ToneMapperWard94()
{
}


ToneMapperWard94::ToneMapperWard94( const ToneMapperWard94& other )
// :	maxDisplayLuminance_m   ( other.maxDisplayLuminance_m ),
//	adaptLuminanceForScene_m( other.adaptLuminanceForScene_m ),
//	scaleFactor_m           ( other.scaleFactor_m ),
//	oneOverScaleFactor_m    ( other.oneOverScaleFactor_m )
{
	ToneMapperWard94::operator=( other );
}


ToneMapperWard94& ToneMapperWard94::operator=( const ToneMapperWard94& other )
{
	if( &other != this )
	{
		maxDisplayLuminance_m    = other.maxDisplayLuminance_m;
		adaptLuminanceForScene_m = other.adaptLuminanceForScene_m;

		scaleFactor_m            = other.scaleFactor_m;
		oneOverScaleFactor_m     = other.oneOverScaleFactor_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void ToneMapperWard94::set
(
	float ,//minDisplayLuminance
	float maxDisplayLuminance
)
{
	if( maxDisplayLuminance < MIN_LUMINANCE )
	{
		maxDisplayLuminance = MIN_LUMINANCE;
	}

	maxDisplayLuminance_m = maxDisplayLuminance;

	setToneMapping();

}


void ToneMapperWard94::set
(
	float adaptLuminanceForScene
)
{
	if( adaptLuminanceForScene < MIN_LUMINANCE )
	{
		adaptLuminanceForScene = MIN_LUMINANCE;
	}

	adaptLuminanceForScene_m = adaptLuminanceForScene;

	setToneMapping();
}




/// queries ----------------------------------------------------------------------------------------
float ToneMapperWard94::map
(
	const float luminance
) const
{
	return gammarize( luminance * scaleFactor_m, GAMMA );
}


float ToneMapperWard94::unMap
(
	const float tone
) const
{
	return gammarize( tone, 1.0f/GAMMA ) * oneOverScaleFactor_m;
}


void ToneMapperWard94::map
(
	Vector3f& xyzToTone
) const
{
	xyzToTone.setXYZ( map( xyzToTone.getX() ),
	                  map( xyzToTone.getY() ),
	                  map( xyzToTone.getZ() ) );
}


void ToneMapperWard94::unMap
(
	Vector3f& toneToXyz
) const
{
	toneToXyz.setXYZ( unMap( toneToXyz.getX() ),
	                  unMap( toneToXyz.getY() ),
	                  unMap( toneToXyz.getZ() ) );
}


float ToneMapperWard94::getAdaptLuminance() const
{
	return adaptLuminanceForScene_m;
}




/// implementation ---------------------------------------------------------------------------------
float ToneMapperWard94::gammarize
(
	float in,
	float gamma
)
{
	if( in < 0.0f )
	{
		in = 0.0f;
	}

	return float(::pow( in, gamma ));
}


void ToneMapperWard94::setToneMapping()
{
	const float displayAdaptationLevel = (maxDisplayLuminance_m / 2.0f) / 2.0f;
	const float sceneAdaptationLevel   = adaptLuminanceForScene_m;

	const float a = 1.219f + float(::pow( displayAdaptationLevel, 0.4f ));
	const float b = 1.219f + float(::pow( sceneAdaptationLevel,   0.4f ));

	scaleFactor_m        = float(::pow( a / b, 2.5f )) / maxDisplayLuminance_m;
	oneOverScaleFactor_m = 1.0f / scaleFactor_m;
}
