/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "BssdfSpike.h"   /// own header is included last


using namespace perceptuum2::interaction;




/// standard object services -----------------------------------------------------------------------
BssdfSpike::BssdfSpike
(
	const UnitVector3f& normal,
	const UnitVector3f& tangent,
	const float         refractiveIndexOutside,
	const float         refractiveIndexInside,
	const Float01       transparency,
	const Light&        reflectivity
)
 :	Bssdf( normal, tangent, refractiveIndexOutside, refractiveIndexInside, transparency )
{
	BssdfSpike::setBssdfSpike( reflectivity );
}


BssdfSpike::~BssdfSpike()
{
}


BssdfSpike::BssdfSpike( const BssdfSpike& other )
{
	BssdfSpike::operator=( other );
}


BssdfSpike& BssdfSpike::operator=( const BssdfSpike& other )
{
	if( &other != this )
	{
		Bssdf::operator=( other );

		reflectivitySpectral_m = other.reflectivitySpectral_m;

		reflectivity_m = other.reflectivity_m;
		color_m        = other.color_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void BssdfSpike::setBssdfSpike
(
	const Light& reflectivity
)
{
	/// precondition reflectivity
	reflectivitySpectral_m = reflectivity;
	reflectivitySpectral_m.clampTo01Positive();

	/// set auxilliarys
	{
		float mean;
		reflectivitySpectral_m.getMeanAndColor( mean, color_m );
		reflectivity_m = mean;
	}
}




/// queries ----------------------------------------------------------------------------------------
Brdf* BssdfSpike::clone() const
{
	Brdf* pClone = new BssdfSpike( *this );
	if( pClone == 0 )
	{
		throw "BssdfSpike::clone()\ncouldnt get memory";
	}

	return  pClone;
}




/// implementation ---------------------------------------------------------------------------------
Float01 BssdfSpike::getReflectivity_() const
{
	return reflectivity_m;
}


void BssdfSpike::getPartitions_
(
	float   proportions[],
	Float01 peakness[],
	dword&  length
) const
{
	proportions[0] = 1.0f;
	peakness[0]    = Float01::MAX();

	length = 1;
}


Float01 BssdfSpike::getCos_
(
	const dword         ,//part,
	const UnitVector3f& //outDirection
) const
{
	return 1.0f;
}


void BssdfSpike::getScaling_
(
	const dword         ,//part,
	const float         ,//IdotN,
	const float         ,//OdotN,
	const UnitVector3f& ,//inDirection,
	const UnitVector3f& ,//outDirection,
	Light&              scaling
) const
{
	/// not really do-able numerically...

	/*UnitVector3f mirrorDirection( inDirection );
	mirrorDirection.reflectEq( getNormal() );

	const float outDotMirror = mirrorDirection.dot( outDirection );
	if( outDotMirror >= 1.0f )
	{
		(scaling = color_m) *= FLOAT_LARGE;
	}
	else
	{
		scaling = Light::ZERO();
	}*/


	scaling = Light::ZERO();
}


void BssdfSpike::getSample_
(
	const dword         ,//part,
	const Vector2f01&   ,//stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	outDirection = inDirection;
	outDirection.reflectEq( getNormal() );

	scaling = color_m;
}
