/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Float01.h"
#include "Float01c.h"
#include <iostream>

#include "EyePathNode.h"   /// own header is included last


using namespace perceptuum2::interaction;




/// standard object services -----------------------------------------------------------------------
EyePathNode::EyePathNode()
 :	SurfacePoint()
{
}


EyePathNode::EyePathNode
(
	const Intersection&       intersection,
	const ObjectRenderableId& outsideObject
)
 :	SurfacePoint( intersection, outsideObject )
{
}


EyePathNode::~EyePathNode()
{
}


EyePathNode::EyePathNode( const EyePathNode& other )
 :	SurfacePoint( other )
{
}


EyePathNode& EyePathNode::operator=( const EyePathNode& other )
{
	assign( other );

	return *this;
}




/// streaming --------------------------------------------------------------------------------------
std::ostream& perceptuum2::interaction::operator<<( std::ostream& out, const EyePathNode& obj )
{
	return out << "{" << *static_cast<const SurfacePoint*>(&obj) << "}";
}




/// queries ----------------------------------------------------------------------------------------
void EyePathNode::getNextPathStep
(
	const Float01       stochastics[7],
	const UnitVector3f& outDirection,
	Ray&                ray,
	Light&              scaling,
	bool&               isDiffuse,
	bool&               isDirectSeparate
) const
{
	/// stochastically choose specular or diffuse, multiply sdf scaling, multiply reflectivity, multiply cos

	dword status;
	getPathStep( 0, stochastics, outDirection,
				 ray, scaling, status );

	isDiffuse        = bool(status & IS_DIFFUSE);
	isDirectSeparate = bool(status & IS_DIRECT_SEPARATE);
}


void EyePathNode::getGatheredLightDistal
(
	const vector<LightStream>& illuminations,
	const UnitVector3f&        outDirection,
	const Float01c             shinynessPoint,
	Light&                     outLight
) const
{
	getGatheredLight( illuminations, outDirection, shinynessPoint, true, outLight );
}


void EyePathNode::getGatheredLightLocal
(
	const vector<LightStream>& illuminations,
	const UnitVector3f&        outDirection,
	const Float01c             shinynessPoint,
	Light&                     outLight
) const
{
	getGatheredLight( illuminations, outDirection, shinynessPoint, false, outLight );
}


void EyePathNode::getEmittedLight
(
	const UnitVector3f& outDirection,
	Light&              emitted
) const
{
	SurfacePoint::getEmittedLight( outDirection, emitted );
}




/// implementation ---------------------------------------------------------------------------------
void EyePathNode::getGatheredLight
(
	const vector<LightStream>& illuminations,
	const UnitVector3f&        outDirection,
	const Float01c             shinynessPoint,
	const bool                 isCos,
	Light&                     outLight
) const
{
	/// inward light * brdf diffuse parts * reflectivity * maybe cos in

	UnitVector3f outDirClamped( outDirection );
	clampNearToFakeSurface( outDirClamped );


	const Brdf& brdf = getBrdf();

	float   proportions[brdf.PARTITIONS_MAX];
	Float01 peakness[brdf.PARTITIONS_MAX];
	dword   partsCount;
	brdf.getPartitions( proportions, peakness, partsCount );


	/// add incoming light
	Light lightSum;
	for( dword i = illuminations.size();  i-- > 0; )
	{
		const LightStream& illumination = illuminations[i];

		UnitVector3f inDirClamped( illumination.direction_m );
		clampNearToFakeSurface( inDirClamped );

		/// sum diffuse parts of brdf
		Light diffuseScaling;
		for( dword part = partsCount;  part-- > 0; )
		{
			if( peakness[part] < shinynessPoint )
			{
				Light scaling;
				brdf.getScaling( part, inDirClamped, outDirClamped, scaling );

				scaling *= proportions[part];
				if( isCos )
				{
					scaling *= brdf.getCos( part, inDirClamped );
				}

				diffuseScaling += scaling;
			}
		}

		lightSum += (diffuseScaling *= illumination.light_m);
	}


	float diffuseProportion = 0.0f;
	for( dword part = partsCount;  part-- > 0; )
	{
		diffuseProportion += (peakness[part] < shinynessPoint) ? proportions[part] : 0.0f;
	}
	const float oneOverDiffuseProportion = (diffuseProportion > 0.0f) ? 1.0f / diffuseProportion : 0.0f;

	outLight = (lightSum *= (oneOverDiffuseProportion * brdf.getReflectivity() * Brdf::PI));
}


void EyePathNode::assign
(
	const SurfacePoint& other
)
{
	if( &other != this )
	{
		SurfacePoint::assign( other );
	}
}


void EyePathNode::assign
(
	const EyePathNode& other
)
{
	if( &other != this )
	{
		SurfacePoint::assign( other );
	}
}
