/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "CieXyz.h"
#include "Serialization.h"

#include "Light.h"   /// own header is included last


using namespace perceptuum2::interaction;




/// standard object services -----------------------------------------------------------------------
Light::Light()
{
	s_m[0] = 0.0f;
	s_m[1] = 0.0f;
	s_m[2] = 0.0f;
}


Light& Light::operator=( const Light& other )
{
	if( &other != this )
	{
		s_m[0] = other.s_m[0];
		s_m[1] = other.s_m[1];
		s_m[2] = other.s_m[2];
	}

	return *this;
}


Light& Light::operator=( const CieXyz& xyz )
{
	s_m[0] = xyz.getX();
	s_m[1] = xyz.getY();
	s_m[2] = xyz.getZ();

	return *this;
}


Light::Light( const float s0, const float s1, const float s2 )
{
	s_m[0] = s0;
	s_m[1] = s1;
	s_m[2] = s2;
}




/// streaming --------------------------------------------------------------------------------------
std::ostream& perceptuum2::interaction::operator<<( std::ostream& out, const Light& obj )
{
	return out << "(" << obj.s_m[0] << ", " << obj.s_m[1] << ", " << obj.s_m[2] << ")";
}


std::istream& perceptuum2::interaction::operator>>( std::istream& in, Light& obj )
{
	char c;
	return in >> c >> obj.s_m[0] >> c >> obj.s_m[1] >> c >> obj.s_m[2] >> c;
}




/// commands ---------------------------------------------------------------------------------------
Light& Light::setToMean1Color()
{
	clampToPositive();

	const float mean = average();
	if( mean == 0.0f )
	{
		operator=( Light::ONE() );
	}
	else if( mean != 1.0f )
	{
		operator/=( mean );
	}

	return *this;
}


Light& Light::clampToPositive()
{
	if( s_m[0] < 0.0f )
	{
		s_m[0] = 0.0f;
	}
	if( s_m[1] < 0.0f )
	{
		s_m[1] = 0.0f;
	}
	if( s_m[2] < 0.0f )
	{
		s_m[2] = 0.0f;
	}

	return *this;
}


Light& Light::clampTo01Positive()
{
	clampToPositive();

	const float mean = average();
	if( mean >= 1.0f )
	{
		operator*=( FLOAT_ALMOST_ONE / mean );
	}

	return *this;
}




/// queries ----------------------------------------------------------------------------------------
CieXyz Light::getXYZ() const
{
	return CieXyz( s_m );
}


void Light::getXYZ( CieXyz& xyz ) const
{
	xyz.setXYZ( s_m );
}


float Light::getLuminance() const
{
	return s_m[1];
}


void Light::getMeanAndColor
(
	float& mean,
	Light& color
) const
{
	color = *this;
	color.clampToPositive();

	const float m = color.average();
	if( m != 0.0f )
	{
		color /= m;
	}
	else
	{
		color = Light::ONE();
	}

	mean = m;
}


float Light::total() const
{
	return s_m[0] + s_m[1] + s_m[2];
}


float Light::average() const
{
	static const float ONE_OVER_THREE = 1.0f / 3.0f;

	return (s_m[0] + s_m[1] + s_m[2]) * ONE_OVER_THREE;
}


float Light::smallest() const
{
	float smallest;
	smallest = s_m[0]   <= s_m[1] ? s_m[0]   : s_m[1];
	smallest = smallest <= s_m[2] ? smallest : s_m[2];

	return smallest;
}


float Light::largest() const
{
	float largest;
	largest = s_m[0]  >= s_m[1] ? s_m[0]  : s_m[1];
	largest = largest >= s_m[2] ? largest : s_m[2];

	return largest;
}




/// arithmetic -------------------------------------------------------------------------------------
Light  Light::operator+ ( const Light& l ) const
{
	return Light( s_m[0] + l.s_m[0],
	              s_m[1] + l.s_m[1],
	              s_m[2] + l.s_m[2] );
}


Light& Light::operator+=( const Light& l )
{
	s_m[0] += l.s_m[0];
	s_m[1] += l.s_m[1];
	s_m[2] += l.s_m[2];

	return *this;
}


Light  Light::operator- ( const Light& l ) const
{
	return Light( s_m[0] - l.s_m[0],
	              s_m[1] - l.s_m[1],
	              s_m[2] - l.s_m[2] );
}


Light& Light::operator-=( const Light& l )
{
	s_m[0] -= l.s_m[0];
	s_m[1] -= l.s_m[1];
	s_m[2] -= l.s_m[2];

	return *this;
}


Light  Light::operator* ( const Light& l ) const
{
	return Light( s_m[0] * l.s_m[0],
	              s_m[1] * l.s_m[1],
	              s_m[2] * l.s_m[2] );
}


Light& Light::operator*=( const Light& l )
{
	s_m[0] *= l.s_m[0];
	s_m[1] *= l.s_m[1];
	s_m[2] *= l.s_m[2];

	return *this;
}


Light  Light::operator/ ( const Light& l ) const
{
	return Light( s_m[0] / l.s_m[0],
	              s_m[1] / l.s_m[1],
	              s_m[2] / l.s_m[2] );
}


Light& Light::operator/=( const Light& l )
{
	s_m[0] /= l.s_m[0];
	s_m[1] /= l.s_m[1];
	s_m[2] /= l.s_m[2];

	return *this;
}


Light  Light::operator* ( const float f ) const
{
	return Light( s_m[0] * f,
	              s_m[1] * f,
	              s_m[2] * f );
}


Light& Light::operator*=( const float f )
{
	s_m[0] *= f;
	s_m[1] *= f;
	s_m[2] *= f;

	return *this;
}


Light  Light::operator/ ( const float f ) const
{
	const float oneOverF = 1.0f / f;

	return Light( s_m[0] * oneOverF,
	              s_m[1] * oneOverF,
	              s_m[2] * oneOverF );
}


Light& Light::operator/=( const float f )
{
	const float oneOverF = 1.0f / f;

	s_m[0] *= oneOverF;
	s_m[1] *= oneOverF;
	s_m[2] *= oneOverF;

	return *this;
}


Light perceptuum2::interaction::operator*( const float f, const Light& l )
{
	return Light( f * l.s_m[0],
	              f * l.s_m[1],
	              f * l.s_m[2] );
}


Light perceptuum2::interaction::operator/( const float f, const Light& l )
{
	return Light( f / l.s_m[0],
	              f / l.s_m[1],
	              f / l.s_m[2] );
}




/// logical ---------------------------------------------------------------------------------------
bool Light::operator==( const Light& l ) const
{
	return (s_m[0] == l.s_m[0]) &
	       (s_m[1] == l.s_m[1]) &
	       (s_m[2] == l.s_m[2]);
}


bool Light::operator!=( const Light& l ) const
{
	return (s_m[0] != l.s_m[0]) |
	       (s_m[1] != l.s_m[1]) |
	       (s_m[2] != l.s_m[2]);
}


bool Light::isZero() const
{
	return bool( (s_m[0] == 0.0f) & (s_m[1] == 0.0f) & (s_m[2] == 0.0f) );
}




/// constants --------------------------------------------------------------------------------------
const Light& Light::ZERO()
{
	static const Light k( CieXyz::ZERO() );
	return k;
}


const Light& Light::ONE()
{
	static const Light k( CieXyz::ONE() );
	return k;
}


const Light& Light::EPSILON()
{
	static const Light k( CieXyz::EPSILON() );
	return k;
}


const Light& Light::ALMOST_ONE()
{
	static const Light k( CieXyz::ALMOST_ONE() );
	return k;
}


const Light& Light::MIN()
{
	static const Light k( CieXyz::MIN() );
	return k;
}


const Light& Light::MAX()
{
	static const Light k( CieXyz::MAX() );
	return k;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::interaction::operator<<( SerializerOut& out, const Light& obj )
{
	out.write( "X", obj.s_m[0] );
	out.write( "Y", obj.s_m[1] );
	out.write( "Z", obj.s_m[2] );

	return out;
}


SerializerIn& perceptuum2::interaction::operator>>( SerializerIn& in, Light& obj )
{
	in.read( "X", obj.s_m[0] );
	in.read( "Y", obj.s_m[1] );
	in.read( "Z", obj.s_m[2] );

	return in;
}


const char* Light::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::interaction::Light";
	return CLASS_NAME;
}


const char* Light::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a light measurement (current implementation is CIE XYZ)";
	return CLASS_DESCRIPTION;
}
