/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "TriangleExplicit.h"
#include "UnitVector3f.h"
#include "Serialization.h"
#include "hxa7241math.h"
#include "hxa7241vector.h"

#include "Mesh.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// statics ----------------------------------------------------------------------------------------
const char Mesh::EXCEPTION_MESSAGE_1[] = "invalid polyhedron structure";
const char Mesh::EXCEPTION_MESSAGE_2[] = "invalid number of vertexs";
const char Mesh::EXCEPTION_MESSAGE_3[] = "invalid number of triangles";
const char Mesh::EXCEPTION_MESSAGE_4[] = "triangle index out of range";
const char Mesh::EXCEPTION_MESSAGE_5[] = "triangle vertexs have same index";




/// standard object services -----------------------------------------------------------------------
Mesh::Mesh()
 :	vertexs_m  (),
	triangles_m(),
	position_m (),
	scaling_m  ()
{
}


Mesh::Mesh
(
	const vector<Vector3f>&         vertexPositions,
	const vector<TriangleImplicit>& triangles
)
 :	vertexs_m  (),
	triangles_m(),
	position_m (),
	scaling_m  ()
{
	Mesh::setMesh( vertexPositions, triangles );
}


Mesh::~Mesh()
{
}


Mesh::Mesh( const Mesh& other )
 :	vertexs_m  ( other.vertexs_m ),
	triangles_m( other.triangles_m ),
	position_m ( other.position_m ),
	scaling_m  ( other.scaling_m )
{
}


Mesh& Mesh::operator=( const Mesh& other )
{
	if( &other != this )
	{
		hxa7241::utility::assignUnreserve( other.vertexs_m, vertexs_m );
		hxa7241::utility::assignUnreserve( other.triangles_m, triangles_m );

		position_m  = other.position_m;
		scaling_m   = other.scaling_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Mesh::setMesh
(
	const vector<Vector3f>&         vertexPositions,
	const vector<TriangleImplicit>& triangles
)
{
	makeTriangles( vertexPositions, triangles, triangles_m );

	vector<UnitVector3f> vertexNormals;
	makeVertexNormals( vertexPositions, triangles_m, vertexNormals );

	makeVertexs( vertexPositions, vertexNormals, vertexs_m, position_m, scaling_m );
}




/// queries ----------------------------------------------------------------------------------------
void Mesh::getMinMax
(
	Vector3f& min,
	Vector3f& max
) const
{
	/// scaling_m  = max - min;
	/// position_m = (scaling_m * 0.5f) + min;

	((min = scaling_m) *= 0.5f).negate() += position_m;
	(max = scaling_m) += min;
}


void Mesh::getVertexPositions
(
	vector<Vector3f>& vertexPositions
) const
{
	vertexPositions.resize( vertexs_m.size() );

	for( dword i = vertexs_m.size();  i-- > 0; )
	{
		vertexs_m[i].getPosition( position_m, scaling_m, vertexPositions[i] );
	}
}


void Mesh::getTriangle
(
	const dword       triangleIndex,
	TriangleExplicit& triangle
) const
{
	const TriangleImplicit& tri = triangles_m[triangleIndex];

	triangle.set( vertexs_m[tri.getVertexIndex0()],
	              vertexs_m[tri.getVertexIndex1()],
	              vertexs_m[tri.getVertexIndex2()],
	              position_m,
	              scaling_m,
	              tri.getInterpolateNormal() );
}


const TriangleImplicit& Mesh::getTriangle
(
	const dword index
) const
{
	return triangles_m[index];
}


const vector<Vertex>& Mesh::getVertexs() const
{
	return vertexs_m;
}


const vector<TriangleImplicit>& Mesh::getTriangles() const
{
	return triangles_m;
}


const Vector3f& Mesh::getPosition() const
{
	return position_m;
}


const Vector3f& Mesh::getScale() const
{
	return scaling_m;
}




/// implementation ---------------------------------------------------------------------------------
void Mesh::makeTriangles
(
	const vector<Vector3f>&         vertexPositions,
	const vector<TriangleImplicit>& trianglesIn,
	vector<TriangleImplicit>&       trianglesOut
)
{
	hxa7241::utility::assignUnreserve( trianglesIn, trianglesOut );

	std::map<dword, Mesh::Edge> edges;

	checkValidity( vertexPositions, trianglesOut, edges );
	normaliseTriangleOrientations( vertexPositions, edges, trianglesOut );
}


void Mesh::makeVertexs
(
	const vector<Vector3f>&     vertexPositions,
	const vector<UnitVector3f>& vertexNormals,
	vector<Vertex>&             vertexs,
	Vector3f&                   position,
	Vector3f&                   scaling
)
{
	/// find min and max
	Vector3f min( Vector3f::MAX() );
	Vector3f max( Vector3f::MIN() );
	for( dword i = vertexPositions.size();  i-- > 0; )
	{
		const Vector3f& vertPos = vertexPositions[i];

		min.clampToMaxOf( vertPos );
		max.clampToMinOf( vertPos );
	}

	/// calc compression factors
	scaling  = max - min;
	position = (scaling * 0.5f) + min;
	const Vector3f oneOverScaling( Vector3f::ONE() / scaling );

	/// compress vertex values
	const dword vertexCount = vertexPositions.size();
	hxa7241::utility::resizeReserve( vertexs, vertexCount );
	for( int i = vertexCount;  i-- > 0; )
	{
		vertexs[i].set( position, oneOverScaling, vertexPositions[i], vertexNormals[i] );
	}
}


void Mesh::checkValidity
(
	const vector<Vector3f>&         vertexPositions,
	const vector<TriangleImplicit>& triangles,
	std::map<dword, Edge>&          edges
)
{
	/// should be: a single, non-intersecting, closed surface

	static const char METHOD_NAME[] = "Mesh::checkValidity";

	const dword faceCount   = triangles.size();
	const dword vertexCount = vertexPositions.size();

	/// check triangle euler relation (V - 0.5F = 2*(1 - Genus))
	{
		/// F should be == even integer
		/// V - 0.5F should be == even integer and <= 2
		const dword vmhf = vertexCount - (faceCount >> 1);
		if( bool( (faceCount & 1) | (vmhf & 1) | int(vmhf > 2) ) )
		{
			throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_1, CLASS_NAME(), METHOD_NAME );
		}
	}

	/// check face edge relation
	dword edgeCount;
	{
		/// every edge should be shared by two faces

		/// build edge list
		for( dword i = faceCount;  i-- > 0; )
		{
			const dword verts[] = { triangles[i].getVertexIndex0(), triangles[i].getVertexIndex2(),
			                        triangles[i].getVertexIndex1(), triangles[i].getVertexIndex0() };

			for( int v = 4;  v-- > 1; )
			{
				const dword edge = Edge::makeKey( verts[v], verts[v-1] );

				const bool moreThanTwoFaces = edges[ edge ].addFace( i );
				/// check every edge has <= 2 faces
				if( moreThanTwoFaces )
				{
					throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_1, CLASS_NAME(), METHOD_NAME );
				}
			}
		}

		/// check every edge has 2 faces
		for( std::map<dword, Edge>::const_iterator it = edges.begin();  it != edges.end();  ++it )
		{
			if( !it->second.hasFaces() )
			{
				throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_1, CLASS_NAME(), METHOD_NAME );
			}
		}

		edgeCount = edges.size();
	}

	/// check general euler relation (F + V - E = 2*(1 - Genus))
	{
		/// F + V - E should be == even integer and <= 2
		/// (multiple objects, unclosed, dangling pieces)
		const dword fpvme = faceCount + vertexCount - edgeCount;
		if( bool( (fpvme & 1) | int(fpvme > 2) ) )
		{
			throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_1, CLASS_NAME(), METHOD_NAME );
		}
	}

	/// check self intersection
	{
		/// index triangles

		///	...
	}
}


void Mesh::normaliseTriangleOrientations
(
	const vector<Vector3f>&      vertexPositions,
	const std::map<dword, Edge>& edges,
	vector<TriangleImplicit>&    triangles
)
{
	/// make supplemental array of triangle adjacencies
	vector<AdjacentTriangles> adjacents( triangles.size() );

	for( std::map<dword, Edge>::const_iterator it = edges.begin();  it != edges.end();  ++it )
	{
		//dword v0;
		//dword v1;
		//Edge::getVertexsFromKey( it->first, v0, v1 );
		const dword key = it->first;

		const Edge& edge = it->second;
		const dword faces[2] = { edge.getFace0(), edge.getFace1() };
		for( dword i = 2;  i-- > 0; )
		{
			const TriangleImplicit& triangle = triangles[ faces[i] ];

			dword adjacent;
			if( key == Edge::makeKey( triangle.getVertexIndex0(), triangle.getVertexIndex1() ) )
			{
				adjacent = 0;
			}
			else if( key == Edge::makeKey( triangle.getVertexIndex1(), triangle.getVertexIndex2() ) )
			{
				adjacent = 1;
			}
			else if( key == Edge::makeKey( triangle.getVertexIndex2(), triangle.getVertexIndex0() ) )
			{
				adjacent = 2;
			}
			else
			{
				throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_1, CLASS_NAME(), "normaliseTriangleOrientations" );
			}

			//dword adjacent;
			//if( ((triangle.getVertex0() == v0) & (triangle.getVertex1() == v1)) ||
			//    ((triangle.getVertex1() == v0) & (triangle.getVertex0() == v1)) )
			//{
			//	adjacent = 0;
			//}
			//else if( ((triangle.getVertex1() == v0) & (triangle.getVertex2() == v1)) ||
			//         ((triangle.getVertex2() == v0) & (triangle.getVertex1() == v1)) )
			//{
			//	adjacent = 1;
			//}
			//else if( ((triangle.getVertex2() == v0) & (triangle.getVertex0() == v1)) ||
			//         ((triangle.getVertex0() == v0) & (triangle.getVertex2() == v1)) )
			//{
			//	adjacent = 2;
			//}

			adjacents[ faces[i] ].index_m[ adjacent ] = faces[ i ^ 1 ];
			adjacents[ faces[i] ].done_m              = false;
		}
	}


	/// find extremal triangle
	dword             extremalTriangle;
	TriangleImplicit* pExtremalTriangle;
	{
		/// find extremal triangles
		vector<const TriangleImplicit*> extremalTriangles;
		dword minSlopeEdgeVertI = 0;
		{
			/// find min (z) vertex by linear search
			dword minI;
			float minZ = FLOAT_MAX;
			for( dword i = vertexPositions.size();  i-- > 0; )
			{
				const float z = vertexPositions[i].getZ();
				if( z < minZ )
				{
					minZ = z;
					minI = i;
				}
			}

			/// find triangles using that vertex by linear search
			for( dword i = triangles.size();  i-- > 0; )
			{
				const TriangleImplicit& triangle = triangles[i];
				if( (triangle.getVertexIndex0() == minI) |
					(triangle.getVertexIndex1() == minI) |
					(triangle.getVertexIndex2() == minI) )
				{
					extremalTriangles.push_back( &(triangles[i]) );
				}
			}

			/// find min slope edge other vertex
			float minDot = 1.0f;
			for( dword i = extremalTriangles.size();  i-- > 0; )
			{
				const TriangleImplicit& tri = *(extremalTriangles[i]);
				const dword verts[] = { tri.getVertexIndex0(), tri.getVertexIndex1(), tri.getVertexIndex2() };

				/// step thru verts, except min z vert
				for( int j = 3;  j-- > 0; )
				{
					const dword vertIndex = verts[j];
					if( vertIndex != minI )
					{
						/// calc slope of edge between vert and min z vert
						const UnitVector3f edgeDir( vertexPositions[vertIndex] - vertexPositions[minI] );
						float edotz = -(edgeDir.dot( UnitVector3f::Z() ));
						if( edotz < 0.0f )
						{
							edotz = -edotz;
						}

						/// find min slope, save vert
						if( edotz < minDot )
						{
							minDot            = edotz;
							minSlopeEdgeVertI = vertIndex;
						}
					}
				}
			}
		}

		/// find triangle containing min slope edge, that has min xy/z slope - most parallel to xy plane
		{
			float maxDot = 0.0f;
			for( dword i = extremalTriangles.size();  i-- > 0; )
			{
				const TriangleImplicit& triangle = *(extremalTriangles[i]);
				if( (triangle.getVertexIndex0() == minSlopeEdgeVertI) |
					(triangle.getVertexIndex1() == minSlopeEdgeVertI) |
					(triangle.getVertexIndex2() == minSlopeEdgeVertI) )
				{
					/// get triangle normal using cross product of edges, unitized
					const Vector3f edge0( vertexPositions[ triangle.getVertexIndex1() ] - vertexPositions[ triangle.getVertexIndex0() ] );
					const Vector3f edge1( vertexPositions[ triangle.getVertexIndex2() ] - vertexPositions[ triangle.getVertexIndex1() ] );
					UnitVector3f normal( edge0.cross( edge1 ) );

					/// choose the triangle with the largest abs(normal.-Z)
					float ndotz = -(normal.dot( UnitVector3f::Z() ));
					if( ndotz < 0.0f )
					{
						ndotz = -ndotz;
					}
					if( ndotz > maxDot )
					{
						maxDot = ndotz;
						extremalTriangle = &triangle - &(triangles[0]);
					}
				}
			}
			pExtremalTriangle = &(triangles[extremalTriangle]);
		}
	}


	///	orient extremal triangle
	{
		/*const Vector3f& vert0 = vertexPositions[ pExtremalTriangle->getVertexIndex0() ];
		const Vector3f& vert1 = vertexPositions[ pExtremalTriangle->getVertexIndex1() ];
		const Vector3f& vert2 = vertexPositions[ pExtremalTriangle->getVertexIndex2() ];

		/// ignore z
		using hxa7241::graphics::Vector2f;
		const Vector2f verts2D[] = { Vector2f( vert0.getX(), vert0.getY() ),
		                             Vector2f( vert1.getX(), vert1.getY() ),
		                             Vector2f( vert2.getX(), vert2.getY() ) };
		/// vertex with min y is first
		const dword first  = vert1.getY() < vert0.getY() ? (vert2.getY() < vert1.getY() ? 2 : 1) : (vert2.getY() < vert0.getY() ? 2 : 0);
		      dword second = (first + 1) % 3;
		      dword third  = (first + 2) % 3;

		/// dot joined edges with X
		const float d2 = Vector2f::X().dot( (verts2D[second] - verts2D[first]).unitizeEq() );
		const float d3 = Vector2f::X().dot( (verts2D[third]  - verts2D[first]).unitizeEq() );

		/// choose min dot vertex as second
		/// remaining vertex is third
		if( d3 < d2 )
		{
			const dword t = second;
			second = third;
			third  = t;
		}

		/// rearrange triangle
		const dword vertIndexs[] = { pExtremalTriangle->getVertexIndex0(), pExtremalTriangle->getVertexIndex1(), pExtremalTriangle->getVertexIndex2() };
		pExtremalTriangle->set( word(vertIndexs[first]), word(vertIndexs[second]), word(vertIndexs[third]) );*/


		const Vector3f& vert0 = vertexPositions[ pExtremalTriangle->getVertexIndex0() ];
		const Vector3f& vert1 = vertexPositions[ pExtremalTriangle->getVertexIndex1() ];
		const Vector3f& vert2 = vertexPositions[ pExtremalTriangle->getVertexIndex2() ];

		/// ignore z
		using hxa7241::graphics::Vector2f;
		const Vector2f verts2D[] = { Vector2f( vert0.getX(), vert0.getY() ),
		                             Vector2f( vert1.getX(), vert1.getY() ),
		                             Vector2f( vert2.getX(), vert2.getY() ) };
		/// vertex with min y is first
		const dword first  = vert1.getY() < vert0.getY() ? (vert2.getY() < vert1.getY() ? 2 : 1) : (vert2.getY() < vert0.getY() ? 2 : 0);
		const dword second = (first + 1) % 3;
		const dword third  = (first + 2) % 3;

		/// dot joined edges with X
		const float d2 = Vector2f::X().dot( (verts2D[second] - verts2D[first]).unitizeEq() );
		const float d3 = Vector2f::X().dot( (verts2D[third]  - verts2D[first]).unitizeEq() );

		/// choose min dot vertex as second
		/// remaining vertex is third
		if( d3 < d2 )
		{
			/// rearrange triangle
			pExtremalTriangle->reverseOrientation();
		}
	}


	///	orient all other triangles
	adjacents[extremalTriangle].done_m = true;
	orientTriangles( extremalTriangle, adjacents[extremalTriangle].index_m[0], adjacents, triangles );
}


void Mesh::orientTriangles
(
	const dword                lastAdjacentTriangle,
	const dword                triangle,
	vector<AdjacentTriangles>& adjacents,
	vector<TriangleImplicit>&  triangles
)
{
	/// orient triangle
	{
		TriangleImplicit&       triangleRef     = triangles[triangle];
		const TriangleImplicit& lastTriangleRef = triangles[lastAdjacentTriangle];

		const dword verts[]     = { triangleRef.getVertexIndex0(),     triangleRef.getVertexIndex1(),     triangleRef.getVertexIndex2() };
		const dword vertsLast[] = { lastTriangleRef.getVertexIndex0(), lastTriangleRef.getVertexIndex1(), lastTriangleRef.getVertexIndex2() };

		/// find shared vertex
		dword match     = -1;
		dword matchLast =  3;
		for( ;  (match == -1) && (--matchLast >= 0); )
		{
			for( match = 3;  (--match >= 0) && (verts[match] != vertsLast[matchLast]); );
		}

		/// check if same ordering
		if( (verts[(match + 1) % 3] == vertsLast[(matchLast + 1) % 3]) ||
		    (verts[(match + 2) % 3] == vertsLast[(matchLast + 2) % 3]) )
		{
			/// reverse vertex ordering
			//triangleRef.set( word(verts[2]), word(verts[1]), word(verts[0]) );
			triangleRef.reverseOrientation();
		}

		/// set done flag
		adjacents[triangle].done_m = true;
	}


	/// recurse
	for( dword i = 3;  i-- > 0; )
	{
		const dword nextAdjacentTri = adjacents[triangle].index_m[i];
		if( !(adjacents[nextAdjacentTri]).done_m )
		{
			orientTriangles( triangle, nextAdjacentTri, adjacents, triangles );
		}
	}
}


void Mesh::makeVertexNormals
(
	const vector<Vector3f>&         vertexPositions,
	const vector<TriangleImplicit>& triangles,
	vector<UnitVector3f>&           vertexNormals
)
{
	/// allocate vertex normals sums
	vector<Vector3f> normalsSum( vertexPositions.size() );

	/// step through triangles
	for( int i = triangles.size();  i-- > 0; )
	{
		const TriangleImplicit& triangle = triangles[i];

		/// get triangle normal
		UnitVector3f normal;
		triangle.getNormal( vertexPositions, normal );

		/// get triangle edges
		const UnitVector3f edge0( vertexPositions[ triangle.getVertexIndex1() ] - vertexPositions[ triangle.getVertexIndex0() ] );
		const UnitVector3f edge1( vertexPositions[ triangle.getVertexIndex2() ] - vertexPositions[ triangle.getVertexIndex1() ] );
		const UnitVector3f edge2( vertexPositions[ triangle.getVertexIndex0() ] - vertexPositions[ triangle.getVertexIndex2() ] );

		/// add normal weighted by angle, to each vertexs normal sum
		normalsSum[ triangle.getVertexIndex0() ] += normal.getVector() * float(::acos( -(edge0.dot( edge2 )) ));
		normalsSum[ triangle.getVertexIndex1() ] += normal.getVector() * float(::acos( -(edge1.dot( edge0 )) ));
		normalsSum[ triangle.getVertexIndex2() ] += normal.getVector() * float(::acos( -(edge2.dot( edge1 )) ));
	}

	/// unitize vertex normal sums
	vertexNormals.resize( normalsSum.size() );
	for( int i = vertexNormals.size();  i-- > 0; )
	{
		vertexNormals[i] = normalsSum[i];
	}
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const Mesh& obj )
{
	/// write vertexs
	{
		/// write vertex count
		const dword vertexCount = obj.vertexs_m.size();
		out.write( "vertexCount", vertexCount );

		/// write vertexs
		for( int i = 0;  i < vertexCount;  ++i )
		{
			Vector3f vertexPosition;
			obj.vertexs_m[i].getPosition( obj.position_m, obj.scaling_m, vertexPosition );

			out.write( "vertex", vertexPosition );
		}
	}

	/// write triangles
	{
		/// write triangle count
		const dword triangleCount = obj.vertexs_m.size();
		out.write( "triangleCount", triangleCount );

		/// write triangles
		for( int i = 0;  i < triangleCount;  ++i )
		{
			out.write( "triangle", obj.triangles_m[i] );
		}
	}

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, Mesh& obj )
{
	static const char METHOD_NAME[] = "Mesh::operator>>SerializerIn";


	/// read vertex positions
	vector<Vector3f> vertexPositions;
	{
		dword vertexCount;
		in.read( "vertexCount", vertexCount );

		if( (vertexCount > obj.VERTEX_COUNT_MAX) |
		    (vertexCount < 0) )
		{
			throw hxa7241::utility::Exception( obj.EXCEPTION_MESSAGE_2, obj.CLASS_NAME(), METHOD_NAME );
		}

		vertexPositions.resize( vertexCount );

		for( int i = 0;  i < vertexCount;  ++i )
		{
			in.read( "vertex", vertexPositions[i] );
		}
	}


	/// read triangles
	vector<TriangleImplicit> triangles;
	{
		dword triangleCount;
		in.read( "triangleCount", triangleCount );

		if( triangleCount < 0 )
		{
			throw hxa7241::utility::Exception( obj.EXCEPTION_MESSAGE_3, obj.CLASS_NAME(), METHOD_NAME );
		}

		const dword vertexCount = vertexPositions.size();

		triangles.resize( triangleCount );

		for( int i = 0;  i < triangleCount;  ++i )
		{
			in.read( "triangle", triangles[i] );

			if( !triangles[i].checkIndexesRange( vertexCount ) )
			{
				throw hxa7241::utility::Exception( obj.EXCEPTION_MESSAGE_4, obj.CLASS_NAME(), METHOD_NAME );
			}
			if( !triangles[i].checkIndexesDistinct() )
			{
				throw hxa7241::utility::Exception( obj.EXCEPTION_MESSAGE_5, obj.CLASS_NAME(), METHOD_NAME );
			}
		}
	}


	try
	{
		/// set mesh
		obj.setMesh( vertexPositions, triangles );

		return in;
	}
	catch( ... )
	{
		/// zero object
		obj = Mesh();

		throw;
	}
}


const char* Mesh::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::Mesh";
	return CLASS_NAME;
}


const char* Mesh::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a triangle mesh with shared vertex pool";
	return CLASS_DESCRIPTION;
}




/// Mesh::Edge:: methods -----------------------------------------------------------------------------------
bool Mesh::Edge::addFace
(
	const dword f
)
{
	bool moreThanTwoFaces = false;

	if( face0_m == FACE_DEFAULT )
	{
		face0_m = f;
	}
	else if( face1_m == FACE_DEFAULT )
	{
		face1_m = f;
	}
	else
	{
		moreThanTwoFaces = true;
	}

	return moreThanTwoFaces;
}


bool Mesh::Edge::hasFaces() const
{
	return (face0_m != FACE_DEFAULT) && (face1_m != FACE_DEFAULT);
}


dword Mesh::Edge::makeKey
(
	dword v0,
	dword v1
)
{
	/// order so higher index in hi word
	if( v1 < v0 )
	{
		const temp = v0;
		v0 = v1;
		v1 = temp;
	}

	return (v0 << 16) | (v1 & 0xFFFF);
}


void Mesh::Edge::getVertexsFromKey
(
	const dword key,
	dword&      v0,
	dword&      v1
)
{
	v1 = key & 0xFFFF;
	v0 = (key >> 16) & 0xFFFF;
}
