/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Mesh_h
#define Mesh_h


#include "Primitives.h"
#include "Vertex.h"
#include "TriangleImplicit.h"
#include "Vector3f.h"
#include <vector>
#include <map>

#include "hxa7241_graphics.h"
#include "hxa7241_utility.h"




#include "perceptuum2_modelling.h"
namespace perceptuum2 { namespace modelling {
	using namespace hxa7241::primitives;
	using hxa7241::graphics::Vector3f;
	using hxa7241::graphics::UnitVector3f;
	using hxa7241::utility::SerializerOut;
	using hxa7241::utility::SerializerIn;
	using std::vector;


/**
 * a single, non-intersecting, closed surface, made of triangle faces.
 *
 * validates structure when setting or deserializing.
 */
class Mesh
{
/// standard object services -----------------------------------------------------------------------
public:
	         Mesh();
	         Mesh( const vector<Vector3f>&         vertexPositions,
	               const vector<TriangleImplicit>& triangles );

	        ~Mesh();
	         Mesh( const Mesh& );
	Mesh& operator=( const Mesh& );


/// commands ---------------------------------------------------------------------------------------
public:
	        void  setMesh( const vector<Vector3f>&         vertexPositions,
	                       const vector<TriangleImplicit>& triangles );


/// queries ----------------------------------------------------------------------------------------
public:
	        void  getMinMax( Vector3f& min,
	                         Vector3f& max )                                                  const;
	        void  getVertexPositions( vector<Vector3f>& )                                     const;
	        void  getTriangle( dword             index,
	                           TriangleExplicit& triangle )                                   const;
	        const TriangleImplicit& getTriangle( dword index )                                const;

	        const vector<Vertex>&           getVertexs()                                      const;
	        const vector<TriangleImplicit>& getTriangles()                                    const;
	        const Vector3f&                 getPosition()                                     const;
	        const Vector3f&                 getScale()                                        const;


/// serialization ----------------------------------------------------------------------------------
	friend  SerializerOut& operator<<( SerializerOut&, const Mesh& );
	friend  SerializerIn&  operator>>( SerializerIn&,        Mesh& );
	static const char* CLASS_NAME();
	static const char* CLASS_DESCRIPTION();


/// implementation ---------------------------------------------------------------------------------
protected:
	class Edge
	{
	public:
		Edge() : face0_m(FACE_DEFAULT), face1_m(FACE_DEFAULT) {}
		~Edge() {}
		Edge( const Edge& other ) : face0_m( other.face0_m ), face1_m( other.face1_m ) {}
		Edge& operator=( const Edge& other ) { face0_m = other.face0_m;   face1_m = other.face1_m;   return *this;   }

		bool  addFace( dword f );

		dword getFace0() const { return face0_m; }
		dword getFace1() const { return face1_m; }
		bool  hasFaces() const;

		static dword makeKey( dword v0, dword v1 );
		static void  getVertexsFromKey( dword key, dword& v0, dword& v1 );

		static const dword FACE_DEFAULT = -1;

	private:
		dword face0_m;
		dword face1_m;
	};

	struct AdjacentTriangles
	{
		dword index_m[3];
		bool  done_m;
	};

	static  void  makeTriangles( const vector<Vector3f>&         vertexPositions,
	                             const vector<TriangleImplicit>& trianglesIn,
	                             vector<TriangleImplicit>&       trianglesOut );
	static  void  makeVertexs( const vector<Vector3f>&     vertexPositions,
	                           const vector<UnitVector3f>& vertexNormals,
	                           vector<Vertex>&             vertexs,
	                           Vector3f&                   position,
	                           Vector3f&                   scaling );
	static  void  checkValidity( const vector<Vector3f>&         vertexPositions,
	                             const vector<TriangleImplicit>& triangles,
	                             std::map<dword, Edge>&          edges );
	static  void  normaliseTriangleOrientations( const vector<Vector3f>&      vertexPositions,
	                                             const std::map<dword, Edge>& edges,
	                                             vector<TriangleImplicit>&    triangles );
	static  void  orientTriangles( dword lastAdjacentTriangle,
	                               dword triangle,
	                               vector<AdjacentTriangles>&,
	                               vector<TriangleImplicit>& );
	static  void  makeVertexNormals( const vector<Vector3f>&         vertexPositions,
	                                 const vector<TriangleImplicit>& triangles,
	                                 vector<UnitVector3f>&           vertexNormals );


/// fields -----------------------------------------------------------------------------------------
private:
	vector<Vertex>           vertexs_m;
	vector<TriangleImplicit> triangles_m;

	Vector3f position_m;
	Vector3f scaling_m;


	static const dword VERTEX_COUNT_MAX = 32767;

	static const char EXCEPTION_MESSAGE_1[];
	static const char EXCEPTION_MESSAGE_2[];
	static const char EXCEPTION_MESSAGE_3[];
	static const char EXCEPTION_MESSAGE_4[];
	static const char EXCEPTION_MESSAGE_5[];
};


}}




#endif//Mesh_h
