/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Vector3f.h"
#include "UnitVector3f.h"
#include "CieXyz.h"
#include "ClipFlags.h"
#include "Intersection.h"
#include "Qualities.h"
#include "ProjectableObjectRef.h"
#include "Vertex.h"
#include "TriangleImplicit.h"

#include "ObjectDefault.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
ObjectDefault::ObjectDefault
(
	const Scene* pScene
)
 :	ObjectRenderable( pScene ),
	refractiveIndex_m()
{
}


//ObjectDefault::ObjectDefault
//(
//	const Scene*  pScene,
//	const CieXyz& color
//)
// :	ObjectRenderable( pScene ),
//	refractiveIndex_m()
//{
//	ObjectDefault::setColor( color );
//}


ObjectDefault::~ObjectDefault()
{
}


ObjectDefault::ObjectDefault( const ObjectDefault& other )
 :	ObjectRenderable( other ),
	color_m( other.color_m ),
	transparency_m( other.transparency_m ),
	refractiveIndex_m( other.refractiveIndex_m )
{
}


ObjectDefault& ObjectDefault::operator=( const ObjectDefault& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void ObjectDefault::setMedium
(
	const Float01 transparency,
	const float   refractiveIndex
)
{
	transparency_m    = transparency;
	refractiveIndex_m = refractiveIndex;
}


//void ObjectDefault::setColor
//(
//	const CieXyz& color
//)
//{
//	color_m = color;
//}


bool ObjectDefault::update
(
	const float //time
)
{
	return false;
}




/// queries ----------------------------------------------------------------------------------------
bool ObjectDefault::getEmitted
(
	const Vector3f& ,//posRands,
	vector<float>&  ,//compositionStorage,
	Intersection&   intersection,
	Qualities&      qualities
) const
{
	intersection = Intersection();

	qualities.normal_m               = -UnitVector3f::Z();
	qualities.tangent_m              = UnitVector3f::X();
	qualities.reflectivityDiffuse_m  = color_m;
	qualities.reflectivitySpecular_m = color_m;
	qualities.roughnessX_m           = FLOAT_ALMOST_ONE;
	qualities.roughnessY_m           = FLOAT_ALMOST_ONE;
	qualities.transparency_m         = 0.0f;
	qualities.refractiveIndex_m      = 1.0f;
	qualities.emitted_m              = CieXyz::ZERO();
	qualities.emittedRoughness_m     = 0.0f;

	return false;
}


bool ObjectDefault::getIntersection
(
	const ObjectRenderableId& ,//objectId,
	const Ray&                ,//ray,
	float&                    distance,
	Intersection&             intersection
) const
{
	intersection.set( Vector3f::ZERO(), -UnitVector3f::Z(), Vector2f(), ObjectRenderableId() );
	distance = 0.0f;

	return false;
}


void ObjectDefault::getObjectsForProjection
(
	const Transclip&              ,//transform,
	const ClipFlags               ,//clipFlags,
	vector<ProjectableObjectRef>& objects
) const
{
	objects.resize( 0 );
}


void ObjectDefault::getMesh
(
	Mesh&        ,//meshBuffer,
	const Mesh** ppMesh
) const
{
	*ppMesh = 0;
}


void ObjectDefault::getQualities
(
	const Intersection& ,//intersection,
	Qualities&          qualities
) const
{
	qualities.normal_m               = -UnitVector3f::Z();
	qualities.tangent_m              = UnitVector3f::X();
	qualities.reflectivityDiffuse_m  = color_m;
	qualities.reflectivitySpecular_m = color_m;
	qualities.roughnessX_m           = FLOAT_ALMOST_ONE;
	qualities.roughnessY_m           = FLOAT_ALMOST_ONE;
	qualities.transparency_m         = 0.0f;
	qualities.refractiveIndex_m      = 1.0f;
	qualities.emitted_m              = CieXyz::ZERO();
	qualities.emittedRoughness_m     = 0.0f;
}


Float01 ObjectDefault::getTransparency
(
	const ObjectRenderableId& ,//objectId,
	const Vector3f&           //position
) const
{
	return transparency_m;
}


float ObjectDefault::getRefractiveIndex
(
	const ObjectRenderableId& ,//objectId,
	const Vector3f&           //position
) const
{
	return refractiveIndex_m;
}


const CieXyz& ObjectDefault::getEmittance
(
	const ObjectRenderableId&
) const
{
	return CieXyz::ZERO();
}


const char* ObjectDefault::getClassName() const
{
	return "perceptuum2::modelling::ObjectDefault";
}




/// implementation ---------------------------------------------------------------------------------
void ObjectDefault::assign
(
	const ObjectRenderable& other
)
{
	if( &other != this )
	{
		ObjectRenderable::assign( other );

		color_m           = CieXyz::ZERO();
		transparency_m    = 0.0f;
		refractiveIndex_m = 1.0f;
	}
}


void ObjectDefault::assign
(
	const ObjectDefault& other
)
{
	if( &other != this )
	{
		ObjectRenderable::assign( other );

		color_m           = other.color_m;
		transparency_m    = other.transparency_m;
		refractiveIndex_m = other.refractiveIndex_m;
	}
}
