/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Intersection.h"
#include "Qualities.h"
#include "Scene.h"
#include "Ray.h"
#include "TriangleExplicit.h"
#include "ProjectableObjectRef.h"
#include "Histogram.h"
#include "Serialization.h"
#include "ClipFlags.h"
#include "hxa7241math.h"
#include "hxa7241vector.h"
#include "Logger.h"
#include <algorithm>

#include "ObjectMesh.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
ObjectMesh::ObjectMesh
(
	const Scene*  pScene,
	SerializerIn& in,
	const bool    b
)
 :	ObjectRenderable( pScene ),
	mesh_m(),
	materialId_m()
{
	if( b )
	{
		in.read( "object", *this );
	}
	else
	{
		in >> *this;
	}
}


ObjectMesh::ObjectMesh
(
	const Scene* pScene
)
 :	ObjectRenderable( pScene ),
	mesh_m(),
	materialId_m()
{
}


ObjectMesh::~ObjectMesh()
{
}


ObjectMesh::ObjectMesh( const ObjectMesh& other )
 :	ObjectRenderable( other ),
	mesh_m          ( other.mesh_m ),
	materialId_m    ( other.materialId_m )
{
}


ObjectMesh& ObjectMesh::operator=( const ObjectMesh& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
bool ObjectMesh::update
(
	const float //time
)
{
	return false;
}




/// queries ----------------------------------------------------------------------------------------
bool ObjectMesh::getEmitted
(
	const Vector3f& posRands,
	vector<float>&  compositionStorage,
	Intersection&   intersection,
	Qualities&      qualities
) const
{
	/// choose triangle with emittance composition and roulette
	dword triangleIndex = 0;
	bool  isAnyEmitted  = false;
	if( getScene() != 0 )
	{
		/// get vertex positions
		vector<Vector3f> vertexPositions;
		mesh_m.getVertexPositions( vertexPositions );

		const vector<TriangleImplicit>& triangles = mesh_m.getTriangles();
		const dword     trianglesSize = triangles.size();
		const Material& material      = getScene()->getMaterial( materialId_m );
		hxa7241::utility::Histogram composition( compositionStorage, trianglesSize );

		/// step through triangles adding emittance to composition
		for( dword i = 0;  i < trianglesSize;  ++i )
		{
			const TriangleImplicit& triangle = triangles[i];

			/// get mean radiant exitance
			CieXyz emittedMean;
			{
				Vector2<word> textureCoords[3];
				triangle.getTextureCoords( textureCoords );

				material.getEmittedMean( textureCoords, emittedMean );
			}

			/// get area
			const float area = triangle.getArea( vertexPositions );

			/// add triangle emittance to composition
			composition.addElement( emittedMean.total() * area );
		}

		isAnyEmitted = composition.findElement( posRands.getZ(), triangleIndex );
	}


	if( isAnyEmitted )
	{
		/// get object-triangle id
		intersection.object_m = ObjectRenderableId( intersection.object_m, triangleIndex & 0xFFFF );

		/// get barycentric coords on triangle
		{
			Vector2<word> textureCoords[3];
			mesh_m.getTriangle( triangleIndex ).getTextureCoords( textureCoords );

			const Material& material = getScene()->getMaterial( materialId_m );
			material.getEmittedPosition( textureCoords, posRands, compositionStorage, intersection.coords_m );
		}

		/// get position and normal
		{
			TriangleExplicit triangle;
			mesh_m.getTriangle( triangleIndex, triangle );

			triangle.getPosition( intersection.coords_m, intersection.position_m );
			triangle.getNormal( intersection.normal_m );
		}

		/// use intersection to lookup qualities (including emitted and emittedRoughness)
		getQualities( intersection, qualities );
	}

	return isAnyEmitted;
}


bool ObjectMesh::getIntersection
(
	const ObjectRenderableId& objectId,
	const Ray&                ray,
	float&                    distance,
	Intersection&             intersection
) const
{
	bool anyHits = false;

	/// check bound intersection first
	if( getBound().getIntersection( ray ) )
	{
		float        nearestDistance = FLOAT_MAX;
		Intersection nearestIntersection;

		/// non-indexed search, by comparing all primitive intersection distances to find minimum
		for( int i = mesh_m.getTriangles().size();  i-- > 0; )
		{
			/// one triangle cant bounce a ray back to itself
			//if( !( (i == ray.getObject().getPrimitiveId()) && ray.getObject().isMainObjectsEqual(objectId)) )
			{
				TriangleExplicit triangle;
				mesh_m.getTriangle( i, triangle );
				const ObjectRenderableId objectTriId( objectId, i & 0xFFFF );

				/// intersect sub part
				float        distanceTrial = 0.0f;
				UnitVector3f normal;
				Vector2f     barycentrics;
				bool         isHit = triangle.getIntersection( ray, distanceTrial, normal, barycentrics );

				/// check for false intersection
				if( isHit )
				{
					isHit &= !isFalseIntersection( ray, objectTriId, mesh_m.getTriangle( i ), triangle, normal );
				}

				/// keep nearest hit data
				anyHits |= isHit;
				if( isHit & (distanceTrial < nearestDistance) )
				{
					nearestDistance = distanceTrial;

					ray.getPosition( nearestDistance, nearestIntersection.position_m );
					nearestIntersection.normal_m = normal;
					nearestIntersection.coords_m = barycentrics;
					nearestIntersection.object_m = objectTriId;
				}
			}
		}

		/// set intersection output params
		if( anyHits )
		{
			distance     = nearestDistance;
			intersection = nearestIntersection;
		}
	}

	return anyHits;
}


void ObjectMesh::getObjectsForProjection
(
	const Transclip&              transform,
	const ClipFlags               enclosingClipFlags,
	vector<ProjectableObjectRef>& objects
) const
{
	/// if wholly or partially inside
	ClipFlags thisClipflags;
	if( getBound().getClipping( enclosingClipFlags, transform, thisClipflags ) )
	{
		/// add object
		objects.push_back( ProjectableObjectRef( this, transform, thisClipflags ) );
	}
}


void ObjectMesh::getMesh
(
	Mesh&        ,//meshBuffer,
	const Mesh** ppMesh
) const
{
	*ppMesh = &mesh_m;
}


void ObjectMesh::getQualities
(
	const Intersection& intersection,
	Qualities&          qualities
) const
{
	const dword     triangleIndex = intersection.object_m.getPrimitiveId();
	const Material& material      = getScene()->getMaterial( materialId_m );
	Vector2<word> textureCoords[3];

	/// get qualities
	if( getScene() != 0 )
	{
		mesh_m.getTriangle( triangleIndex ).getTextureCoords( textureCoords );

		material.getQualities( intersection.coords_m, textureCoords, qualities );
	}
	else
	{
		qualities.setToDefault();
	}

	/// set qualities normal and tangent
	{
		TriangleExplicit triangle;
		mesh_m.getTriangle( triangleIndex, triangle );

		/// get basic normal
		UnitVector3f normal;
		if( mesh_m.getTriangle( triangleIndex ).getInterpolateNormal() )
		{
			/// get interpolated normal
			triangle.getNormal( intersection.coords_m, normal );
		}
		else
		{
			/// use facet normal
			normal = intersection.normal_m;
		}

		/// make triangle-texture coord frame by aligning: basic normal, first edge, and cross prod; with texture x axis
		Vector3f frameX;
		Vector3f frameZ;
		Vector3f frameY( normal.getVector() );
		{
			Vector3f edge01;
			{
				Vector3f vert0;
				Vector3f vert1;
				triangle.getVertexPosition0( vert0 );
				triangle.getVertexPosition1( vert1 );
				edge01 = (vert1 -= vert0);
				edge01.unitizeEq();
			}

			Vector3f edgeCross( normal.getVector() );
			edgeCross.crossEq( edge01 );

			/// rotate frame in triangle plane to align with texture x axis
			const Vector2<word> edgeTx01w( textureCoords[1] - textureCoords[0] );
			Vector2f edgeTx01( edgeTx01w.getX(), edgeTx01w.getY() );
			edgeTx01.unitizeEq();
			((frameX = edge01) *= edgeTx01.getX()) += (edgeCross * -edgeTx01.getY());
			((frameZ = edge01) *= edgeTx01.getY()) += (edgeCross *  edgeTx01.getX());
		}

		/// adjust basic normal with qualities (bump-mapped) normal
		if( !qualities.normal_m.isZero() )
		{
			/// set resultant normal to bump normal relative to coord frame
			const UnitVector3f& bump = qualities.normal_m;
			qualities.normal_m = (((frameX * bump.getX())  +=
			                       (frameY * bump.getY())) +=
			                       (frameZ * bump.getZ()));
		}
		else
		{
			qualities.normal_m = normal;
		}

		/// set tangent
		if( !qualities.tangent_m.isZero() )
		{
			/// set resultant tangent to mapped tangent relative to coord frame
			/// (using bumped frame instead would mean interpreting bump mapping as a moulding,
			/// rather than carving, of the surface. either seems sensible...)
			const UnitVector3f& tangent = qualities.tangent_m;
			qualities.tangent_m = (((frameX * tangent.getX())  +=
			                        (frameY)) +=
			                        (frameZ * tangent.getZ()));
		}
		else
		{
			/// set resultant tangent to base tangent projected to coord frame
			qualities.tangent_m = material.getQualitiesBase().tangent_m;
			qualities.tangent_m.orthogonalizeEq( UnitVector3f(frameY) );
		}
	}
}


Float01 ObjectMesh::getTransparency
(
	const ObjectRenderableId& ,//objectId,
	const Vector3f&           position
) const
{
	Float01 transparency;
	if( getScene() != 0 )
	{
		const Material& material = getScene()->getMaterial( materialId_m );
		transparency = material.getTransparency( position );
	}

	return transparency;
}


float ObjectMesh::getRefractiveIndex
(
	const ObjectRenderableId& ,//objectId,
	const Vector3f&           position
) const
{
	float refractiveIndex = 1.0f;
	if( getScene() != 0 )
	{
		const Material& material = getScene()->getMaterial( materialId_m );
		refractiveIndex = material.getRefractiveIndex( position );
	}

	return refractiveIndex;
}


const CieXyz& ObjectMesh::getEmittance
(
	const ObjectRenderableId&
) const
{
	return getEmittanceSum();
}


const char* ObjectMesh::getClassName() const
{
	return ObjectMesh::CLASS_NAME();
}




/// implementation ---------------------------------------------------------------------------------
void ObjectMesh::setMesh
(
	const vector<Vector3f>&         vertexPositions,
	const vector<TriangleImplicit>& triangles,
	const dword                     materialId
)
{
	mesh_m.setMesh( vertexPositions, triangles );

	materialId_m = materialId;
	setMesh();
}


void ObjectMesh::setMesh()
{
	{
		Vector3f min;
		Vector3f max;
		mesh_m.getMinMax( min, max );
		makeBound( min, max, bound() );
	}

	setEmittanceSum( makeEmittance() );
	setSpecularness( makeSpecularness() );
}


void ObjectMesh::makeBound
(
	const Vector3f& min,
	const Vector3f& max,
	Bound&          bound
) const
{
	/// axis aligned box
	bound.setMinMax( min, max );
}


CieXyz ObjectMesh::makeEmittance() const
{
	/// emittance is watts (== mean(emit texels) * area, or emit * area)
	CieXyz emittance;

	if( getScene() != 0 )
	{
		/// get vertex positions
		vector<Vector3f> vertexPositions;
		mesh_m.getVertexPositions( vertexPositions );

		const Material& material = getScene()->getMaterial( materialId_m );

		/// step through triangles summing emittance
		const vector<TriangleImplicit>& triangles = mesh_m.getTriangles();
		for( int i = triangles.size();  i-- > 0; )
		{
			const TriangleImplicit& triangle = triangles[i];

			/// get mean radiant exitance
			CieXyz emittedMean;
			{
				Vector2<word> textureCoords[3];
				triangle.getTextureCoords( textureCoords );

				material.getEmittedMean( textureCoords, emittedMean );
			}

			/// get area
			const float area = triangle.getArea( vertexPositions );

			/// add triangle emittance to total
			emittance += (emittedMean *= area);
		}
	}

	return emittance;
}


Float01 ObjectMesh::makeSpecularness() const
{
	float specularness = 0.0f;

	if( getScene() != 0 )
	{
		const Material& material = getScene()->getMaterial( materialId_m );

		/// step thru triangles accumulating specularness
		const vector<TriangleImplicit>& triangles = mesh_m.getTriangles();
		const int triangleCount = triangles.size();
		for( int i = triangleCount;  i-- > 0; )
		{
			/// get mean specularness
			Float01 specularnessMean;
			{
				Vector2<word> textureCoords[3];
				triangles[i].getTextureCoords( textureCoords );

				material.getSpecularnessMean( textureCoords, specularnessMean );
			}

			/// accumulate triangle specularness to total
			specularness += (specularnessMean / float(triangleCount));
		}
	}

	return specularness;
}


void ObjectMesh::zeroObject()
{
	mesh_m       = Mesh();
	materialId_m = 0;
}


dword ObjectMesh::getMaterialId() const
{
	return materialId_m;
}


string ObjectMesh::getMaterialName() const
{
	return getScene()->getMaterialName( materialId_m );
}


dword ObjectMesh::getMaterialIndex( const string& name ) const
{
	return getScene()->getMaterialIndex( name );
}


const vector<TriangleImplicit>& ObjectMesh::triangles() const
{
	return mesh_m.getTriangles();
}


bool ObjectMesh::isFalseIntersection
(
	const Ray&                ray,
	const ObjectRenderableId& objectId,
	const TriangleImplicit&   triImp,
	const TriangleExplicit&   triExp,
	const UnitVector3f&       normal
) const
{
	bool isFalseIntersection = false;

	const bool isMainObjectsEqual = objectId.isMainObjectsEqual( ray.getObject() );
	if( isMainObjectsEqual )
	{
		/// check for inaccuracy along ray
		{
			/// false intersection if:
			/// 	query and previous main objects are the same, and
			/// 	ray left from the outside and hit the inside (or vice versa)
			const float nDotd           = normal.dot( ray.getDirection() );
			const bool  isDifferentSide = ray.isInside() == (nDotd < 0.0f);

			isFalseIntersection = isDifferentSide;
		}

		/// check for inaccuracy causing overlapping edges
		/// (but overlapping corners are not completely cured...)
		if( !isFalseIntersection )
		{
			/// determine if triangles are connected
			dword sharedVertexsCount;
			bool  isSharedVertex[3];
			{
				const dword previousTriId              = ray.getObject().getPrimitiveId();
				const TriangleImplicit& previousTriImp = mesh_m.getTriangle( previousTriId );
				sharedVertexsCount = triImp.getCommonVertexs( previousTriImp, isSharedVertex[0], isSharedVertex[1], isSharedVertex[2] );
			}

			/// false intersection if:
			///		query triangle is directly connected to previous triangle, and
			///		query triangle is entirely in different halfspace (as separated by previous triangle) to ray
			/// (if the query triangle is connected by a single vertex, and is angled so its in both halfspaces,
			/// it can cause false intersections that are *not* found and cured by this check...)
			if( (sharedVertexsCount > 0) & (sharedVertexsCount < 3))
			{
				if( sharedVertexsCount == 2 )
				{
					const dword nonshared = dword(!isSharedVertex[1]) | (dword(!isSharedVertex[2]) << 1);
					const dword shared    = (nonshared + 1) % 3;

					const Vector3f tangent( triExp.getEdge( shared, nonshared ) );

					isFalseIntersection = (tangent.dot( ray.getNormal().getVector() ) < 0.0f) != ray.isInside();
				}
				else if( sharedVertexsCount == 1 )
				{
					const dword shared     = dword(isSharedVertex[1]) | (dword(isSharedVertex[2]) << 1);
					const dword nonshared0 = (shared + 1) % 3;
					const dword nonshared1 = (nonshared0 + 1) % 3;

					const Vector3f tangent0( triExp.getEdge( shared, nonshared0 ) );
					const Vector3f tangent1( triExp.getEdge( shared, nonshared1 ) );

					isFalseIntersection = ((tangent0.dot( ray.getNormal().getVector() ) < 0.0f) != ray.isInside()) &&
					                      ((tangent1.dot( ray.getNormal().getVector() ) < 0.0f) != ray.isInside());
				}
			}
		}
	}

	return isFalseIntersection;
}


void ObjectMesh::assign
(
	const ObjectRenderable& other
)
{
	if( &other != this )
	{
		ObjectRenderable::assign( other );

		mesh_m       = Mesh();
		materialId_m = 0;
	}
}


void ObjectMesh::assign
(
	const ObjectMesh& other
)
{
	if( &other != this )
	{
		ObjectRenderable::assign( other );

		mesh_m       = other.mesh_m;
		materialId_m = other.materialId_m;
	}
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const ObjectMesh& obj )
{
	/// write mesh
	out << obj.mesh_m;

	/// write material id
	out.write( "materialRef", obj.getMaterialName() );

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, ObjectMesh& obj )
{
	try
	{
		/// read mesh
		in >> obj.mesh_m;

		/// read material id
		{
			string materialName;
			in.read( "materialRef", materialName );
			obj.materialId_m = obj.getMaterialIndex( materialName );
		}


		/// set fields
		obj.setMesh();


		return in;
	}
	catch( ... )
	{
		obj.zeroObject();

		throw;
	}
}


const char* ObjectMesh::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::ObjectMesh";
	return CLASS_NAME;
}


const char* ObjectMesh::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a renderable object, as a triangle mesh with shared vertex pool";
	return CLASS_DESCRIPTION;
}
