/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Scene.h"
#include "hxa7241math.h"
#include "Exception.h"
#include "Serialization.h"

#include "ObjectRenderable.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
ObjectRenderable::ObjectRenderable
(
	const Scene* pScene
)
 :	bound_m         (),
	emittance_m     (),
	specularness_m  (),
	hierarchyLevel_m(),
	subPartCount_m  ( 1 ),
	pScene_m        ( pScene )
{
	if( pScene == 0 )
	{
		throw hxa7241::utility::Exception( "null scene pointer given", CLASS_NAME(), "constructor" );
	}
}


ObjectRenderable::~ObjectRenderable()
{
}


ObjectRenderable::ObjectRenderable( const ObjectRenderable& other )
{
	ObjectRenderable::assign( other );
}


ObjectRenderable& ObjectRenderable::operator=( const ObjectRenderable& other )
{
	assign( other );

	return *this;
}




/// queries ----------------------------------------------------------------------------------------
const CieXyz& ObjectRenderable::getEmittanceSum() const
{
	return emittance_m;
}


Float01 ObjectRenderable::getSpecularness() const
{
	return specularness_m;
}


dword ObjectRenderable::getHierarchyLevel() const
{
	return hierarchyLevel_m;
}


dword ObjectRenderable::getSubPartCount() const
{
	return subPartCount_m;
}


const Bound& ObjectRenderable::getBound() const
{
	return bound_m;
}


float ObjectRenderable::getExtinctionCoefficient
(
	const ObjectRenderableId& objectId,
	const Vector3f&           position
) const
{
	const Float01 transparency( getTransparency( objectId, position ) );
	return (transparency != 0.0f) ? -float( ::log( transparency ) ) : FLOAT_MAX;
}


const char* ObjectRenderable::getClassName() const
{
	return ObjectRenderable::CLASS_NAME();
}




/// implementation ---------------------------------------------------------------------------------
void ObjectRenderable::setEmittanceSum
(
	const CieXyz& emittance
)
{
	if( emittance.average() >= pScene_m->getLightsourceThreshold() )
	{
		emittance_m = emittance;
	}
	else
	{
		emittance_m = CieXyz();
	}
}


void ObjectRenderable::setSpecularness
(
	const Float01 specularness
)
{
	specularness_m = specularness;
}


void ObjectRenderable::setHierarchyLevel
(
	const dword level
)
{
	hierarchyLevel_m = level;
}


void ObjectRenderable::setSubPartCount
(
	const dword count
)
{
	subPartCount_m = count;
}


Bound& ObjectRenderable::bound()
{
	return bound_m;
}


const Scene* ObjectRenderable::getScene() const
{
	return pScene_m;
}


void ObjectRenderable::assign
(
	const ObjectRenderable& other
)
{
	if( &other != this )
	{
		bound_m          = other.bound_m;

		emittance_m      = other.emittance_m;
		specularness_m   = other.specularness_m;
		hierarchyLevel_m = other.hierarchyLevel_m;
		subPartCount_m   = other.subPartCount_m;

		pScene_m         = other.pScene_m;
	}
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const ObjectRenderable& )
{
	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, ObjectRenderable& )
{
	return in;
}


const char* ObjectRenderable::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::ObjectRenderable";
	return CLASS_NAME;
}


const char* ObjectRenderable::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a renderable object";
	return CLASS_DESCRIPTION;
}
