/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include <iostream>

#include "ObjectRenderableId.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
ObjectRenderableId::ObjectRenderableId()
 :	pObject_m( 0 ),
	stackEnd_m( 0 ),
	primitive_m( 0 )
{
	ObjectRenderableId::zeroStack( stack_m );
}


ObjectRenderableId::ObjectRenderableId
(
	const ObjectRenderable* pObject,
	const dword             primitive
)
 :	pObject_m( pObject ),
	stackEnd_m( 0 ),
	primitive_m( primitive )
{
	ObjectRenderableId::zeroStack( stack_m );
}


ObjectRenderableId::ObjectRenderableId
(
	const ObjectRenderableId& other,
	const dword               primitive
)
 :	pObject_m  ( other.pObject_m ),
	stackEnd_m ( other.stackEnd_m ),
	primitive_m( primitive )
{
	ObjectRenderableId::copyStack( other.stack_m, stack_m );
}


ObjectRenderableId::~ObjectRenderableId()
{
}


ObjectRenderableId::ObjectRenderableId( const ObjectRenderableId& other )
 :	pObject_m  ( other.pObject_m ),
	stackEnd_m ( other.stackEnd_m ),
	primitive_m( other.primitive_m )
{
	ObjectRenderableId::copyStack( other.stack_m, stack_m );
}


ObjectRenderableId& ObjectRenderableId::operator=( const ObjectRenderableId& other )
{
	pObject_m   = other.pObject_m;
	copyStack( other.stack_m, stack_m );
	stackEnd_m  = other.stackEnd_m;
	primitive_m = other.primitive_m;

	return *this;
}




/// streaming --------------------------------------------------------------------------------------
std::ostream& perceptuum2::modelling::operator<<( std::ostream& out, const ObjectRenderableId& obj )
{
	out << "(" << obj.pObject_m << ", (";

	for( int i = 0, end = obj.STACK_SIZE_MAX;  i < end;  ++i )
	{
		out << dword(ubyte(obj.stack_m[i])) << " ";
	}

	out << "| " << obj.stackEnd_m << "), " << obj.primitive_m << ")";

	return out;
}




/// commands ---------------------------------------------------------------------------------------
bool ObjectRenderableId::push
(
	const hxa7241::primitives::byte subPartId
)
{
	const bool wouldOverflow = stackEnd_m >= STACK_SIZE_MAX;
	if( !wouldOverflow )
	{
		stack_m[stackEnd_m] = subPartId;
		++stackEnd_m;
	}

	return wouldOverflow;
}


bool ObjectRenderableId::push
(
	const ObjectRenderable* const   pObject,
	const hxa7241::primitives::byte subPartId
)
{
	const bool wouldOverflow = stackEnd_m >= STACK_SIZE_MAX;
	if( !wouldOverflow )
	{
		pObject_m = pObject;

		stack_m[stackEnd_m] = subPartId;
		++stackEnd_m;
	}

	return wouldOverflow;
}


bool ObjectRenderableId::pop
(
	const ObjectRenderable* const pObject
)
{
	const bool wouldUnderflow = stackEnd_m <= 0;
	if( !wouldUnderflow )
	{
		pObject_m = pObject;

		--stackEnd_m;
		stack_m[stackEnd_m] = 0;
	}

	return wouldUnderflow;
}


bool ObjectRenderableId::shift
(
	const ObjectRenderable* const pObject
)
{
	const bool wouldUnderflow = stackEnd_m <= 0;
	if( !wouldUnderflow )
	{
		pObject_m = pObject;

		const byte* sourceCursor      = stack_m + 1;
			  byte* destinationCursor = stack_m;
			  byte* destinationEnd    = stack_m + stackEnd_m - 1;
		for( ;  destinationCursor < destinationEnd; )
		{
			*(destinationCursor++) = *(sourceCursor++);
		}

		*destinationCursor = 0;
		--stackEnd_m;
	}

	return wouldUnderflow;
}




/// queries ----------------------------------------------------------------------------------------
hxa7241::primitives::byte ObjectRenderableId::getTop() const
{
	byte top = 0;

	if( stackEnd_m > 0 )
	{
		top = stack_m[stackEnd_m - 1];
	}

	return top;
}


hxa7241::primitives::byte ObjectRenderableId::getBottom() const
{
	return stack_m[0];
}


bool ObjectRenderableId::operator==( const ObjectRenderableId& other ) const
{
	return ((pObject_m   == other.pObject_m) & (primitive_m == other.primitive_m)) &&
	       isEqualStack( stack_m, other.stack_m, stackEnd_m );
}


bool ObjectRenderableId::operator!=( const ObjectRenderableId& other ) const
{
	return !operator==( other );
}


bool ObjectRenderableId::isMainObjectsEqual
(
	const ObjectRenderableId& other
) const
{
	return (pObject_m == other.pObject_m) && isEqualStack( stack_m, other.stack_m, stackEnd_m );
}


dword ObjectRenderableId::getMaxStackSize()
{
	return STACK_SIZE_MAX;
}


dword ObjectRenderableId::getMaxStackWidth()
{
	return STACK_WIDTH_MAX;
}




/// implementation ---------------------------------------------------------------------------------
void ObjectRenderableId::zeroStack
(
	hxa7241::primitives::byte array[]
)
{
	//if( STACK_SIZE_MAX != 8 )
	#ifndef STACK_2_DWORDS
	{
		byte* cursor = array + STACK_SIZE_MAX;

		for( ;  cursor > array; )
		{
			*(--cursor) = 0;
		}
	}
	//else
	#else
	{
		dword* quads = reinterpret_cast<dword*>(array);
		*quads       = 0;
		*(quads + 1) = 0;
	}
	#endif
}


void ObjectRenderableId::copyStack
(
	const hxa7241::primitives::byte source[],
	hxa7241::primitives::byte       destination[]
)
{
	//if( STACK_SIZE_MAX != 8 )
	#ifndef STACK_2_DWORDS
	{
		const byte* sourceCursor      = source      + STACK_SIZE_MAX;
			  byte* destinationCursor = destination + STACK_SIZE_MAX;

		for( ;  destinationCursor > destination; )
		{
			*(--destinationCursor) = *(--sourceCursor);
		}
	}
	//else
	#else
	{
		const dword* quadsSource      = reinterpret_cast<const dword*>(source);
		dword*       quadsDestination = reinterpret_cast<dword*>(destination);
		*quadsDestination       = *quadsSource;
		*(quadsDestination + 1) = *(quadsSource + 1);
	}
	#endif
}


bool ObjectRenderableId::isEqualStack
(
	const hxa7241::primitives::byte first[],
	const hxa7241::primitives::byte second[],
	#ifndef STACK_2_DWORDS
	const dword                     size
	#else
	const dword
	#endif
)
{
	bool isEqual;

	//if( STACK_SIZE_MAX != 8 )
	#ifndef STACK_2_DWORDS
	{
		isEqual = true;

		const byte* firstCursor  = first  + size;
		const byte* secondCursor = second + size;
		for( ;  (firstCursor > first) & isEqual; )
		{
			isEqual = *(--firstCursor) == *(--secondCursor);
		}
	}
	//else
	#else
	{
		const dword* quadsFirst  = reinterpret_cast<const dword*>(first);
		const dword* quadsSecond = reinterpret_cast<const dword*>(second);

		isEqual = (*quadsSecond == *quadsFirst) & (*(quadsSecond + 1) == *(quadsFirst + 1));
	}
	#endif

	return isEqual;
}
