/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Float01.h"
#include "Ray.h"

#include "TriangleExplicit.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
TriangleExplicit::TriangleExplicit()
// :	vertex0_m (),
//	vertex1_m (),
//	vertex2_m (),
//	position_m(),
//	scaling_m (),
//	interpolateNormal_m()
{
}


TriangleExplicit::TriangleExplicit
(
	const Vertex&   vert0,
	const Vertex&   vert1,
	const Vertex&   vert2,
	const Vector3f& position,
	const Vector3f& scaling,
	const bool      interpolateNormal
)
 :	vertex0_m ( vert0 ),
	vertex1_m ( vert1 ),
	vertex2_m ( vert2 ),
	position_m( position ),
	scaling_m ( scaling ),
	interpolateNormal_m( interpolateNormal )
{
}


TriangleExplicit::~TriangleExplicit()
{
}


TriangleExplicit::TriangleExplicit( const TriangleExplicit& other )
 :	vertex0_m ( other.vertex0_m ),
	vertex1_m ( other.vertex1_m ),
	vertex2_m ( other.vertex2_m ),
	position_m( other.position_m ),
	scaling_m ( other.scaling_m ),
	interpolateNormal_m( other.interpolateNormal_m )
{
}


TriangleExplicit& TriangleExplicit::operator=( const TriangleExplicit& other )
{
	if( &other != this )
	{
		vertex0_m  = other.vertex0_m;
		vertex1_m  = other.vertex1_m;
		vertex2_m  = other.vertex2_m;

		position_m = other.position_m;
		scaling_m  = other.scaling_m;

		interpolateNormal_m = other.interpolateNormal_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void TriangleExplicit::set
(
	const Vertex&   vert0,
	const Vertex&   vert1,
	const Vertex&   vert2,
	const Vector3f& position,
	const Vector3f& scaling,
	const bool      interpolateNormal
)
{
	vertex0_m  = vert0;
	vertex1_m  = vert1;
	vertex2_m  = vert2;

	position_m = position;
	scaling_m  = scaling;

	interpolateNormal_m = interpolateNormal;
}




/// queries ----------------------------------------------------------------------------------------
void TriangleExplicit::getVertexPosition0( Vector3f& pos ) const
{
	vertex0_m.getPosition( position_m, scaling_m, pos );
}


void TriangleExplicit::getVertexPosition1( Vector3f& pos ) const
{
	vertex1_m.getPosition( position_m, scaling_m, pos );
}


void TriangleExplicit::getVertexPosition2( Vector3f& pos ) const
{
	vertex2_m.getPosition( position_m, scaling_m, pos );
}


void TriangleExplicit::getVertexNormal0( UnitVector3f& norm ) const
{
	vertex0_m.getNormal( norm );
}


void TriangleExplicit::getVertexNormal1( UnitVector3f& norm ) const
{
	vertex1_m.getNormal( norm );
}


void TriangleExplicit::getVertexNormal2( UnitVector3f& norm ) const
{
	vertex2_m.getNormal( norm );
}


Vector3f TriangleExplicit::getEdge
(
	const dword vertex0,
	const dword vertex1
) const
{
	const Vertex*const vertexs[] = { &vertex0_m, &vertex1_m, &vertex2_m };

	Vector3f pos0;
	Vector3f pos1;
	vertexs[vertex0 % 3]->getPosition( position_m, scaling_m, pos0 );
	vertexs[vertex1 % 3]->getPosition( position_m, scaling_m, pos1 );

	return Vector3f( pos1 - pos0 );
}


int
intersect_triangle( const float orig[3], const float dir[3],
                    const float vert0[3], const float vert1[3], const float vert2[3],
                    const int twoSided,
                    float* t, float* u, float* v );

bool TriangleExplicit::getIntersection
(
	const Ray&    ray,
	float&        distance,
	UnitVector3f& normal,
	Vector2f&     barycentrics
) const
{
	bool isHit = false;
	{
		/// cant intersect if ray direction zero
		if( !ray.getDirection().isZero() )
		{
			/// repackage ray
			const Vector3f&     origin    = ray.getPosition();
			const UnitVector3f& direction = ray.getDirection();
			const float ori[] = { origin.getX(),    origin.getY(),    origin.getZ() };
			const float dir[] = { direction.getX(), direction.getY(), direction.getZ() };

			/// repackage vertex positions
			Vector3f vert0;
			Vector3f vert1;
			Vector3f vert2;
			getVertexPosition0( vert0 );
			getVertexPosition1( vert1 );
			getVertexPosition2( vert2 );
			const float v0[] = { vert0.getX(), vert0.getY(), vert0.getZ() };
			const float v1[] = { vert1.getX(), vert1.getY(), vert1.getZ() };
			const float v2[] = { vert2.getX(), vert2.getY(), vert2.getZ() };

			/// do intersection
			float d;
			float u, v;
			isHit = bool( intersect_triangle( ori, dir, v0, v1, v2, int(true), &d, &u, &v ) );
			if( isHit )
			{
				distance = d;
				barycentrics.setXY( u, v );

				/// calc non-interpolated facet normal
				getNormal( normal );
			}
		}
	}

	return isHit;
}


bool TriangleExplicit::getBarycentrics
(
	const Vector3f& position,
	Vector2f&       barycentrics
) const
{
	/// get vertex positions
	Vector3f vert0;
	Vector3f vert1;
	Vector3f vert2;
	getVertexPosition0( vert0 );
	getVertexPosition1( vert1 );
	getVertexPosition2( vert2 );

	/// make triangle edges
	const Vector3f edge01( vert1 - vert0 );
	const Vector3f edge12( vert2 - vert1);
	const Vector3f edge20( vert0 - vert2 );

	/// make centroidal edges of verts to position
	const Vector3f edge0Pos( position - vert0 );
	const Vector3f edge1Pos( position - vert1 );
	const Vector3f edge2Pos( position - vert2 );

	/// calculate areas of centroidal subtriangles (scaled by 2 since cross calcs parallelogram area)
	const Vector3f cross010P( edge01.cross( edge0Pos ) );
	const Vector3f cross121P( edge12.cross( edge1Pos ) );
	const Vector3f cross202P( edge20.cross( edge2Pos ) );

	/// make normal of triangle
	const Vector3f norm( edge01.cross( edge12 ) );

	/// check if point outside triangle edges, and if triangle is degenerate
	const bool isOutsideEdge0   = cross010P.dot( norm ) < 0.0f;
	const bool isOutsideEdge1   = cross121P.dot( norm ) < 0.0f;
	const bool isOutsideEdge2   = cross202P.dot( norm ) < 0.0f;
	const bool isInsideTriangle = !(isOutsideEdge0 | isOutsideEdge1 | isOutsideEdge2);
	bool isZeroAreaTriangle = norm.length() == 0.0f;

	/// calculate barycentric coords of position inside non-zero triangle
	if( isInsideTriangle & !isZeroAreaTriangle )
	{
		/// these crosses in this order match the output of intersect_triangle()
		const float b0 = cross202P.length();
		const float b1 = cross010P.length();
		const float b2 = cross121P.length();

		const float sum = b0 + b1 + b2;
		isZeroAreaTriangle = sum == 0.0f;
		if( !isZeroAreaTriangle )
		{
			/// normalize to total of 1, effectively projecting position onto plane of triangle
			const float div = 1.0f / sum;
			barycentrics.setXY( b0 * div, b1 * div );
		}
	}

	/// calculate barycentric coords of position clamped to an edge
	if( !isInsideTriangle | isZeroAreaTriangle )
	{
		dword clampEdge;
		if( !isInsideTriangle )
		{
			/// pick an edge the position is on the outside of
			clampEdge = isOutsideEdge0 ? 0 : (isOutsideEdge1 ? 1 : 2);
		}
		else
		{
			/// pick non shortest edge, for good results with line-like zero-area triangles
			const float edge01Len = edge01.length();
			const float edge12Len = edge12.length();
			clampEdge = edge01Len > edge12Len ? 0 : 1;
		}

		float b[3];
		{
			const Vector3f* const edge[]    = { &edge01,   &edge12,   &edge20   };
			const Vector3f* const edgePos[] = { &edge0Pos, &edge1Pos, &edge2Pos };

			/// calc one barycentric as proportional distance along clamp edge
			const float edgeLen = edge[clampEdge]->length();
			if( edgeLen > 0.0f )
			{
				b[1] = edgePos[clampEdge]->dot( *(edge[clampEdge]) ) / (edgeLen * edgeLen);
			}
			else
			{
				b[1] = 0.0f;
			}
			/// constrain between vertexs
			if( b[1] < 0.0f )
			{
				b[1] = 0.0f;
			}
			else if( b[1] > 1.0f )
			{
				b[1] = 1.0f;
			}
			/// second barycentric is complement of first, last barycentric is zero
			b[0] = 1.0f - b[1];
			b[2] = 0.0f;
		}

		/// order the barycentrics dependent on chosen clamp edge
		barycentrics.setXY( b[(1 - clampEdge) % 3], b[(2 - clampEdge) % 3] );//, b[(3 - clampEdge) % 3] );
	}

	return isInsideTriangle;
}


void TriangleExplicit::getPosition
(
	const Vector2f& barycentrics,
	Vector3f&       position
) const
{
	Vector3f vert0;
	Vector3f vert1;
	Vector3f vert2;
	getVertexPosition0( vert0 );
	getVertexPosition1( vert1 );
	getVertexPosition2( vert2 );

	/// calculate interpolation by using two edges as axes scaled by the barycentrics
	/// weighting association matches the output of intersect_triangle()
	Vector3f edge0( vert1 - vert0 );
	Vector3f edge1( vert2 - vert0 );
	position = (((edge0 *= barycentrics.getX()) += (edge1 *= barycentrics.getY())) += vert0);

//	/// not accurate, or is it just wrong...
//	/// weighting association matches the output of intersect_triangle()
//	position = ((vert0 * (1.0f - (barycentrics.getX() + barycentrics.getY()))) +=
//	            (vert1 * barycentrics.getY()) +=
//	            (vert2 * barycentrics.getX()));
}


void TriangleExplicit::getNormal
(
	const Vector2f& barycentrics,
	UnitVector3f&   normal
) const
{
	if( interpolateNormal_m )
	{
		UnitVector3f norm0;
		UnitVector3f norm1;
		UnitVector3f norm2;
		getVertexNormal0( norm0 );
		getVertexNormal1( norm1 );
		getVertexNormal2( norm2 );

		/// weighting association matches the output of intersect_triangle()
		Vector3f edge0( norm1.getVector() - norm0.getVector() );
		Vector3f edge1( norm2.getVector() - norm0.getVector() );
		normal = (((edge0 *= barycentrics.getX()) += (edge1 *= barycentrics.getY())) += norm0.getVector());

		///// weighting association matches the output of intersect_triangle()
		//normal = ((norm0.getVector() * (1.0f - (barycentrics.getX() + barycentrics.getY()))) +=
		//          (norm1.getVector() * barycentrics.getY()) +=
		//	        (norm2.getVector() * barycentrics.getX()));
	}
	else
	{
		getNormal( normal );
	}
}


void TriangleExplicit::getNormal
(
	UnitVector3f& normal
) const
{
	Vector3f vert0;
	Vector3f vert1;
	Vector3f vert2;
	getVertexPosition0( vert0 );
	getVertexPosition1( vert1 );
	getVertexPosition2( vert2 );

	const Vector3f edge01( vert1 - vert0 );
	const Vector3f edge12( vert2 - vert1);

	normal = edge01.cross( edge12 );
}




/// implementation ---------------------------------------------------------------------------------
/**
 * 'fast, minimum storage ray-triangle intersection'
 * - moller, trumbore
 * (journal of graphics tools, v2n1p21 1997)
 *
 * We present a clean algorithm for determining whether a ray intersects a
 * triangle. The algorithm translates the origin of the ray and then changes the
 * base to yield a vector <i>(t&nbsp;u&nbsp;v)</i><tt><sup>T</sup></tt>, where
 * <i>t</i> is the distance to the plane in which the triangle lies and <i>(u,v)
 * </i> represents the coordinates inside the triangle.
 * <br>One advantage of this method is that the plane equation need not be computed
 * on the fly nor be stored, which can amount to significant memory savings for
 * triangle meshes. As we found our method to be comparable in speed to previous
 * methods, we believe it is the fastest ray-triangle intersection routine for
 * triangles that do not have precomputed plane equations.
 *
 * hxa7241 modifications:
 * - changed doubles to floats
 * - added 'twoSided' bool param, and replaced TEST_CULL sections with
 *   runtime branches
 * - changed final return to give true only for intersections along the
 *   forward (positive t) direction of the ray
 */

#define EPSILON 0.000001f
#define CROSS(dest,v1,v2) \
          dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
          dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
          dest[2]=v1[0]*v2[1]-v1[1]*v2[0];
#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])
#define SUB(dest,v1,v2) \
          dest[0]=v1[0]-v2[0]; \
          dest[1]=v1[1]-v2[1]; \
          dest[2]=v1[2]-v2[2];

int
intersect_triangle( const float orig[3], const float dir[3],
                    const float vert0[3], const float vert1[3], const float vert2[3],
                    const int twoSided,
                    float* t, float* u, float* v )
{
	float edge1[3], edge2[3], tvec[3], pvec[3], qvec[3];
	float det,inv_det;

	/* find vectors for two edges sharing vert0 */
	SUB(edge1, vert1, vert0);
	SUB(edge2, vert2, vert0);

	/* begin calculating determinant - also used to calculate U parameter */
	CROSS(pvec, dir, edge2);

	/* if determinant is near zero, ray lies in plane of triangle */
	det = DOT(edge1, pvec);

//#ifdef TEST_CULL           /* define TEST_CULL if culling is desired */
	if( !twoSided )
	{
		if (det < EPSILON)
			return 0;

		/* calculate distance from vert0 to ray origin */
		SUB(tvec, orig, vert0);

		/* calculate U parameter and test bounds */
		*u = DOT(tvec, pvec);
		if (*u < 0.0f || *u > det)
			return 0;

		/* prepare to test V parameter */
		CROSS(qvec, tvec, edge1);

		/* calculate V parameter and test bounds */
		*v = DOT(dir, qvec);
		if (*v < 0.0f || *u + *v > det)
			return 0;

		/* calculate t, scale parameters, ray intersects triangle */
		*t = DOT(edge2, qvec);
		inv_det = 1.0f / det;
		*t *= inv_det;
		*u *= inv_det;
		*v *= inv_det;
	}
//#else                    /* the non-culling branch */
	else
	{
		if (det > -EPSILON && det < EPSILON)
			return 0;
		inv_det = 1.0f / det;

		/* calculate distance from vert0 to ray origin */
		SUB(tvec, orig, vert0);

		/* calculate U parameter and test bounds */
		*u = DOT(tvec, pvec) * inv_det;
		if (*u < 0.0f || *u > 1.0f)
			return 0;

		/* prepare to test V parameter */
		CROSS(qvec, tvec, edge1);

		/* calculate V parameter and test bounds */
		*v = DOT(dir, qvec) * inv_det;
		if (*v < 0.0f || *u + *v > 1.0f)
			return 0;

		/* calculate t, ray intersects triangle */
		*t = DOT(edge2, qvec) * inv_det;
//#endif
	}

	/// only allow intersections in the forward ray direction
	return int(*t >= 0.0f);
}
