/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "Vector3f.h"
#include "UnitVector3f.h"

#include "TriangleImplicit.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// check byte size
#define DESIRED_BYTE_SIZE 20
#if sizeof(TriangleImplicit) > DESIRED_BYTE_SIZE
	#error size of perceptuum2::modelling::TriangleImplicit is larger than DESIRED_BYTE_SIZE
#endif




/// standard object services -----------------------------------------------------------------------
TriangleImplicit::TriangleImplicit()
 :	index0_m(),
	index1_m(),
	index2_m(),
	interpolateNormal_m()
{
	TriangleImplicit::zeroTextureCoords();
}


TriangleImplicit::TriangleImplicit
(
	const word vertex0,
	const word vertex1,
	const word vertex2,
	const bool interpolateNormal
)
{
	TriangleImplicit::set( vertex0, vertex1, vertex2 );
	TriangleImplicit::set( interpolateNormal );
	TriangleImplicit::zeroTextureCoords();
}


TriangleImplicit::~TriangleImplicit()
{
}


TriangleImplicit::TriangleImplicit( const TriangleImplicit& other )
{
	TriangleImplicit::operator=( other );
}


TriangleImplicit& TriangleImplicit::operator=( const TriangleImplicit& other )
{
	if( &other != this )
	{
		index0_m = other.index0_m;
		index1_m = other.index1_m;
		index2_m = other.index2_m;

		interpolateNormal_m = other.interpolateNormal_m;

		for( int i = 3;  i-- > 0; )
		{
			for( int j = 2;  j-- > 0; )
			{
				txys_m[i][j] = other.txys_m[i][j];
			}
		}
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void TriangleImplicit::set
(
	const word vertex0,
	const word vertex1,
	const word vertex2
)
{
	index0_m = vertex0 >= 0 ? vertex0 : 0;
	index1_m = vertex1 >= 0 ? vertex1 : 0;
	index2_m = vertex2 >= 0 ? vertex2 : 0;
}


void TriangleImplicit::set
(
	const bool interpolateNormal
)
{
	interpolateNormal_m = int(interpolateNormal);
}


void TriangleImplicit::set
(
	const Vector2<word> txy[3]
)
{
	for( int i = 3;  i-- > 0; )
	{
		txys_m[i][0] = txy[i].x_m;
		txys_m[i][1] = txy[i].y_m;
	}
}


void TriangleImplicit::reverseOrientation()
{
	/// swap 1 and 2

	/// vertex indexs
	{
		const word temp = word(index1_m);
		index1_m = index2_m;
		index2_m = temp;
	}

	/// texture coords
	{
		word temp    = txys_m[1][0];
		txys_m[1][0] = txys_m[2][0];
		txys_m[2][0] = temp;

		temp         = txys_m[1][1];
		txys_m[1][1] = txys_m[2][1];
		txys_m[2][1] = temp;
	}
}




/// queries ----------------------------------------------------------------------------------------
bool TriangleImplicit::checkIndexesRange
(
	const dword vertexCount
) const
{
	return (index0_m < vertexCount) &
	       (index1_m < vertexCount) &
	       (index2_m < vertexCount);
}


bool TriangleImplicit::checkIndexesDistinct() const
{
	return (index0_m != index1_m) &
	       (index0_m != index2_m) &
	       (index1_m != index2_m);
}


void TriangleImplicit::getNormal
(
	const vector<Vector3f>& vertexPositions,
	UnitVector3f&           normal
) const
{
	Vector3f edge1( vertexPositions[index1_m] - vertexPositions[index0_m] );
	Vector3f edge2( vertexPositions[index2_m] - vertexPositions[index1_m] );

	normal = edge1.crossEq( edge2 );
}


float TriangleImplicit::getArea
(
	const vector<Vector3f>& vertexPositions
) const
{
	Vector3f edge1( vertexPositions[index1_m] - vertexPositions[index0_m] );
	Vector3f edge2( vertexPositions[index2_m] - vertexPositions[index1_m] );

	/// area of parallelogram
	const float length = edge1.crossEq( edge2 ).length();

	return length * 0.5f;
}


void TriangleImplicit::getTextureCoords
(
	Vector2<word> txy[3]
) const
{
	for( int i = 3;  i-- > 0; )
	{
		txy[i].x_m = txys_m[i][0];
		txy[i].y_m = txys_m[i][1];
	}
}


dword TriangleImplicit::getCommonVertexs
(
	const TriangleImplicit& other,
	bool& isFirst,
	bool& isSecond,
	bool& isThird
) const
{
	return dword(isFirst  = (index0_m == other.index0_m) | (index0_m == other.index1_m) | (index0_m == other.index2_m)) +
	       dword(isSecond = (index1_m == other.index0_m) | (index1_m == other.index1_m) | (index1_m == other.index2_m)) +
	       dword(isThird  = (index2_m == other.index0_m) | (index2_m == other.index1_m) | (index2_m == other.index2_m));
}




/// implementation ---------------------------------------------------------------------------------
void TriangleImplicit::zeroTextureCoords()
{
	for( int i = 3;  i-- > 0; )
	{
		for( int j = 2;  j-- > 0; )
		{
			txys_m[i][j] = 0;
		}
	}
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const TriangleImplicit& obj )
{
	/// write vertex indexs
	out.write( "vertexIndex0",      word(obj.index0_m) );
	out.write( "vertexIndex1",      word(obj.index1_m) );
	out.write( "vertexIndex2",      word(obj.index2_m) );
	out.write( "interpolateNormal", bool(obj.interpolateNormal_m) );

	/// write texture indexs
	{
		static const char* TX_NAMES[3][2] = { { "tx0", "ty0" }, { "tx1", "ty1" }, { "tx2", "ty2" } };
		for( int i = 0;  i < 3;  ++i )
		{
			for( int j = 0;  j < 2;  ++j )
			{
				out.write( TX_NAMES[i][j], obj.txys_m[i][j] );
			}
		}
		//out.write( "tx0", obj.txys_m[0][0] );
		//out.write( "ty0", obj.txys_m[0][1] );
		//out.write( "tx1", obj.txys_m[1][0] );
		//out.write( "ty1", obj.txys_m[1][1] );
		//out.write( "tx2", obj.txys_m[2][0] );
		//out.write( "ty2", obj.txys_m[2][1] );
	}

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, TriangleImplicit& obj )
{
	/// read vertex indexs
	{
		word vertex0;
		word vertex1;
		word vertex2;
		bool interpolateNormal;

		in.read( "vertexIndex0",      vertex0 );
		in.read( "vertexIndex1",      vertex1 );
		in.read( "vertexIndex2",      vertex2 );
		in.read( "interpolateNormal", interpolateNormal );

		obj.set( vertex0, vertex1, vertex2 );
		obj.set( interpolateNormal );
	}

	/// read texture indexs
	{
		static const char* TX_NAMES[3][2] = { { "tx0", "ty0" }, { "tx1", "ty1" }, { "tx2", "ty2" } };
		for( int i = 0;  i < 3;  ++i )
		{
			for( int j = 0;  j < 2;  ++j )
			{
				word t = 0;
				in.read( TX_NAMES[i][j], t );
				if( t < 0 )
				{
					t = 0;
				}
				obj.txys_m[i][j] = t;
			}
		}
	}

	return in;
}


const char* TriangleImplicit::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::TriangleImplicit";
	return CLASS_NAME;
}


const char* TriangleImplicit::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a triangle, of indexes into a vertex pool";
	return CLASS_DESCRIPTION;
}
