/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Projection.h"
#include "SamplingControl.h"

#include "Camera.h"   /// own header is included last


using namespace perceptuum2::transport;




/// statics
const char Camera::CLASS_NAME[] = "perceptuum2::transport::Camera";




/// standard object services -----------------------------------------------------------------------
Camera::Camera
(
	const Projection& projection,
	const dword       imageWidth,
	const dword       imageHeight
)
 :	pProjection_m      ( &projection ),
	imageWidth_m       ( imageWidth ),
	imageHeight_m      ( imageHeight ),
	halfImageWidth_m   (),
	halfImageHeight_m  (),
	oneOverImageWidth_m()
{
	imageWidth_m  = imageWidth_m  > 0 ? imageWidth_m  : 1;
	imageHeight_m = imageHeight_m > 0 ? imageHeight_m : 1;

	halfImageWidth_m    = float(imageWidth_m)  * 0.5f;
	halfImageHeight_m   = float(imageHeight_m) * 0.5f;
	oneOverImageWidth_m = 1.0f / float(imageWidth_m);
}


Camera::~Camera()
{
}


Camera::Camera( const Camera& other )
 :	pProjection_m      ( other.pProjection_m ),
	imageWidth_m       ( other.imageWidth_m ),
	imageHeight_m      ( other.imageHeight_m ),
	halfImageWidth_m   ( other.halfImageWidth_m ),
	halfImageHeight_m  ( other.halfImageHeight_m ),
	oneOverImageWidth_m( other.oneOverImageWidth_m )
{
}


Camera& Camera::operator=( const Camera& other )
{
	assign( other );

	return *this;
}




/// queries ----------------------------------------------------------------------------------------
void Camera::getRay
(
	const dword            x,
	const dword            y,
	const SamplingControl& samplingControl,
	Vector2f01&            subPixelPosition,
	Vector3f&              start,
	UnitVector3f&          direction
) const
{
	/// sub-pixel sample point 0-1, 0-1
	subPixelPosition = samplingControl.getPixelPoint();

	/// proportional position of pixel, relative to screen center
	float xProportional =  (float(x) + subPixelPosition.getX() - halfImageWidth_m ) * oneOverImageWidth_m;
	float yProportional = -(float(y) + subPixelPosition.getY() - halfImageHeight_m) * oneOverImageWidth_m;

	pProjection_m->getRay( xProportional, yProportional, samplingControl, start, direction );
}


void Camera::getWorldToViewTransform
(
	const SamplingControl& samplingControl,
	Matrix4f&              matrix
) const
{
	using hxa7241::utility::Float01;
	const Vector2f01 pixelPoint( samplingControl.getPixelPoint() );

	/// sub-pixel sample point [-0.5, +0.5)
	Vector2f subPixelPosition( pixelPoint.getX() - 0.5f, pixelPoint.getY() - 0.5f );
	subPixelPosition *= oneOverImageWidth_m;

	pProjection_m->getWorldToViewTransform( samplingControl, subPixelPosition, matrix );
}


void Camera::getViewToClipTransform
(
	const SamplingControl& samplingControl,
	Matrix4f&              matrix
) const
{
	const Vector2f oneOverWidthHeight( 1.0f / imageWidth_m, 1.0f / imageHeight_m );

	pProjection_m->getViewToClipTransform( samplingControl, oneOverWidthHeight, matrix );
}


void Camera::getClipToScreenScaling
(
	Vector2f& xyScaling
) const
{
	xyScaling.setXY( imageWidth_m, imageHeight_m );
}




/// implementation ---------------------------------------------------------------------------------
void Camera::assign
(
	const Camera& other
)
{
	if( &other != this )
	{
		pProjection_m       = other.pProjection_m;
		imageWidth_m        = other.imageWidth_m;
		imageHeight_m       = other.imageHeight_m;
		halfImageWidth_m    = other.halfImageWidth_m;
		halfImageHeight_m   = other.halfImageHeight_m;
		oneOverImageWidth_m = other.oneOverImageWidth_m;
	}
}
