/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "MediaStack.h"
#include "SpacetimeIndex.h"
#include "SamplingControl.h"
#include "SamplingControlPath.h"
#include "Illumination.h"
#include "EyePathNode.h"
#include "Medium.h"
#include "LightStream.h"
#include "Scene.h"
#include "ObjectRenderableId.h"
#include "Ray.h"
#include "Intersection.h"
#include "Vector2.h"
#include "Vector3f.h"
#include "UnitVector3f.h"
#include "Float01.h"

#include "EyePath.h"   /// own header is included last


using namespace perceptuum2::transport;




/// standard object services -----------------------------------------------------------------------
EyePath::EyePath
(
	const dword    pathLengthLimit,
	const Float01c pathEndThreshold
)
 :	pathLengthLimit_m ( pathLengthLimit > 0 ? pathLengthLimit : 1 ),
	pathEndThreshold_m( pathEndThreshold ),
	pathLength_m      ( 0 ),
	eyeEndLight_m     ()
{
}


EyePath::~EyePath()
{
}




/// commands ---------------------------------------------------------------------------------------
void EyePath::trace
(
	const SamplingControlPath&    pathSamplingControl,
	const SamplingControlDiffuse& diffuseSamplingControl,
	const SpacetimeIndex&         sceneIndex,
	const Illumination&           illumination,
	const Ray&                    initialRay
)
{
	makeIlluminationFromPath( pathSamplingControl, diffuseSamplingControl, sceneIndex, illumination, initialRay );


//	/// add path pseudo terminator (located at the eye)
//	{
//		const perceptuum2::modelling::ObjectRenderableId defaultObject( sceneIndex.getScene().getDefaultObject() );
//		perceptuum2::modelling::Intersection it;
//		it.object_m   = defaultObject;
//		it.position_m = initialRay.getPosition();
//		nodes_m[0].setEyePathNode( it, defaultObject, perceptuum2::interaction::LightStream( initialRay.getDirection(), Light::ONE() ) );
//	}
//
//	/// do the tracing
//	buildAndIlluminatePath( samplingControl, sceneIndex, illumination, initialRay );
}


void EyePath::traceFromFirstHit
(
	const SamplingControlPath&    ,//pathSamplingControl,
	const SamplingControlDiffuse& ,//diffuseSamplingControl,
	const SpacetimeIndex&         ,//sceneIndex,
	const Illumination&           ,//illumination,
	const EyePathNode&            //eyePathNode
)
{



//	/// get initial ray
//	Ray initialRay;
//	eyePathNode.getInwardPathStepRay( initialRay );
//
//	/// add path terminator
//	nodes_m[0] = eyePathNode;
//
//	/// do the tracing
//	buildAndIlluminatePath( samplingControl, sceneIndex, illumination, initialRay );
}




/// queries ----------------------------------------------------------------------------------------
void EyePath::getEyeEndLight
(
	bool&  isHit,
	Light& light
) const
{
	isHit = pathLength_m > 0;
	light = eyeEndLight_m;
}


void EyePath::getEyeEndLight
(
	Light& light
) const
{
	light = eyeEndLight_m;
}


dword EyePath::getPathLength() const
{
	return pathLength_m;
}




/// implementation ---------------------------------------------------------------------------------
void EyePath::makeIlluminationFromPath
(
	const SamplingControlPath&    pathSamplingControl,
	const SamplingControlDiffuse& diffuseSamplingControl,
	const SpacetimeIndex&         sceneIndex,
	const Illumination&           illumination,
	const Ray&                    initialRay
)
{
	/// set initial scaling to one
	/// set initial emitted sum to zero
	Light cumulativeScaling( Light::ONE() );
	Light emittedSum;

	/// make tracing things
	Ray ray( initialRay );
	MediaStack media( sceneIndex.getScene().getDefaultObject() );

	/// step through eye path bounces
	dword pathLength            = 0;
	bool  isSceneEndDiffuse     = false;
	bool  isDirectIllumSeparate = false;
	EyePathNode  eyePathNode;
	UnitVector3f prevStepDirection;
	do
	{
		/// intersect ray, and end path if nothing hit
		perceptuum2::modelling::Intersection intersection;
		if( !sceneIndex.getNextIntersection( ray, intersection ) )
		{
			break;
		}

		/// make eye path node for hit point
		{
			using perceptuum2::modelling::ObjectRenderableId;
			const ObjectRenderableId outsideObject( media.getOutsideObject( ray.getDirection().getVector(), intersection.normal_m.getVector(), intersection.object_m ) );
			eyePathNode.setSurfacePoint( intersection, outsideObject );
		}

		/// accumulate medium participation scaling
		{
			perceptuum2::interaction::Medium medium( media.getTop() );
			medium.participateLight( ray.getPosition(), eyePathNode.getPosition(), cumulativeScaling );
		}

		/// get emitted from node, scale by cumulative scaling, add to emitted sum
		if( !isDirectIllumSeparate )
		{
			Light emitted;
			eyePathNode.getEmittedLight( -ray.getDirection(), emitted );
			emitted *= cumulativeScaling;
			emittedSum += emitted;
		}

		/// get next path step, or not if node is diffuse
		Light stepScaling;
		{
			hxa7241::utility::Float01 stochastics[8];
			pathSamplingControl.getSampleCoords( pathLength, stochastics );

			prevStepDirection = ray.getDirection();
			eyePathNode.getNextPathStep( stochastics, prevStepDirection, ray, stepScaling, isSceneEndDiffuse, isDirectIllumSeparate );
		}

		if( !isSceneEndDiffuse )
		{
			/// get direct illumination for node separately through brdf
			if( isDirectIllumSeparate )
			{
				const Float01c shinynessPoint( Float01c::ONE() );

				Light directLight;
				illumination.getDirectIllumination( diffuseSamplingControl, sceneIndex, eyePathNode, media, prevStepDirection.negate(), shinynessPoint,
				                                    directLight );
				directLight *= cumulativeScaling;
				emittedSum += directLight;
			}

			/// update media stack
			media.update( ray );

			/// accumulate step scaling
			cumulativeScaling *= stepScaling;
		}

		/// if diffuse, break loop
		/// if path length limit reached, break loop
		/// if cumulative scaling less than threshold, break loop
	} while( !isSceneEndDiffuse & (++pathLength < pathLengthLimit_m) & (cumulativeScaling.average() >= pathEndThreshold_m) );


	/// if scene end is diffuse, join
	if( isSceneEndDiffuse )
	{
		const Float01c shinynessPoint( Float01c::ONE() );

		/// get illumination for last eye path node (including emitted)
		Light eyewardLight;
		illumination.getIllumination( diffuseSamplingControl, sceneIndex, eyePathNode, media, prevStepDirection.negate(), shinynessPoint,
		                              eyewardLight );

		/// scale eyeward light by cumulative scaling
		eyewardLight *= cumulativeScaling;

		/// make eye light from emitted sum added to eyeward light
		eyeEndLight_m = (eyewardLight += emittedSum);
		pathLength_m  = pathLength;
	}
	else
	{
		eyeEndLight_m = Light::ZERO();
		pathLength_m  = pathLength;
	}
}




/*void EyePath::buildAndIlluminatePath
(
	const SamplingControl& samplingControl,
	const SpacetimeIndex&  sceneIndex,
	const Illumination&    illumination,
	const Ray&             initialRay
)
{
	/// send a path out
	pathLength_m = buildPath( samplingControl, sceneIndex, initialRay, pathLengthLimit_m, nodes_m );

	/// bring the light back along the path
	illuminatePath( sceneIndex, illumination, nodes_m, pathLength_m, eyeEndLight_m );
}


dword EyePath::buildPath
(
	const SamplingControl& samplingControl,
	const SpacetimeIndex&  sceneIndex,
	const Ray&             initialRay,
	const dword            pathLengthLimit,
	vector<EyePathNode>&   nodes
) const
{
	/// make tracing things
	Ray ray( initialRay );
	MediaStack media( sceneIndex.getScene().getDefaultObject() );

	/// step thru eye path ray bounces, acquiring eye path nodes
	dword pathLength = 0;
	while( true )
	{
		/// intersect ray, and end path if nothing hit
		perceptuum2::modelling::Intersection intersection;
		if( !sceneIndex.getNextIntersection( ray, intersection ) )
		{
			break;
		}

		/// advance loop
		EyePathNode& eyePathNode = nodes[ ++pathLength ];

		/// set eye path node in list
		{
			using perceptuum2::modelling::ObjectRenderableId;
			const ObjectRenderableId outsideObject( media.getOutsideObject( ray.getDirection().getVector(), intersection.normal_m.getVector(), intersection.object_m ) );
			eyePathNode.setSurfacePoint( intersection, outsideObject );
		}

		/// decide if path ends
		if( (pathLength >= pathLengthLimit) ||
			(eyePathNode.getSpecularness() <= samplingControl.getSamplingPoint(samplingControl.SPECULARNESS)) )
		{
			break;
		}

		/// get bounced ray
		{
			using hxa7241::utility::Float01;
			using hxa7241::graphics::Vector2;
			const Vector2<Float01> stochastic( samplingControl.getSamplingPoint(samplingControl.EYE_INTERACTION_X),
			                                   samplingControl.getSamplingPoint(samplingControl.EYE_INTERACTION_Y) );
			const Float01          roulette( samplingControl.getSamplingPoint(samplingControl.EYE_INTERACTION_R) );

			eyePathNode.addInwardPathStep( ray.getDirection(), stochastic, roulette );
			eyePathNode.getInwardPathStepRay( ray );
		}

		/// decide if path ends
		if( ray.getDirection().isZero() )
		{
			break;
		}

		/// update media stack
		media.update( ray );
	}

	return pathLength;
}


void EyePath::illuminatePath
(
	const SpacetimeIndex&      sceneIndex,
	const Illumination&        illumination,
	const vector<EyePathNode>& nodes,
	const dword                pathLength,
	Light&                     eyeEndLight
) const
{
	Light light;
	if( pathLength > 0 )
	{
		dword nodeIndex = pathLength;

		using hxa7241::graphics::UnitVector3f;
		const EyePathNode*  pEyePathNode      = &(nodes[nodeIndex]);
		const UnitVector3f* pEyewardDirection = &(nodes[nodeIndex - 1].getNextDirection());

		/// illuminate last eye path node
		illumination.getIllumination( sceneIndex, *pEyePathNode, *pEyewardDirection, light );

		/// step thru eye path nodes propagating illumination, last to first - in photon direction
		while( true )
		{
			/// calc light change through medium between successive eye path nodes
			{
				using perceptuum2::modelling::ObjectRenderableId;
				const ObjectRenderableId&          mediumObject = pEyePathNode->getObject( pEyewardDirection->getVector() );
				const hxa7241::graphics::Vector3f& nextPosition = nodes[nodeIndex - 1].getPosition();
				perceptuum2::interaction::Medium   medium( mediumObject );
				medium.participateLight( pEyePathNode->getPosition(), nextPosition, light );
			}

			/// advance loop
			if( --nodeIndex <= 0 )
			{
				break;
			}
			pEyePathNode      = &(nodes[nodeIndex]);
			pEyewardDirection = &(nodes[nodeIndex - 1].getNextDirection());

			/// get bounced light from eye path node
			pEyePathNode->propagateLight( light, *pEyewardDirection, light );
		}
	}

	/// set eye end value
	eyeEndLight = light;
}*/
