/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef OctreeVisitorPhoton_h
#define OctreeVisitorPhoton_h


#include "Primitives.h"
#include "Photon.h"
#include "Octree.h"
#include <utility>
#include <vector>

#include "hxa7241_graphics.h"




#include "perceptuum2_transport.h"
namespace perceptuum2 { namespace transport {
	using namespace hxa7241::primitives;
	using hxa7241::graphics::OctreeVisitorT;
	using hxa7241::graphics::OctreeVisitor;
	using hxa7241::graphics::OctreeBound;
	using hxa7241::graphics::Vector3f;
	using std::vector;


/**
 * an octree visitor for photon mapping density estimation.<br/><br/>
 *
 * the selection condition is intersection with a sphere 'shell' - that is a hollow
 * sphere with a thick crust.<br/><br/>
 *
 * the visit processing is to append a pointer to the item, and its distance squared
 * from the shell center, to a collection.<br/><br/>
 *
 * the selection implementation is conservative not precise, which means occasional
 * descent further down the octree than necessary.
 */
class OctreeVisitorPhoton
	: public OctreeVisitorT<Photon>
{
/// standard object services -----------------------------------------------------------------------
public:
	         /**
	          * for inserting or removing.
	          */
	         OctreeVisitorPhoton();
	         /**
	          * for visiting.
	          */
	         OctreeVisitorPhoton( vector<std::pair<float, const Photon*> >* pSelectedItemsBuffer,
	                              const Vector3f& shellCenter,
	                              float           shellRadiusOuter,
	                              float           shellRadiusInner );

	virtual ~OctreeVisitorPhoton();
private:
	         OctreeVisitorPhoton( const OctreeVisitorPhoton& );
	OctreeVisitorPhoton& operator=( const OctreeVisitorPhoton& );


/// commands ---------------------------------------------------------------------------------------
protected:
	virtual void  visitItem( const Photon& item,
	                         bool          isInside );


/// queries ----------------------------------------------------------------------------------------
	virtual bool  isOverlappingCell( const Photon&   item,
	                                 const Vector3f& position,
	                                 float           size )                                   const;
	virtual dword getSubcellOverlaps( const Photon&   item,
	                                  const Vector3f& planePositions )                        const;

public:
	virtual ESelection isSelected( const OctreeBound&,
	                               bool isInside )                                            const;
	virtual dword      getSubcellSelections( const OctreeBound&,
	                                         bool isInside )                                  const;
	virtual dword      getVisitCount()                                                        const;


/// implementation ---------------------------------------------------------------------------------
protected:
	static  bool       isPhotonOverlappingShell( const Photon&   item,
	                                             const Vector3f& center,
	                                             float           radiusOuter,
	                                             float           radiusInner );
	static  ESelection isCellOverlappingShell( const OctreeBound& cellBound,
	                                           const Vector3f&    center,
	                                           float              outerRadius,
	                                           float              innerRadius );
	static  dword      getSubcellShellOverlaps( const OctreeBound& cellBound,
	                                            const Vector3f&    center,
	                                            float              outerRadius,
	                                            float              innerRadius );

	static  void  getComponentDistancesSquared( const Vector3f& cellCenter,
	                                            float           cellSize,
	                                            const Vector3f& position,
	                                            Vector3f        dxyzs[2] );
	static  float getDistanceSquared( const Photon&,
	                                  const Vector3f& );


/// fields -----------------------------------------------------------------------------------------
private:
	/// external selection buffer
	vector<std::pair<float, const Photon*> >* pSelectedItems_m;

	/// shell
	Vector3f center_m;
	float    radiusOuter_m;
	float    radiusInner_m;

	dword visitCount_m;
};








/// inlines ///

/// implementation ---------------------------------------------------------------------------------
inline
float OctreeVisitorPhoton::getDistanceSquared
(
	const Photon&   photon,
	const Vector3f& point
)
{
	const Vector3f& photonPosition = photon.getPosition();

	const float dx = photonPosition.getX() - point.getX();
	const float dy = photonPosition.getY() - point.getY();
	const float dz = photonPosition.getZ() - point.getZ();

	return (dx * dx) + (dy * dy) + (dz * dz);
}


}}




#endif//OctreeVisitorPhoton_h
