/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Projection.h"
#include "samplingControl.h"
#include "Illumination.h"
#include "Light.h"
#include "Ray.h"
#include "ImageAccumulation.h"
#include "CieXyz.h"
#include "Float01.h"
#include "Logger.h"

#include "RasterizerTracing.h"   /// own header is included last


using namespace perceptuum2::transport;




/// standard object services -----------------------------------------------------------------------
RasterizerTracing::RasterizerTracing
(
	const Projection& projection,
	const dword       imageWidth,
	const dword       imageHeight,
	const dword       eyePathLengthLimit,
	const Float01c    eyePathEndThreshold,
	const bool        isLogging
)
 :	camera_m                 ( projection, imageWidth, imageHeight ),
	eyePath_m                ( eyePathLengthLimit, eyePathEndThreshold ),
	totalEyePathCount_m      ( 0 ),
	totalEyePathLengthCount_m( 0 ),
	totalOverdrawCount_m     ( 0 ),
	isLogging_m              ( isLogging )
{
}


RasterizerTracing::~RasterizerTracing()
{
}




/// commands ---------------------------------------------------------------------------------------
void RasterizerTracing::rasterizeFrame
(
	const SamplingControl& samplingControl,
	const SpacetimeIndex&  sceneIndex,
	const Illumination&    illumination,
	ImageAccumulation&     accumulation
)
{
	const dword width  = accumulation.getWidth();
	const dword height = accumulation.getHeight();

	/// step through pixels
	for( dword y = 0;  y < height;  ++y )
	{
		for( dword x = 0;  x < width;  ++x )
		{
			if( accumulation.shouldSample( x, y ) )
			{
				using hxa7241::utility::Float01;

				/// get initial ray from camera
				perceptuum2::modelling::Ray         initialRay;
				hxa7241::graphics::Vector2<Float01> subPixelPosition;
				{
					hxa7241::graphics::Vector3f     start;
					hxa7241::graphics::UnitVector3f direction;
					camera_m.getRay( x, y, samplingControl, subPixelPosition, start, direction );
					initialRay.set( start, direction );
				}

				/// trace eye path for pixel
				eyePath_m.trace( samplingControl.getEyePathControl(), samplingControl.getDiffuseControl(), sceneIndex, illumination, initialRay );

				/// get path values at eye end
				hxa7241::graphics::CieXyz xyz;
				float alpha;
				{
					bool   didHit;
					Light  light;
					eyePath_m.getEyeEndLight( didHit, light );

					light.getXYZ( xyz );
					alpha = didHit ? FLOAT_ALMOST_ONE : 0.0f;
				}

				/// add eye end values to image accumulation
				accumulation.addSample( x, y, subPixelPosition.getX(), subPixelPosition.getY(), xyz, alpha );

				/// track overall stats
				++totalEyePathCount_m;
				totalEyePathLengthCount_m += eyePath_m.getPathLength();
			}
		}
	}


	if(isLogging_m) hxa7241::utility::Logger() << "RasterizerTracing -  totalEyePathCount: " << totalEyePathCount_m << "  totalEyePathLengthCount: " << totalEyePathLengthCount_m;
}
