/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"

#include "SamplingControlPath.h"   /// own header is included last


using namespace perceptuum2::transport;




/// standard object services -----------------------------------------------------------------------
SamplingControlPath::SamplingControlPath
(
	const dword sobolDimensions[],
	const dword randomSeeds[],
	const dword specialLevels,
	const dword laterSeed
)
 :	sobols_m     ( SOBOL_COUNT  * (specialLevels >= 0 ? specialLevels : 0) ),
	randoms_m    ( RANDOM_COUNT * (specialLevels >= 0 ? specialLevels : 0) ),
	laterRandom_m( laterSeed ),
	laterCoords_m()
{
	for( dword i = sobols_m.size();  i-- > 0; )
	{
		sobols_m[i].setDimension( sobolDimensions[i] );
	}
	for( dword i = randoms_m.size();  i-- > 0; )
	{
		randoms_m[i].setSeed( randomSeeds[i] );
	}
}


SamplingControlPath::~SamplingControlPath()
{
}


SamplingControlPath::SamplingControlPath
(
	const SamplingControlPath& other
)
 :	sobols_m     ( other.sobols_m ),
	randoms_m    ( other.randoms_m ),
	laterRandom_m( other.laterRandom_m ),
	laterCoords_m( other.laterCoords_m )
{
}


SamplingControlPath& SamplingControlPath::operator=
(
	const SamplingControlPath& other
)
{
	if( &other != this )
	{
		sobols_m      = other.sobols_m;
		randoms_m     = other.randoms_m;
		laterRandom_m = other.laterRandom_m;
		laterCoords_m = other.laterCoords_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void SamplingControlPath::nextPath()
{
	for( dword i = sobols_m.size();  i-- > 0; )
	{
		sobols_m[i].next();
	}
	for( dword i = randoms_m.size();  i-- > 0; )
	{
		randoms_m[i].next();
	}

	laterCoords_m.resize( 0 );
}




/// queries ----------------------------------------------------------------------------------------
void SamplingControlPath::getSampleCoords
(
	dword   pathNodeIndex,
	Float01 coords[COORDS_COUNT]
) const
{
	/// precondition input
	if( pathNodeIndex < 0 )
	{
		pathNodeIndex = 0;
	}


	/// first nodes use sobols and randoms
	const dword specialNodes = sobols_m.size() / (COORDS_COUNT >> 1);
	if( pathNodeIndex < specialNodes )
	{
		const dword j = pathNodeIndex * (COORDS_COUNT >> 1);
		for( dword i = (COORDS_COUNT >> 1);  i-- > 0; )
		{
			coords[i]               = sobols_m [j + i].getItem();
			coords[i + SOBOL_COUNT] = randoms_m[j + i].getItem();
		}

		//const dword s = pathNodeIndex * SOBOL_COUNT;
		//const dword r = pathNodeIndex * RANDOM_COUNT;
		//
		//coords[0] = sobols_m [s + 0].getItem();
		//coords[1] = sobols_m [s + 1].getItem();
		//coords[2] = sobols_m [s + 2].getItem();
		//coords[3] = sobols_m [s + 3].getItem();
		//coords[4] = randoms_m[r + 0].getItem();
		//coords[5] = randoms_m[r + 1].getItem();
		//coords[6] = randoms_m[r + 2].getItem();
		//coords[7] = randoms_m[r + 3].getItem();
	}
	/// later nodes use the later rand cache
	else
	{
		const dword laterCoordSetIndex  = pathNodeIndex - specialNodes;
		const dword laterCoordSetLength = dword(laterCoords_m.size()) / COORDS_COUNT;

		/// check if rand cache needs to be extended
		if( laterCoordSetIndex >= laterCoordSetLength )
		{
			/// add more of the rand sequence on the end
			const dword extend = (laterCoordSetIndex - laterCoordSetLength) + 1;
			for( dword i = extend;  i-- > 0; )
			{
				for( dword j = COORDS_COUNT;  j-- > 0; )
				{
					laterRandom_m.next();
					laterCoords_m.push_back( laterRandom_m.getFloat() );
				}
			}
		}

		/// read coords from cache
		for( dword i = COORDS_COUNT;  i-- > 0; )
		{
			coords[i] = laterCoords_m[(laterCoordSetIndex * COORDS_COUNT) + i];
		}
	}
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::transport::operator<<( SerializerOut& out, const SamplingControlPath& obj )
{
	{
		const dword sobolsLength = obj.sobols_m.size();
		out.write( "sobolsLength", sobolsLength );

		for( dword i = 0;  i < sobolsLength;  ++i )
		{
			out.write( "sobol", obj.sobols_m[i] );
		}
	}
	{
		const dword randomsLength = obj.randoms_m.size();
		out.write( "randomsLength", randomsLength );

		for( dword i = 0;  i < randomsLength;  ++i )
		{
			out.write( "random", obj.randoms_m[i] );
		}
	}

	out.write( "laterRandom", obj.laterRandom_m );
	{
		const dword laterCoordsLength = obj.laterCoords_m.size();
		out.write( "laterCoordsLength", laterCoordsLength );

		for( dword i = 0;  i < laterCoordsLength;  ++i )
		{
			out.write( "laterCoord", obj.laterCoords_m[i] );
		}
	}

	return out;
}


SerializerIn& perceptuum2::transport::operator>>( SerializerIn& in, SamplingControlPath& obj )
{
	{
		dword sobolsLength = 0;
		in.read( "sobolsLength", sobolsLength );
		sobolsLength = sobolsLength >= 0 ? sobolsLength : 0;

		obj.sobols_m.resize( sobolsLength );
		for( dword i = 0;  i < sobolsLength;  ++i )
		{
			in.read( "sobol", obj.sobols_m[i] );
		}
	}
	{
		dword randomsLength = 0;
		in.read( "randomsLength", randomsLength );
		randomsLength = randomsLength >= 0 ? randomsLength : 0;

		obj.randoms_m.resize( randomsLength );
		for( dword i = 0;  i < randomsLength;  ++i )
		{
			in.read( "random", obj.randoms_m[i] );
		}
	}

	in.read( "laterRandom", obj.laterRandom_m );
	{
		dword laterCoordsLength = 0;
		in.read( "laterCoordsLength", laterCoordsLength );
		laterCoordsLength = laterCoordsLength >= 0 ? laterCoordsLength : 0;

		obj.laterCoords_m.resize( laterCoordsLength );
		for( dword i = 0;  i < laterCoordsLength;  ++i )
		{
			in.read( "laterCoord", obj.laterCoords_m[i] );
		}
	}

	return in;
}


const char* SamplingControlPath::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::transport::SamplingControlPath";
	return CLASS_NAME;
}


const char* SamplingControlPath::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "grouping of stochastic sequences used in path sampling";
	return CLASS_DESCRIPTION;
}
