/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Camera.h"
#include "ClipFlags.h"
#include "Vector3f.h"
#include "UnitVector3f.h"
#include "hxa7241math.h"

#include "Projector.h"   /// own header is included last


using namespace perceptuum2::transport::projection;




/// standard object services -----------------------------------------------------------------------
Projector::Projector
(
	const Camera&          camera,
	const SamplingControl& samplingControl
)
 :	worldToClip_m (),
	modelToClip_m (),
	modelToWorld_m(),
	clipToScreen_m()
{
	/// set world to clip
	{
		Matrix4f viewToClip;
		camera.getWorldToViewTransform( samplingControl, worldToClip_m );
		camera.getViewToClipTransform( samplingControl, viewToClip );

		worldToClip_m.multiplyEqPost( viewToClip );
	}

	/// set model to clip
	modelToClip_m = worldToClip_m;

	/// set clip to screen
	camera.getClipToScreenScaling( clipToScreen_m );
}


Projector::~Projector()
{
}


Projector::Projector
(
	const Projector& other
)
 :	worldToClip_m ( other.worldToClip_m  ),
	modelToClip_m ( other.modelToClip_m  ),
	modelToWorld_m( other.modelToWorld_m ),
	clipToScreen_m( other.clipToScreen_m )
{
}


Projector& Projector::operator=
(
	const Projector& other
)
{
	if( &other != this )
	{
		worldToClip_m  = other.worldToClip_m;
		modelToClip_m  = other.modelToClip_m;
		modelToWorld_m = other.modelToWorld_m;
		clipToScreen_m = other.clipToScreen_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Projector::setModelToClip
(
	const Matrix4f& modelToClip
)
{
	modelToClip_m = modelToClip;
}




/// queries ----------------------------------------------------------------------------------------
const Matrix4f& Projector::getWorldToClip() const
{
	return worldToClip_m;
}


/*void Projector::transformModelToScreen
(
	const Vector3f& modelPosition,
	Vector3f&       screenPosition
) const
{
	Vector4f position;

	/// transform model point to screen space (model -> screen)
	modelToScreen_m.multiply( modelPosition, position );

	/// do perspective divide
	if( position.getW() != 0.0f )
	{
		const float oneOverW = 1.0f / position.getW();
		screenPosition.setXYZ( position.getX() * oneOverW,
							   position.getY() * oneOverW,
							   position.getZ() * oneOverW );
	}
	else
	{
		screenPosition = Vector3f::ZERO();
	}
}*/


void Projector::transformModelToScreen
(
	const ClipFlags objectClipflags,
	const Vector3f& modelPosition,
	Vector3f&       screenPosition,
	ClipFlags&      clipflags
) const
{
	Vector4f clipPosition;

	/// transform model point to clip space (model -> clip)
	modelToClip_m.multiply( modelPosition, clipPosition );

	/// check if definitely unclipped
	if( objectClipflags.isUnclipped() )
	{
		clipflags.zero();
	}
	else
	{
		/// clip point, set point clipflags
		clipflags.setForPoint( clipPosition );
	}

	/// transform clip space point to screen space (clip -> screen)
	if( clipPosition.getW() != 0.0f )
	{
		const float positionX = (clipPosition.getX() - 0.5f) * clipToScreen_m.getX();
		const float positionY = (clipPosition.getY() - 0.5f) * clipToScreen_m.getY();

		const float oneOverW = 1.0f / clipPosition.getW();
		screenPosition.setXYZ( positionX           * oneOverW,
							   positionY           * oneOverW,
							   clipPosition.getZ() * oneOverW );
	}
	else
	{
		screenPosition = Vector3f::ZERO();
	}
}


void Projector::transformModelToWorld
(
	const Vector3f& modelPoint,
	Vector3f&       worldPoint
) const
{
	modelToWorld_m.multiply( modelPoint, worldPoint );
}


void Projector::transformWorldToClip
(
	const Vector3f& worldPoint,
	Vector4f&       clipPoint
) const
{
	worldToClip_m.multiply( worldPoint, clipPoint );
}
