/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "perceptuum2.rc.h"
#include "Perceptuum2.h"
#include <commctrl.h>

#include "InfoDialog.h"   /// own header is included last


using namespace perceptuum2::platform_ui;




/// standard object services -----------------------------------------------------------------------
InfoDialog::InfoDialog
(
	HINSTANCE hInstance,
	HWND      hWndParent
)
 :	Win32Dialog( hInstance, hWndParent ),
	hWndTabs_m( 0 ),
	hWndTexts_m(),
	hFont_m( 0 ),
	pNames_m( 0 ),
	pSummaries_m( 0 )
{
	static const int BUF_LEN = 32;
	TCHAR buf[BUF_LEN] = TEXT("");
	::LoadString( hInstance, IDS_INFO_TEXT_FONTNAME, buf, BUF_LEN );

	hFont_m = ::CreateFont( INFO_TEXT_FONTSIZE, 0, 0, 0, 0,
	                        false, false, false,
	                        DEFAULT_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
	                        FIXED_PITCH | FF_MODERN, buf );
}


InfoDialog::~InfoDialog()
{
	if( hFont_m != 0 )
	{
		::DeleteObject( hFont_m );
	}
}




/// commands ---------------------------------------------------------------------------------------
int InfoDialog::run
(
	vector<string>* pNames,
	vector<string>* pSummaries
)
{
	int ok = FALSE;

	if( (pNames != 0) && (pSummaries != 0) )
	{
		if( pNames->size() == pSummaries->size() )
		{
			pNames_m = pNames;
			for( int i = pSummaries->size();  i-- > 0; )
			{
				convertText( true, (*pSummaries)[i] );
			}
			pSummaries_m = pSummaries;

			Win32Dialog::run();

			ok = TRUE;
		}
	}

	return ok;
}


int InfoDialog::run()
{
	return Win32Dialog::run();
}




/// overrides --------------------------------------------------------------------------------------
int InfoDialog::getDialogResourceId() const
{
	return IDD_INFO;
}


bool InfoDialog::dialogProc
(
	HWND   hWndDlg,
	UINT   message,
	WPARAM wParam,
	LPARAM lParam
)
{
	bool didHandleMessage = true;

	switch( message )
	{
	case WM_INITDIALOG :
		{
			bool ok;

			/// create tabs
			hWndTabs_m = createTabs( hWndDlg, hInstance_m, *pNames_m );
			ok = (hWndTabs_m != NULL);

			/// create text controls
			if( ok )
			{
				ok &= createTextDisplays( hInstance_m, hWndTabs_m, hFont_m, *pSummaries_m, hWndTexts_m );
			}

			/// set initial tab to display
			if( ok )
			{
				::SendMessage( hWndTabs_m, TCM_SETCURSEL, 0, 0 );
				::ShowWindow( hWndTexts_m[0], SW_SHOW );
			}

			if( !ok )
			{
				::EndDialog( hWndDlg, FALSE );
			}

			break;
		}
	case WM_NOTIFY :
		{
			const LPNMHDR pNotification = (LPNMHDR)lParam;
			const UINT    code          = pNotification->code;
			switch( code )
			{
			case TCN_SELCHANGING :
			case TCN_SELCHANGE :
				{
					const int tabIndex = ::SendMessage( hWndTabs_m, TCM_GETCURSEL, 0, 0 );
					if( tabIndex != -1 )
					{
						const int showOrHide = (code == TCN_SELCHANGING) ? SW_HIDE : SW_SHOW;
						::ShowWindow( hWndTexts_m[tabIndex], showOrHide );
					}

					break;
				}
			default :
				{
					didHandleMessage = false;
					break;
				}
			}

			break;
		}
	case WM_COMMAND :
		{
			switch( wParam )
			{
			case IDOK :
			case IDCANCEL :
				{
					::EndDialog( hWndDlg, wParam == IDOK ? TRUE : FALSE );
					break;
				}
			default :
				{
					didHandleMessage = false;
					break;
				}
			}

			break;
		}
	default:
		didHandleMessage = false;
		break;
	}

	return didHandleMessage;
}




/// implementation ---------------------------------------------------------------------------------
HWND InfoDialog::createTabs
(
	HWND                  hWndDlg,
	HINSTANCE             hInstance,
	const vector<string>& names
)
{
	/// get position and size
	RECT tabsRect;
	bool ok = getControlRect( hWndDlg, IDC_INFO_TABS_POS, tabsRect );

	/// create the window
	HWND hWndTabs = NULL;
	if( ok )
	{
		hWndTabs = ::CreateWindow( WC_TABCONTROL, NULL,
		                           WS_CHILD | WS_VISIBLE | WS_TABSTOP | WS_CLIPSIBLINGS | TCS_FIXEDWIDTH,
		                           tabsRect.left, tabsRect.top,
		                           tabsRect.right - tabsRect.left, tabsRect.bottom - tabsRect.top,
		                           hWndDlg, NULL, hInstance, NULL );
		if( hWndTabs != NULL )
		{
			HFONT font = (HFONT)::GetStockObject( DEFAULT_GUI_FONT );
			::SendMessage( hWndTabs, WM_SETFONT, (WPARAM)font, (LPARAM)0 );

			TC_ITEM tabItem = { TCIF_TEXT, 0, 0, 0, 0, -1, 0 };

			for( udword i = 0;  (i < names.size()) && ok;  ++i )
			{
				tabItem.pszText = const_cast<char*>( names[i].c_str() );
				const int t = ::SendMessage( hWndTabs, TCM_INSERTITEM, (WPARAM)i, (LPARAM)(&tabItem) );
				ok = (t != -1);
			}
		}
	}

	return ok ? hWndTabs : 0;
}


bool InfoDialog::createTextDisplays
(
	HINSTANCE             hInstance,
	HWND                  hWndTabs,
	HFONT                 hFont,
	const vector<string>& summaries,
	vector<HWND>&         hWndTexts
)
{
	/// get position and size
	RECT rect;
	bool ok = bool( ::GetClientRect( hWndTabs, &rect ) );
	if( ok )
	{
		::SendMessage( hWndTabs, TCM_ADJUSTRECT, (WPARAM)FALSE, (LPARAM)(&rect) );
		rect.top += 4;
		::InflateRect( &rect, -1, -1 );
	}
	const POINT pos  = { rect.left, rect.top };
	const POINT size = { rect.right - rect.left, rect.bottom - rect.top };


	/// create the windows
	if( ok )
	{
		const udword length = summaries.size();
		hWndTexts.resize( length );

		for( udword i = 0;  (i < length) && ok;  ++i )
		{
			HWND hWndText = ::CreateWindow( TEXT("EDIT"), NULL,
			                                WS_CHILD | WS_VSCROLL | ES_LEFT | ES_MULTILINE | ES_AUTOVSCROLL | ES_READONLY,
			                                pos.x, pos.y,
			                                size.x, size.y,
			                                hWndTabs, NULL, hInstance, NULL );
			ok &= (hWndText != NULL);
			if( ok )
			{
				/// set font
				if( hFont != 0 )
				{
					::SendMessage( hWndText, WM_SETFONT, (WPARAM)hFont, 0 );
				}

				/// set tab size
				const udword tab = 16;
				::SendMessage( hWndText, EM_SETTABSTOPS, (WPARAM)1, (LPARAM)(&tab) );

				/// set the margins
				const int scrollWidth = ::GetSystemMetrics( SM_CXVSCROLL );
				RECT border = { 0, 0, size.x - scrollWidth, size.y };
				::InflateRect( &border, -4, -4 );
				::SendMessage( hWndText, EM_SETRECT, 0, (LPARAM)(&border) );

				//const udword margin = 2;
				//::SendMessage( hWndText, EM_SETMARGINS, (WPARAM)(EC_LEFTMARGIN | EC_RIGHTMARGIN), (LPARAM)(margin | (margin << 16)) );

				/// add the text content
				ok &= bool( ::SetWindowText( hWndText, summaries[i].c_str() ) );

				hWndTexts[i] = hWndText;
			}
		}
	}

	return ok;
}
