/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "perceptuum2.rc.h"
#include "Win32Exception.h"
#include "LogThreadSafe.h"
#include "Logger.h"
#include "BackgroundDialog.h"
#include "RenderDesktopWin32.h"
#include "StopDialog.h"
#include "commctrl.h"
#include <fstream>

#include "Perceptuum2.h"


using namespace perceptuum2::platform_ui;




/// statics
const TCHAR Perceptuum2::WND_CLASS_NAME[]   = TEXT("Perceptuum2WindowClass");
const TCHAR Perceptuum2::WINDOW_TITLE[]     = TEXT("Perceptuum2");
const char  Perceptuum2::CLASS_NAME[]       = "perceptuum2::platform_ui::Perceptuum2";
const char  Perceptuum2::ERROR_MESSAGE1[]   = "couldnt enable buttons";
const char  Perceptuum2::ERROR_MESSAGE2[]   = "couldnt get memory for render object";
const char  Perceptuum2::ERROR_MESSAGE3[]   = "no render object to use";
const char  Perceptuum2::ERROR_MESSAGE4[]   = "timer failed to start";
const char  Perceptuum2::NAME_AND_VERSION[] = "Perceptuum 2.1";
const char  Perceptuum2::LOGFILE_NAME[]     = "perceptuum2.log";


/// application instance
HINSTANCE           hInstance_g = 0;
static Perceptuum2* pApp_g      = 0;




/// standard object services -----------------------------------------------------------------------
Perceptuum2::Perceptuum2
(
	HINSTANCE hInstance,
	HWND      hWnd
)
 :	hInstance_m( hInstance ),
	hWnd_m( hWnd ),
//	hPalette_m( 0 ),
	buttonBar_m( hWnd ),
	hWndStatusBar_m( createStatusBar() ),
	imagePane_m( hInstance, hWnd ),
	pRender_m( 0 ),
	timerId_m( 0 )
{
	static const char METHOD_NAME[] = "constructor";

	augmentSysMenu();

//	hPalette_m = makePalette();

	/// enable buttons
	bool ok = buttonBar_m.enableButtonsForStop();
	ok     &= buttonBar_m.enableButtonsForNoRender();
	if( !ok )
	{
		const DWORD error = ::GetLastError();
		throw hxa7241::win32::Win32Exception( IDS_EXC_BUTTONS_ENABLE, error, CLASS_NAME, METHOD_NAME );
	}
}


Perceptuum2::~Perceptuum2()
{
	enableTimer( false );
	deleteRender();
}




/// commands ---------------------------------------------------------------------------------------
void Perceptuum2::setTonemapAdaptationLuminance
(
	const float f
)
{
	static const char METHOD_NAME[] = "setTonemapAdaptationLuminance";

	RenderDesktopWin32*const pRender = getRenderThrow( METHOD_NAME );
	pRender->setTonemapAdaptationLuminance( f );

	imagePane_m.updateImage();
}


void Perceptuum2::stop()
{
	RenderDesktopWin32*const pRender = getRender();
	if( pRender != 0 )
	{
		pRender->stop();
	}
}


void Perceptuum2::stopWhenFrameDone()
{
	RenderDesktopWin32*const pRender = getRender();
	if( pRender != 0 )
	{
		pRender->stopWhenFrameDone();
	}
}


void Perceptuum2::stopForcibly()
{
	RenderDesktopWin32*const pRender = getRender();
	if( pRender != 0 )
	{
		pRender->stopForcibly();
	}
}


void Perceptuum2::unStop()
{
	RenderDesktopWin32*const pRender = getRender();
	if( pRender != 0 )
	{
		pRender->unStop();
	}
}


void Perceptuum2::enableTimer( const bool enable )
{
	static const char METHOD_NAME[] = "enableTimer";

	if( enable )
	{
		if( timerId_m == 0 )
		{
			timerId_m = ::SetTimer( hWnd_m, TIMER_ID, TIMER_PERIOD_MILLISECS, 0 );
			if( timerId_m == 0 )
			{
				hxa7241::utility::Exception e( ERROR_MESSAGE4, CLASS_NAME, METHOD_NAME );
				hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC2 << e.what();
			}
		}
	}
	else
	{
		if( timerId_m != 0 )
		{
			::KillTimer( hWnd_m, timerId_m );
			timerId_m = 0;
		}
	}
}


void Perceptuum2::start
(
	const string& filePathname,
	const int     nameOffset,
	const int     extensionOffset
)
{
	static const char METHOD_NAME[] = "start";

	try
	{
		/// write message to status bar
		{
			TCHAR buf[64];
			const bool ok = bool( ::LoadString( hInstance_g, IDS_RENDER_START, (LPSTR)buf, sizeof(buf)/sizeof(TCHAR) ) );
			if( ok )
			{
				::SendMessage( hWndStatusBar_m, WM_SETTEXT, 0, (LPARAM)buf );
			}
		}
		/// change cursor
		::SetCursor( ::LoadCursor( NULL, IDC_WAIT ) );

		const string projectName( filePathname.substr( nameOffset, extensionOffset - nameOffset ) );

		/// make and start the render
		makeNewRender( projectName );
		RenderDesktopWin32*const pRender = getRenderThrow( METHOD_NAME );

		/// modify app response
		changeEnablementForRender();

		/// set window title with project name
		{
			string windowTitle;
			windowTitle.reserve( 20 + (extensionOffset * 2) );
			windowTitle += WINDOW_TITLE;
			windowTitle += "  |  ";
			windowTitle += projectName;
			windowTitle += "  (";
			windowTitle += filePathname.substr( 0, extensionOffset );
			windowTitle += ")";
			::SetWindowText( hWnd_m, windowTitle.c_str() );
		}

		/// start rendering
		pRender->start();

		/// change cursor back
		::SetCursor( ::LoadCursor( NULL, IDC_ARROW ) );
	}
	catch( ... )
	{
		/// undo
		::SetWindowText( hWnd_m, WINDOW_TITLE );
		changeEnablementForStop();
		deleteRender();

		TCHAR buf[64] = "\0";
		::LoadString( hInstance_g, IDS_RENDER_START_EXCEPTION, (LPSTR)buf, sizeof(buf)/sizeof(TCHAR) );
		::SendMessage( hWndStatusBar_m, WM_SETTEXT, 0, (LPARAM)buf );

		throw;
	}


	hxa7241::utility::Logger() << "started rendering: " << filePathname;
}


void Perceptuum2::stop
(
	const int //stopCode
)
{
	static const char METHOD_NAME[] = "stop";

//	/// write message to status bar
//	{
//		int id = -1;
//		switch( stopCode )
//		{
//			case StopDialog::STOP_FRAME  : id = IDS_RENDER_STOP1;   break;
//			case StopDialog::STOP_MIDDLE : id = IDS_RENDER_STOP2;   break;
//			case StopDialog::STOP_FORCE  : id = IDS_RENDER_STOP3;   break;
//		}
//		if( id != -1 )
//		{
//			TCHAR buf[64];
//			const bool ok = bool( ::LoadString( hInstance_g, id, (LPSTR)buf, sizeof(buf)/sizeof(TCHAR) ) );
//			if( ok )
//			{
//				::SendMessage( hWndStatusBar_m, WM_SETTEXT, 0, (LPARAM)buf );
//			}
//		}
//	}

	/// modify app response
	const bool ok = changeEnablementForStop();

//	/// persist
//	if( stopCode != StopDialog::STOP_FORCE )
//	{
//		RenderDesktopWin32*const pRender = getRenderThrow( METHOD_NAME );
//		pRender->persist();
//	}

	if( !ok )
	{
		throw hxa7241::utility::Exception( ERROR_MESSAGE1, CLASS_NAME, METHOD_NAME );
	}


	hxa7241::utility::Logger() << "stopped rendering";
}


void Perceptuum2::save
(
	const string& filePathname,
	const int     extensionOffset
)
{
	static const char METHOD_NAME[] = "save";

	const RenderDesktopWin32*const pRender = getRenderThrow( METHOD_NAME );

	/// find format number
	dword format = -1;
	{
		vector<string> names;
		getFormattingNames( names );

		const string extension( filePathname.substr( extensionOffset + 1, (filePathname.length() - extensionOffset - 1) ) );
		for( int i = names.size();  i-- > 0; )
		{
			if( extension == names[i] )
			{
				format = i;
				break;
			}
		}
	}

	/// make file stream
	std::ofstream imageFile( filePathname.c_str(), std::ofstream::binary | std::ofstream::trunc );
	try
	{
		/// request exceptions
		imageFile.exceptions( imageFile.badbit | imageFile.failbit );

		/// format image into stream
		pRender->writeImage( format, imageFile );
	}
	catch( ... )
	{
		/// remove faulty file
		imageFile.close();
		::DeleteFile( filePathname.c_str() );
		throw;
	}

	hxa7241::utility::Logger() << "saved image file: " << filePathname;
}




/// queries ----------------------------------------------------------------------------------------
float Perceptuum2::getTonemapAveLuminance() const
{
	float f = 30.0f;

	const RenderDesktopWin32*const pRender = getRender();
	if( pRender != 0 )
	{
		f = pRender->getTonemapAveLuminance();
	}

	return f;
}


float Perceptuum2::getTonemapAdaptationLuminance() const
{
	float f = 30.0f;

	const RenderDesktopWin32*const pRender = getRender();
	if( pRender != 0 )
	{
		f = pRender->getTonemapAdaptationLuminance();
	}

	return f;
}


bool Perceptuum2::isThreadRunning() const
{
	const RenderDesktopWin32*const pRender = getRender();
	return (pRender != 0) && pRender->isThreadRunning();
}


void Perceptuum2::getSummaryNames
(
	vector<string>& names
) const
{
	static const char METHOD_NAME[] = "getSummaryNames";

	const RenderDesktopWin32*const pRender = getRenderThrow( METHOD_NAME );
	pRender->getSummaryNames( names );
}


void Perceptuum2::getSummaries
(
	vector<string>& summaries
) const
{
	static const char METHOD_NAME[] = "getSummaries";

	const RenderDesktopWin32*const pRender = getRenderThrow( METHOD_NAME );
	pRender->getSummaries( summaries );
}


void Perceptuum2::getStartFileNames( vector<string>& names ) const
{
	RenderDesktopWin32::getStartFileNames( names );
}


void Perceptuum2::getFormattingNames
(
	vector<string>& names
) const
{
	static const char METHOD_NAME[] = "getFormattingNames";

	const RenderDesktopWin32*const pRender = getRenderThrow( METHOD_NAME );
	pRender->getFormattingNames( names );
}




/// implementation ---------------------------------------------------------------------------------
HWND Perceptuum2::createStatusBar()
{
	static const char METHOD_NAME[] = "createStatusBar";

	HWND hWnd = ::CreateStatusWindow( WS_CHILD | WS_VISIBLE, NULL, hWnd_m, IDR_STATUSBAR );
	if( hWnd == 0 )
	{
		const DWORD error = ::GetLastError();
		throw hxa7241::win32::Win32Exception( IDS_EXC_STATUSBAR_CREATE, error, CLASS_NAME, METHOD_NAME );
	}

	return hWnd;
}


void Perceptuum2::augmentSysMenu()
{
	HMENU hMenu = ::GetSystemMenu( hWnd_m, FALSE );

	if( hMenu != NULL )
	{
		const int itemCount = ::GetMenuItemCount( hMenu );
		if( itemCount != -1 )
		{
			TCHAR buf[16];
			const bool ok = bool( ::LoadString( hInstance_m, IDS_BACKGROUND, (LPSTR)&buf, sizeof(buf)/sizeof(TCHAR) ) );
			if( ok )
			{
				::InsertMenu( hMenu, itemCount-1, MF_BYPOSITION | MF_STRING,    IDM_BACKGROUND, buf );
				::InsertMenu( hMenu, itemCount,   MF_BYPOSITION | MF_SEPARATOR, 0, NULL );
			}
		}
	}
}


//HPALETTE Perceptuum2::makePalette()
//{
//	LOGPALETTE lp;
//	lp.palVersion    = 0x300;
//	lp.palNumEntries = 1;
//	HPALETTE hPal = ::CreatePalette( &lp );
//	::ResizePalette( hPal, 256 );
//
//	hxa7241::utility::Array<PALETTEENTRY> entries( 256 );
//	for( int i = entries.getLength();  i-- > 0; )
//	{
//		entries[ i ].peRed   = BYTE( ((i & 0x07) * 0x49) >> 1 );
//		entries[ i ].peGreen = BYTE( ((i & 0x38) * 0x49) >> 4 );
//		entries[ i ].peBlue  = BYTE( ((i & 0xC0) >> 6) * 0x55 );
//		entries[ i ].peFlags = 0;
//	}
//	::SetPaletteEntries( hPal, 0, entries.getLength(), entries.getMemory() );
//
//	return hPal;
//}


bool Perceptuum2::getClientRectAroundImagePane( RECT& rect ) const
{
	bool ok = true;

	RECT mainRect;
	ok &= bool( ::GetClientRect( hWnd_m, &mainRect ) );
	RECT buttonBarRect;
	ok &= buttonBar_m.getWindowRect( buttonBarRect );
	RECT statusBarRect;
	ok &= bool( ::GetWindowRect( hWndStatusBar_m, &statusBarRect ) );

	POINT buttonBarRightBottom = { buttonBarRect.right, buttonBarRect.bottom };
	ok &= bool( ::ScreenToClient( hWnd_m, &buttonBarRightBottom ) );
	POINT statusBarLeftTop     = { statusBarRect.left, statusBarRect.top };
	ok &= bool( ::ScreenToClient( hWnd_m, &statusBarLeftTop ) );

	rect.left   = mainRect.left;
	rect.top    = buttonBarRightBottom.y;
	rect.right  = mainRect.right;
	rect.bottom = statusBarLeftTop.y;

	return ok;
}


bool Perceptuum2::getImagePaneRect( POINT& leftTop, POINT& widthHeight ) const
{
	RECT rect;
	const bool ok = getClientRectAroundImagePane( rect );

	/// calc the rect, inset within border
	const LONG IMAGEPANE_BORDER_WIDTH = 2;

	leftTop.x     = rect.left + IMAGEPANE_BORDER_WIDTH;
	leftTop.y     = rect.top  + IMAGEPANE_BORDER_WIDTH;
	widthHeight.x = (rect.right - rect.left) - (IMAGEPANE_BORDER_WIDTH << 1);
	widthHeight.y = (rect.bottom - rect.top) - (IMAGEPANE_BORDER_WIDTH << 1);

	return ok;
}


void Perceptuum2::makeNewRender
(
	const string& projectName
)
{
	static const char METHOD_NAME[] = "makeNewRender";

	deleteRender();

	pRender_m = new RenderDesktopWin32( projectName, hWnd_m );
	if( pRender_m == 0 )
	{
		throw hxa7241::win32::Exception( ERROR_MESSAGE2, CLASS_NAME, METHOD_NAME );
	}

	/// setup rgb image
	{
		dword width;
		dword height;
		pRender_m->getImageSize( width, height );
		imagePane_m.setImageSize( width, height );

		dword* pPixels;
		imagePane_m.getImage( pPixels, width, height );
		pRender_m->setRgbImage( pPixels, width, height, RenderDesktopWin32::ARGB );
	}
}


void Perceptuum2::deleteRender()
{
	delete pRender_m;
	pRender_m = 0;

	buttonBar_m.enableButtonsForNoRender();
}


RenderDesktopWin32* Perceptuum2::getRender() const
{
	return pRender_m;
}


RenderDesktopWin32* Perceptuum2::getRenderThrow
(
	const char* methodName
) const
{
	RenderDesktopWin32*const pRender = getRender();
	if( pRender == 0 )
	{
		throw hxa7241::utility::Exception( ERROR_MESSAGE3, CLASS_NAME, methodName );
	}

	return pRender;
}


void Perceptuum2::changeEnablementForRender()
{
	static const char METHOD_NAME[] = "changeEnablementForRender";

	try
	{
		const bool ok = buttonBar_m.enableButtonsForStart();
		if( !ok )
		{
			const DWORD error = ::GetLastError();
			throw hxa7241::win32::Win32Exception( IDS_EXC_BUTTONS_ENABLE, error, CLASS_NAME, METHOD_NAME );
		}

		enableTimer( true );
	}
	catch( ... )
	{
		changeEnablementForStop();
		throw;
	}
}


bool Perceptuum2::changeEnablementForStop()
{
	enableTimer( false );

	const bool ok = buttonBar_m.enableButtonsForStop();

	return ok;
}




/// message handlers -------------------------------------------------------------------------------
void Perceptuum2::onSize
(
	HWND   ,//hWnd,
	UINT   ,//message,
	WPARAM wParam,
	LPARAM lParam
)
{
	/// resize bars
	buttonBar_m.onSize( lParam, wParam );
	::SendMessage( hWndStatusBar_m, WM_SIZE, lParam, wParam );

	/// resize image pane
	{
		POINT leftTop;
		POINT widthHeight;
		getImagePaneRect( leftTop, widthHeight );

		imagePane_m.moveWindow( leftTop, widthHeight );
	}
}


void Perceptuum2::onPaint()
{
	PAINTSTRUCT ps;
	HDC dc = ::BeginPaint( hWnd_m, &ps );

	if( dc != NULL )
	{
		/// draw a thin shaded inner window border around the image pane
		RECT rect;
		const bool ok = getClientRectAroundImagePane( rect );
		if( ok )
		{
			::DrawEdge( dc, &rect, EDGE_SUNKEN, BF_RECT );
		}
	}

	::EndPaint( hWnd_m, &ps );
}


//LRESULT Perceptuum2::onQueryNewPalette()
//{
//	HDC dc = ::GetDC( hWnd_m );
//
//	HPALETTE p1 = ::SelectPalette( dc, hPalette_m, FALSE );
//	UINT noOfColsChanged = ::RealizePalette( dc );
//	::SelectPalette( dc, p1, FALSE );
//
//	LRESULT lResult;
//	if( noOfColsChanged > 0 )
//	{
//		::InvalidateRect( hWnd_m, NULL, FALSE );
//
//		lResult = TRUE;
//	}
//	else
//	{
//		lResult = FALSE;
//	}
//
//	return lResult;
//}


bool Perceptuum2::onCommand
(
	WPARAM wParam
)
{
	return buttonBar_m.onCommand( wParam, *this );
}


bool Perceptuum2::onCommandKey
(
	WPARAM wParam
)
{
	return buttonBar_m.onCommandKey( wParam, *this );
}


void Perceptuum2::onBackground()
{
	/// get current color
	const COLORREF background = imagePane_m.getBackground();
	udword grayValue = udword((GetRValue(background) + GetGValue(background) + GetBValue(background)) / 3);

	/// run dialog to maybe get new gray value
	BackgroundDialog dialog( hInstance_m, hWnd_m );
	const bool ok = bool( dialog.run( grayValue ) );
	if( ok )
	{
		/// set new color
		imagePane_m.setBackground( RGB( grayValue, grayValue, grayValue ) );
	}
}


void Perceptuum2::onLineUpdate
(
	WPARAM wParam,
	LPARAM lParam
)
{
	const dword lineNumber = static_cast<dword>( lParam );
	if( lineNumber >= 0 )
	{
		/// induce repaint (partial)
		imagePane_m.updateImage( lineNumber );

		/// set status line content
		const char* pStatusLine = reinterpret_cast<const char*>( wParam );
		::SendMessage( hWndStatusBar_m, WM_SETTEXT, 0, (LPARAM)pStatusLine );
	}
}


void Perceptuum2::onFrameUpdate
(
	WPARAM wParam,
	LPARAM lParam
)
{
	const bool repaint = static_cast<bool>( lParam );
	if( repaint )
	{
		/// induce repaint
		imagePane_m.updateImage();

		/// set status line content
		const char* pStatusLine = reinterpret_cast<const char*>( wParam );
		::SendMessage( hWndStatusBar_m, WM_SETTEXT, 0, (LPARAM)pStatusLine );
	}
}


void Perceptuum2::onRenderUpdate
(
	WPARAM wParam,
	LPARAM lParam
)
{
	static const char METHOD_NAME[] = "onRenderUpdate";

	onFrameUpdate( wParam, lParam );

	/// modify app response
	const bool ok = changeEnablementForStop();
	if( !ok )
	{
		throw hxa7241::utility::Exception( ERROR_MESSAGE1, CLASS_NAME, METHOD_NAME );
	}

	hxa7241::utility::Logger() << "rendering ended";
}


void Perceptuum2::onTimer()
{
	if( timerId_m != 0 )
	{
		const RenderDesktopWin32*const pRender = getRender();
		if( pRender != 0 )
		{
			/// if render thread not running
			if( !pRender->isThreadRunning() )
			{
				changeEnablementForStop();

				/// write stopped message to status bar
				TCHAR buf[64];
				bool ok = bool( ::LoadString( hInstance_g, IDS_RENDER_STOP_ODD2, (LPSTR)buf, sizeof(buf)/sizeof(TCHAR) ) );
				if( ok )
				{
					::SendMessage( hWndStatusBar_m, WM_SETTEXT, 0, (LPARAM)buf );
				}

				/// show message box
				ok = bool( ::LoadString( hInstance_g, IDS_RENDER_STOP_ODD1, (LPSTR)buf, sizeof(buf)/sizeof(TCHAR) ) );
				if( ok )
				{
					hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC1 << buf;
					::MessageBox( hWnd_m, buf, WINDOW_TITLE, MB_OK | MB_ICONEXCLAMATION );
				}
			}
		}
	}
}


bool Perceptuum2::onClose()
{
	bool stopCancelled = false;

	if( isThreadRunning() )
	{
		onCommand( IDM_STOP );
		stopCancelled = isThreadRunning();
	}

	return !stopCancelled;
}



/// win32 ------------------------------------------------------------------------------------------
LRESULT CALLBACK WindowProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );


int WINAPI WinMain
(
	HINSTANCE hInstance,
	HINSTANCE hPrevInstance,
	LPSTR     ,//lpszCmdLine,
	int       nCmdShow
)
{
	static const char METHOD_NAME[] = "WinMain";


	/// detect if theres another instance, and abort this one
	/// (deleted automatically on process termination)
	::CreateMutex( NULL, TRUE, TEXT("perceptuum2") );
	if( ::GetLastError() == ERROR_ALREADY_EXISTS )
	{
		return FALSE;
	}


	hInstance_g = hInstance;


	/// init logging, using compiler switch defined symbols
	hxa7241::utility::LogThreadSafe::init( Perceptuum2::LOGFILE_NAME, LOGGING_ENABLED );
	using hxa7241::utility::Logger;


	Logger() << "# " << Perceptuum2::NAME_AND_VERSION << " : built " << __TIME__ << " " << __DATE__ << " #";
	Logger() << METHOD_NAME << " start  [" << Perceptuum2::CLASS_NAME << "]";


	/// register the window class for the main window
	WNDCLASSEX wc;

	if( !hPrevInstance )
	{
		wc.cbSize        = sizeof(WNDCLASSEX);
		wc.style         = CS_OWNDC | CS_DBLCLKS | CS_SAVEBITS | CS_HREDRAW | CS_VREDRAW;
		wc.lpfnWndProc   = (WNDPROC)WindowProc;
		wc.cbClsExtra    = 0;
		wc.cbWndExtra    = 0;
		wc.hInstance     = hInstance;
		wc.hIcon         = ::LoadIcon( (HINSTANCE)hInstance, MAKEINTRESOURCE(IDI_PERCEPTUUM2) );
		wc.hCursor       = ::LoadCursor( (HINSTANCE)NULL, IDC_ARROW );
		wc.hbrBackground = NULL;//::GetSysColorBrush( COLOR_BTNFACE );///NULL to paint own background
		wc.lpszMenuName  = NULL;
		wc.lpszClassName = Perceptuum2::WND_CLASS_NAME;
		wc.hIconSm       = ::LoadIcon( (HINSTANCE)hInstance, MAKEINTRESOURCE(IDI_PERCEPTUUM2_SMALL) );

		if( ::RegisterClassEx( &wc ) == 0 )
		{
			hxa7241::win32::Win32Exception e( "::RegisterClassEx failed", ::GetLastError(), Perceptuum2::CLASS_NAME, METHOD_NAME );
			Logger() << Logger::EXC2 << e.what();
			return FALSE;
		}
	}


	::InitCommonControls();

	/// create the main window
	HWND hWndMain = ::CreateWindow( Perceptuum2::WND_CLASS_NAME, Perceptuum2::WINDOW_TITLE,
	                                WS_OVERLAPPEDWINDOW,
	                                CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
	                                (HWND)NULL, (HMENU)NULL, hInstance, (LPVOID)NULL );
	if( hWndMain == NULL )
	{
		hxa7241::win32::Win32Exception e( "::CreateWindow failed", ::GetLastError(), Perceptuum2::CLASS_NAME, METHOD_NAME );
		Logger() << Logger::EXC2 << e.what();
		return FALSE;
	}


	::ShowWindow( hWndMain, nCmdShow );
	::UpdateWindow( hWndMain );

	MSG    msg;
	int    status;
	//HACCEL hAccel = ::LoadAccelerators( hInstance, MAKEINTRESOURCE(IDR_ACCELERATORS) );

	while( status = ::GetMessage( &msg, (HWND)NULL, 0, 0 ),
	       (status != 0) && (status != -1) )
	{
		//if( !::TranslateAccelerator( hWndMain, hAccel, &msg ) )
		{
			::TranslateMessage( &msg );
			::DispatchMessage( &msg );
		}
	}


	Logger() << METHOD_NAME << " end  [" << Perceptuum2::CLASS_NAME << "]";

	return msg.wParam;
}


LRESULT CALLBACK WindowProc
(
	HWND   hWnd,
	UINT   message,
	WPARAM wParam,
	LPARAM lParam
)
{
	bool    createOk         = true;

	bool    ok               = false;
	string  exceptionString;

	bool    didHandleMessage = true;
	LRESULT lResult          = 0;

	try
	{
		switch( message )
		{
			case WM_CREATE :
			{
				hxa7241::utility::Logger() << "wm_create";

				createOk = false;
				pApp_g   = new Perceptuum2( hInstance_g, hWnd );
				createOk = true;

				break;
			}
			case WM_SIZE :
			{
				pApp_g->onSize( hWnd, message, wParam, lParam );
				break;
			}
			case WM_PAINT :
			{
				pApp_g->onPaint();
				break;
			}
//			case WM_PALETTECHANGED :
//			{
//				if( (HWND)wParam == hWnd )
//				{
//					break;
//				}
//			}
//			case WM_QUERYNEWPALETTE :
//			{
//				lResult = pApp_g->onQueryNewPalette();
//				break;
//			}
			case WM_COMMAND :
			{
				didHandleMessage = pApp_g->onCommand( wParam );
				break;
			}
			case WM_KEYUP :
			{
				didHandleMessage = pApp_g->onCommandKey( wParam );
				break;
			}
			case WM_SYSCOMMAND :
			{
				switch( wParam )
				{
					case  IDM_BACKGROUND :
					{
						pApp_g->onBackground();
						break;
					}
					default :
					{
						didHandleMessage = false;
						break;
					}
				}
				break;
			}
			case MSG_LINE_UPDATE :
			{
				pApp_g->onLineUpdate( wParam, lParam );
				break;
			}
			case MSG_FRAME_UPDATE :
			{
				pApp_g->onFrameUpdate( wParam, lParam );
				break;
			}
			case MSG_RENDER_UPDATE :
			{
				pApp_g->onRenderUpdate( wParam, lParam );
				break;
			}
			case WM_TIMER :
			{
				pApp_g->onTimer();

				break;
			}
			case WM_CLOSE :
			{
				const bool okToClose = pApp_g->onClose();
				if( okToClose )
				{
					::DestroyWindow( hWnd );
				}

				break;
			}
			case WM_DESTROY :
			{
				hxa7241::utility::Logger() << "wm_destroy";

				::PostQuitMessage( 0 );
				delete pApp_g;
				break;
			}
			default :
			{
				didHandleMessage = false;
				break;
			}
		}
		ok = true;
	}
	catch( const std::exception& e )
	{
		exceptionString = e.what();
	}
	catch( ... )
	{
		hxa7241::utility::Exception e;
		exceptionString = e.what();
	}

	/// common exception handling
	if( !ok )
	{
		hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC2 << exceptionString;
		::MessageBox( hWnd, exceptionString.c_str(), Perceptuum2::WINDOW_TITLE, MB_OK | MB_ICONERROR );

		if( !createOk )
		{
			/// fatal create problem, tell ::CreateWindow() to retreat and fail
			return -1;
		}
	}


	/// default handling
	if( !didHandleMessage )
	{
		lResult = ::DefWindowProc( hWnd, message, wParam, lParam );
	}

	return lResult;
}
