/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Win32Exception.h"
#include "Win32Dialog.h"
#include "perceptuum2.rc.h"
#include "Logger.h"
#include <memory>
#include <cctype>

#include "Win32FileDialog.h"   /// own header is included last


using namespace perceptuum2::platform_ui;




/// statics
const char Win32FileDialog::CLASS_NAME[]   = "perceptuum2::platform_ui::Win32FileDialog";


/// win32 callback
UINT APIENTRY fileDialogHook( HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam );




/// standard object services -----------------------------------------------------------------------
Win32FileDialog::Win32FileDialog
(
	const TCHAR dialogTitle[],
	const TCHAR initialDirectory[],
	const TCHAR filter[],
	const bool  createable
)
{
	Win32FileDialog::construct( dialogTitle, initialDirectory, filter, true, createable );
}


Win32FileDialog::Win32FileDialog
(
	const TCHAR dialogTitle[],
	const TCHAR initialDirectory[],
	const TCHAR filter[]
)
{
	Win32FileDialog::construct( dialogTitle, initialDirectory, filter, false, false );
}


Win32FileDialog::Win32FileDialog
(
	const TCHAR dialogTitle[],
	const TCHAR initialDirectory[],
	const vector<string>& filterNames,
	const bool  createable
)
{
	string filter;
	makeFilter( filterNames, filter );

	Win32FileDialog::construct( dialogTitle, initialDirectory, filter.c_str(), true, createable );
}


Win32FileDialog::Win32FileDialog
(
	const TCHAR dialogTitle[],
	const TCHAR initialDirectory[],
	const vector<string>& filterNames
)
{
	string filter;
	makeFilter( filterNames, filter );

	Win32FileDialog::construct( dialogTitle, initialDirectory, filter.c_str(), false, false );
}


Win32FileDialog::~Win32FileDialog()
{
}




/// commands ---------------------------------------------------------------------------------------
bool Win32FileDialog::run
(
	HINSTANCE hInstance,
	HWND      hWndParent,
	string&   filePathname,
	int&      namePos,
	int&      extensionPos
)
{
	static const char METHOD_NAME[] = "run";


	const int BUF_LENGTH = 1024;
	std::auto_ptr<TCHAR> filePathnameBuf( new TCHAR[BUF_LENGTH] );
	filePathnameBuf.get()[0] = 0;

	OPENFILENAME ofn =
	{
		sizeof(OPENFILENAME),
		hWndParent,
		0,
		filter_m,
		0,
		0,
		1,
		filePathnameBuf.get(),   /// out
		BUF_LENGTH,
		0,
		0,
		initialDirectory_m,
		dialogTitle_m,
		flags_m,
		0,   /// out
		0,   /// out
		pDefaultExtension_m,
		reinterpret_cast<DWORD>(this),
		(LPOFNHOOKPROC)fileDialogHook_m,
		0
	};

	hInstance_m = hInstance;

	BOOL ok;
	if( isOpenNotSave_m )
	{
		ok = ::GetOpenFileName( &ofn );
	}
	else
	{
		ok = ::GetSaveFileName( &ofn );
	}

	if( !ok )
	{
		const DWORD win32Error = ::GetLastError();
		const DWORD errorCode  = ::CommDlgExtendedError();
		if( errorCode != 0 )
		{
			throw hxa7241::win32::Win32Exception( IDS_EXC_FILEDIALOG, win32Error, CLASS_NAME, METHOD_NAME );
		}
	}

	filePathname = filePathnameBuf.get();

	namePos      = ofn.nFileOffset;
	extensionPos = extensionPos_m + filePathname.length();


//using hxa7241::utility::Logger;
//Logger() << "filePathname: {" << filePathname << "}";
//Logger() << "path: {" << filePathname.substr( 0, namePos ) << "}   "
//<< "name: {" << filePathname.substr( namePos, extensionPos-namePos ) << "}   "
//<< "extension: {" << filePathname.substr( extensionPos ) << "}";

	return ok;
}


void Win32FileDialog::setExtensionPos
(
	const int pos
)
{
	extensionPos_m = pos;
}




/// queries ----------------------------------------------------------------------------------------
const vector<string>& Win32FileDialog::getExtensions() const
{
	return extensions_m;
}


HINSTANCE Win32FileDialog::getInstance() const
{
	return hInstance_m;
}




/// implementation ---------------------------------------------------------------------------------
void Win32FileDialog::construct
(
	const TCHAR dlgTitle[],
	const TCHAR initialDir[],
	const TCHAR filter[],
	const bool  openNotSave,
	const bool  createable
)
{
	dialogTitle_m      = dlgTitle;
	initialDirectory_m = initialDir;
	filter_m           = filter;

	if( openNotSave )
	{
		flags_m = (createable ? OFN_CREATEPROMPT : OFN_FILEMUSTEXIST) |
				  OFN_PATHMUSTEXIST | OFN_HIDEREADONLY |
				  OFN_ENABLEHOOK | OFN_EXPLORER;
	}
	else
	{
		flags_m = OFN_NOREADONLYRETURN | OFN_NOTESTFILECREATE | OFN_OVERWRITEPROMPT |
				  OFN_PATHMUSTEXIST | OFN_HIDEREADONLY |
				  OFN_ENABLEHOOK | OFN_EXPLORER;
	}

	pDefaultExtension_m = 0;
	fileDialogHook_m    = fileDialogHook;

	isOpenNotSave_m = openNotSave;
	extensionPos_m  = 0;
	hInstance_m     = 0;


	extractExtensions( filter_m, extensions_m );
}


void Win32FileDialog::makeFilter
(
	const vector<string>& names,
	string&               filter
)
{
	string filterPart1( "(" );
	string filterPart2( "" );
	const dword namesLength = names.size();
	for( int i = 0;  i < namesLength;  ++i )
	{
		filterPart1 += names[i];
		filterPart2 += "*.";
		filterPart2 += names[i];

		if( i >= (namesLength-1) ) break;

		filterPart1 += ", ";
		filterPart2 += ";";
	}
	filterPart1 += ")";
	filterPart1 += '\0';
	filterPart2 += '\0';

	filter  = filterPart1;
	filter += filterPart2;
}


void Win32FileDialog::extractExtensions
(
	const TCHAR     filter[],
	vector<string>& extensions
)
{
	/// from the win32 format, eg: "image files (jpg, png, bmp)\0*.jpg;*.png;*.bmp\0other (oth1, oth2)\0*.oth1;*.oth2\0"

	static const TCHAR EXT_SEPARATOR = ';';

	const TCHAR* pF = filter;

//using hxa7241::utility::Logger;

	string section;
	int    length;
	bool   extSection = false;
	while( true )
	{
		section = pF;
		length = section.length();
//Logger() << "section: {" << section << "}" << "  length: " << length;
		if( length <= 0 )
		{
			break;
		}
		pF += length + 1;

		if( extSection )
		{
			int extPos = 1;
			int colonPos;
			do
			{
				colonPos = section.find( EXT_SEPARATOR, extPos );
				if( colonPos == int(section.npos) )
				{
					colonPos = length;
				}
				string extension( section, extPos, colonPos - extPos );
//Logger() << "extension: {" << extension << "}  length: " << extension.length() << "  extPos: " << extPos << "  colonPos: " << colonPos;
				for( int i = extension.length();  i-- > 0; )
				{
					extension[i] = static_cast<char>( std::tolower( extension[i] ) );
				}
				extensions.insert( extensions.end(), extension );
				extPos = colonPos + 2;
			} while( colonPos < length );
		}

		extSection = !extSection;
	}

//Logger() << "extension count: " << extensions.size();
//for( string::size_type i = 0;  i < extensions.size();  ++i )
//{
//	Logger() << "\t{" << extensions[i] << "}";
//}
}




/// win32 ------------------------------------------------------------------------------------------
UINT APIENTRY fileDialogHook
(
	HWND   hDlg,
	UINT   msg,
	WPARAM ,//wParam,
	LPARAM lParam
)
{
	UINT returnVal = 1;

	switch( msg )
	{
	case WM_NOTIFY :
		{
			const LPOFNOTIFY pNotification = (LPOFNOTIFY)lParam;
			const UINT       code          = pNotification->hdr.code;
			switch( code )
			{
//			/// wont work...
//			case CDN_INITDONE :
//				{
//					using hxa7241::win32::Win32Dialog;
//					const bool ok = Win32Dialog::setPositionCentered( hDlg );
//hxa7241::utility::Logger() << "CDN_INITDONE " << ok;
//					break;
//				}
			case CDN_FILEOK :
				{
					const OPENFILENAME* const pOfn = pNotification->lpOFN;
					const LPTSTR pFileName = pOfn->lpstrFile + pOfn->nFileOffset;

					/// check if filename has a permitted extension
					bool ok = false;
					try
					{
						string fileName( pFileName );

						/// to lower case the filename
						for( int i = fileName.length();  i-- > 0; )
						{
							fileName[i] = static_cast<char>( std::tolower( fileName[i] ) );
						}

						/// look for extensions at the end
						Win32FileDialog* const pDialogObject = reinterpret_cast<Win32FileDialog*>(pOfn->lCustData);
						const vector<string>& extensions( pDialogObject->getExtensions() );
						string::size_type extPos = fileName.npos;
						for( int i = extensions.size();  i-- > 0; )
						{
							const string& ext = extensions[i];
							if( fileName.length() > ext.length() )
							{
								const string::size_type pos = fileName.find( ext, fileName.length() - ext.length() );
								/// find longest match
								if( pos < extPos )
								{
									extPos = pos;
								}
							}
//hxa7241::utility::Logger() << "checking: {" << fileName << "}  {" << ext << "}  " << ok;
						}
						if( extPos != fileName.npos )
						{
							ok = true;
							pDialogObject->setExtensionPos( extPos - fileName.length() );
						}
					}
					catch( const std::exception& e )
					{
						using hxa7241::utility::Logger;
						Logger() << Logger::EXC2 << e.what();
					}
					catch( ... )
					{
						hxa7241::utility::Exception e;
						using hxa7241::utility::Logger;
						Logger() << Logger::EXC2 << e.what();
					}

					if( !ok )
					{
						Win32FileDialog* const pDialogObject = reinterpret_cast<Win32FileDialog*>(pOfn->lCustData);
						HINSTANCE hInstance = pDialogObject->getInstance();
						TCHAR buf1[64];
						const int c1 = ::LoadString( hInstance, IDS_FILEDLG_MSG1, (LPSTR)&buf1, sizeof(buf1)/sizeof(TCHAR) );
						TCHAR buf2[24];
						::LoadString( hInstance, IDS_FILEDLG_MSG1_TITLE, (LPSTR)&buf2, sizeof(buf2)/sizeof(TCHAR) );

						if( c1 != 0 )
						{
							::MessageBox( hDlg, buf1, buf2, MB_OK | MB_ICONEXCLAMATION );
						}

						returnVal = 1;
						::SetWindowLong( hDlg, DWL_MSGRESULT, 1 );
					}

					break;
				}
			default :
				returnVal = 0;
				break;
			}

			break;
		}
	default :
		returnVal = 0;
		break;
	}

	return returnVal;
}
