/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "ByteSerializerOut.h"
#include "Exception.h"

#include "ByteSerializerIn.h"   /// own header is included last


using namespace hxa7241::utility;




/// statics
const char ByteSerializerIn::CLASS_NAME[] = "hxa7241::utility::ByteSerializerIn";

const char ByteSerializerIn::EXCEPTION_MESSAGE1[] = "byte serialization version not supported";
const char ByteSerializerIn::EXCEPTION_MESSAGE2[] = "byte file doesnt match object type";




/// standard object services -----------------------------------------------------------------------
ByteSerializerIn::ByteSerializerIn
(
	istream& in,
	const char* fileName
)
 :	SerializerIn( in, fileName )
{
}


ByteSerializerIn::~ByteSerializerIn()
{
}




/// implementation ---------------------------------------------------------------------------------
ByteSerializerIn& ByteSerializerIn::operator>>( void*& a )
{
	udword d;
	operator>>( d );
	a = reinterpret_cast<void*>( d );

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( bool& a )
{
	ubyte b;
	operator>>( b );
	a = static_cast<bool>( b );

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( hxa7241::utility::byte& a )
{
	ubyte b;
	operator>>( b );
	a = static_cast<hxa7241::utility::byte>( b );

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( ubyte& a )
{
	a = ubyte(in_m.get());

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( word& a )
{
	uword w;
	operator>>( w );
	a = static_cast<word>( w );

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( uword& a )
{
	const ubyte hiByte = ubyte(in_m.get());
	const ubyte loByte = ubyte(in_m.get());

	a = static_cast<uword>( static_cast<udword>(loByte) | (static_cast<udword>(hiByte) << 8) );

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( dword& a )
{
	udword d;
	operator>>( d );
	a = static_cast<dword>( d );

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( udword& a )
{
	const ubyte byte3 = ubyte(in_m.get());
	const ubyte byte2 = ubyte(in_m.get());
	const ubyte byte1 = ubyte(in_m.get());
	const ubyte byte0 = ubyte(in_m.get());

	a = (static_cast<udword>( byte0 )      ) |
	    (static_cast<udword>( byte1 ) <<  8) |
	    (static_cast<udword>( byte2 ) << 16) |
	    (static_cast<udword>( byte3 ) << 24);

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( long& a )
{
	udword d;
	operator>>( d );
	a = static_cast<long>( d );

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( unsigned long& a )
{
	udword d;
	operator>>( d );
	a = static_cast<unsigned long>( d );

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( float& a )
{
	udword d;
	operator>>( d );
	a = *(reinterpret_cast<float*>( &d ));

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( double& a )
{
	udword d;
	operator>>( d );
	a = double( *(reinterpret_cast<float*>( &d )) );

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( char& c )
{
	c = char(in_m.get());

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( char* cp )
{
	udword length;
	operator>>( length );

	while( (length-- > 0) && (*cp != 0) )
	{
		*(cp++) = char(in_m.get());
	}
	*cp = '\0';

	return *this;
}


ByteSerializerIn& ByteSerializerIn::operator>>( string& s )
{
	udword length;
	operator>>( length );

	s.resize( length );
	for( udword i = 0;  i < length;  ++i )
	{
		s[i] = char(in_m.get());
	}

	return *this;
}




/// implementation ---------------------------------------------------------------------------------
void ByteSerializerIn::readHeader
(
	const char* name
)
{
	/// check general serializer header
	const char* pHeader = ByteSerializerOut::HEADER();
	checkHeaderLine( in_m, pHeader, EXCEPTION_MESSAGE1 );

	/// check specific object type
	const char* pTypeName = name;
	checkHeaderLine( in_m, pTypeName, EXCEPTION_MESSAGE2 );

	/// move past description
	while( char(in_m.get()) != '\n' );

	/// move to start of next dword
	const udword pos = in_m.tellg();
	in_m.ignore( (4 - (pos & 0x03)) & 0x03 );
}




void ByteSerializerIn::checkHeaderLine
(
	istream&    in,
	const char* pLine,
	const char* pMessage
)
{
	char c;
	while( (c = char(in.get())) != '\n' )
	{
		if( c != *(pLine++) )
		{
			throw hxa7241::utility::Exception( pMessage, CLASS_NAME, "readHeader" );
		}
	}
}
