/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Float01.h"
#include "Logger.h"
#include "RandomGood.h"
#include <time>

#include "HistogramSmall.h"   /// own header is included last


using namespace hxa7241::utility;




//#define ENABLE_TEST_CODE


/// standard object services -----------------------------------------------------------------------
HistogramSmall::HistogramSmall()
 :	length_m( 1 )
{
	widths_m[0] = 1.0f;
}


HistogramSmall::HistogramSmall
(
	const dword length
)
 :	length_m( length > 0 ? (length < MAX_LENGTH ? length : MAX_LENGTH) : 1 )
{
	for( dword i = length_m;  i-- > 1; )
	{
		widths_m[i] = 0.0f;
	}

	widths_m[0] = 1.0f;
}


HistogramSmall::HistogramSmall
(
	const dword length,
	const float values[]
)
{
	HistogramSmall::setElements( length, values );
}


HistogramSmall::~HistogramSmall()
{
}


HistogramSmall::HistogramSmall
(
	const HistogramSmall& other
)
{
	HistogramSmall::operator=( other );
}


HistogramSmall& HistogramSmall::operator=
(
	const HistogramSmall& other
)
{
	if( &other != this )
	{
		length_m = other.length_m;

		for( dword i = length_m;  i-- > 0; )
		{
			widths_m[i] = other.widths_m[i];
		}
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void HistogramSmall::setElements
(
	const dword length,
	const float values[]
)
{
	length_m = length > 0 ? (length < MAX_LENGTH ? length : MAX_LENGTH) : 1;

	switch( length_m )
	{
		case 0 :
		case 1 :
		{
			widths_m[0] = 1.0f;

			break;
		}
		case 2 :
		{
			widths_m[0] = values[0]   >= 0.0f ? values[0]   : 0.0f;
			widths_m[0] = widths_m[0] <= 1.0f ? widths_m[0] : 1.0f;
			widths_m[1] = 1.0f - widths_m[0];

			break;
		}
		default :
		{
			float sum = 0.0f;

			for( dword i = 0;  i < (length_m - 1);  ++i )
			{
				float width = values[i];
				width = width >= 0.0f ? width : 0.0f;

				const float nextSum = (sum + width);
				if( nextSum < 1.0f )
				{
					widths_m[i] = width;
					sum         = nextSum;
				}
				else
				{
					widths_m[i] = 1.0f - sum;
					sum         = 1.0f;
				}
			}

			widths_m[length_m - 1] = 1.0f - sum;
		}
	}
}




/// queries ----------------------------------------------------------------------------------------
dword HistogramSmall::getChoice
(
	const Float01 stochastic
) const
{
	Float01 dummy;
	return getChoice( stochastic, dummy );
}


dword HistogramSmall::getChoice
(
	const Float01 stochastic,
	Float01&      stochasticRescaled
) const
{
	dword choice;

	switch( length_m )
	{
		case 0 :
		case 1 :
		{
			choice             = 0;
			stochasticRescaled = stochastic;

			break;
		}
		case 2 :
		{
			choice             = dword( stochastic >= widths_m[0] );
			stochasticRescaled = (bool(choice) ? stochastic - widths_m[0] : stochastic) / widths_m[choice];

			//const float thresholds[] = { 0.0f, widths_m[0] };
			//
			//choice             = dword( stochastic >= widths_m[0] );
			//stochasticRescaled = (stochastic - thresholds[choice]) / widths_m[choice];

			break;
		}
		case 3 :
		{
			const float thresholds[3] = { 0.0f, widths_m[0], widths_m[0] + widths_m[1] };

			choice             = dword(stochastic >= thresholds[1]) + dword(stochastic >= thresholds[2]);
			stochasticRescaled = (stochastic - thresholds[choice]) / widths_m[choice];

			break;
		}
		default :
		{
			float below = 0.0f;
			for( choice = 0;  choice < (length_m - 1);  ++choice )
			{
				const float above = below + widths_m[choice];
				if( stochastic < above )
				{
					break;
				}
				below = above;
			}

			stochasticRescaled = (stochastic - below) / widths_m[choice];

			break;
		}
	}

	return choice;
}


bool HistogramSmall::isChosen
(
	const Float01 proportion,
	const Float01 stochastic,
	Float01&      stochasticRescaled
)
{
	const bool isChosen = stochastic < proportion;

	if( isChosen )
	{
		stochasticRescaled = stochastic / proportion;
	}
	else
	{
		stochasticRescaled = (stochastic - proportion) / (1.0f - proportion);
	}

	return isChosen;


//	const float r[] = { proportion, 0.0f };
//	const HistogramSmall h( 2, r );
//	return h.getChoice( stochastic, stochasticRescaled ) == 0;
}








#ifndef ENABLE_TEST_CODE
bool HistogramSmall::test
(
	const bool isLogging
)
{
	if(isLogging) hxa7241::utility::Logger() << "HistogramSmall::test  [disabled]";

	return false;
}
#else
bool HistogramSmall::test
(
	const bool isLogging
)
{
	bool isSucceeded = true;

	static const char PREFIX[] = "HistogramSmall::test  ";
	if(isLogging) hxa7241::utility::Logger() << PREFIX << "=== general test ===";

	const dword seed = dword(std::time(0));
	hxa7241::utility::RandomGood rand( seed );


	/// constructor (length)
	{
		bool isConstructorOk = true;

		const dword l[] = { -2, 0, 1, 5, MAX_LENGTH, 20 };
		for( dword i = 0, end = (sizeof(l)/sizeof(*l));  (i < end) & isConstructorOk;  ++i )
		{
			HistogramSmall hs( l[i] );

			bool isLengthOk;
			if( (l[i] >= 1) & (l[i] <= MAX_LENGTH) )
			{
				isLengthOk = (hs.getLength() == l[i]);
			}
			else
			{
				isLengthOk = (hs.getLength() >= 1) & (hs.getLength() <= MAX_LENGTH);
			}

			bool isElementsOk = true;
			for( dword j = hs.getLength();  j-- > 1; )
			{
				isElementsOk &= (hs.getElements()[j] == 0.0f);
			}
			isElementsOk &= (hs.getElements()[0] == 1.0f);


			isConstructorOk &= (isLengthOk & isElementsOk);
			if( isLogging & !isConstructorOk ) hxa7241::utility::Logger() << PREFIX << "*** constructor test failed  (" << i << ")";
		}

		isSucceeded &= isConstructorOk;
	}


	/// setElements
	{
		bool isSetElementsOk = true;

		for( dword i = 0, end = 10000;  (i < end) & isSetElementsOk;  ++i )
		{
			static const dword LENGTH = dword(float(MAX_LENGTH) * 1.25f);
			float items[ LENGTH ];

			HistogramSmall hs( 1 );
			{
				dword length = rand.getFloat() * (sizeof(items)/sizeof(*items));
				length = length <= (sizeof(items)/sizeof(*items)) ? length : (sizeof(items)/sizeof(*items));
				rand.next();
				//const dword length = (sizeof(items)/sizeof(*items));
				for( dword j = length;  j-- > 0; )
				{
					const float range = rand.getFloat() * 2.0f;
					rand.next();
					const float item = (rand.getFloat() * range) - (range * 0.125f);
					rand.next();

					items[j] = item;
				}

				hs.setElements( length, items );
			}

			const bool isLengthOk = (hs.getLength() >= 1) & (hs.getLength() <= MAX_LENGTH);

			bool isPositiveOk = true;
			bool isSumOk;
			{
				float sum = 0.0f;
				for( dword j = 0, end = hs.getLength();  j < end;  ++j )
				{
					isPositiveOk &= hs.getElements()[j] >= 0.0f;
					sum += hs.getElements()[j];
				}
				isSumOk = float(::fabs( sum - 1.0f ) ) < 0.000001f;
			}

			bool isMatchingOk = true;
			{
				float sum = 0.0f;
				for( dword j = 0, end = hs.getLength() - 1;  (j < end) & (sum < 1.0f);  ++j )
				{
					sum += hs.getElements()[j];
					if( sum < 1.0f )
					{
						isMatchingOk &= (hs.getElements()[j] == (items[j] >= 0.0f ? items[j] : 0.0f));
					}
				}
			}


			isSetElementsOk = (isLengthOk & isPositiveOk & isSumOk & isMatchingOk);
			if( isLogging & !isSetElementsOk ) hxa7241::utility::Logger() << PREFIX << "*** setElements test failed  (" << i << ") seed: " << seed;
		}

		isSucceeded &= isSetElementsOk;
	}


	/// getChoice
	{
		bool isGetChoiceOk = true;

		for( dword i = 0, end = 500;  (i < end) & isGetChoiceOk;  ++i )
		{
			HistogramSmall hs( 1 );
			{
				static const dword LENGTH = dword(float(MAX_LENGTH) * 1.25f);
				float items[ LENGTH ];
				dword length = rand.getFloat() * (sizeof(items)/sizeof(*items));
				length = length <= (sizeof(items)/sizeof(*items)) ? length : (sizeof(items)/sizeof(*items));
				rand.next();
				//const dword length = (sizeof(items)/sizeof(*items));
				for( dword j = length;  j-- > 0; )
				{
					const float range = rand.getFloat() * 2.0f;
					rand.next();
					const float item = (rand.getFloat() * range) - (range * 0.125f);
					rand.next();

					items[j] = item;
				}

				hs.setElements( length, items );
			}


			float sums[MAX_LENGTH];
			sums[0] = 0.0f;
			for( dword j = 1, end = hs.getLength();  j < end;  ++j )
			{
				sums[j] = sums[j - 1] + hs.getElements()[j - 1];
			}


			dword counters[MAX_LENGTH];
			for( dword j = MAX_LENGTH;  j-- > 0; )
			{
				counters[j] = 0;
			}

			bool isRescalingOk = true;
			dword ii = 0;
			for( float f = 0;  f < 1.0f;  f += (1.0f / 131072.0f), ++ii )
			{
				Float01 r;
				const dword c = hs.getChoice( f, r );
				counters[c] += 1;

				isRescalingOk &= ::fabs( r - ((f - sums[c]) / hs.getElements()[c]) ) < 0.000001f;
//hxa7241::utility::Logger(true) << PREFIX << ii << ":  r= " << f << "  rr= " << r << "  dif= " << (r - ((f - sums[c]) / hs.getElements()[c]));
			}

//hxa7241::utility::Logger() << PREFIX << "length= " << hs.getLength() << "   " << ii;
//for( dword jjj = 0, end = hs.getLength();  jjj < end;  ++jjj )
//{
//	hxa7241::utility::Logger() << PREFIX << jjj << ":  elem= " << hs.getElements()[jjj] << "  sum= " << sums[jjj] << "  count= " << counters[jjj] << "  " << (float(counters[jjj]) / float(ii));
//}

			bool isChoicesOk = true;
			for( dword j = 0, end = hs.getLength();  j < end;  ++j )
			{
//hxa7241::utility::Logger(true) << PREFIX << j << ":  dif= " << ::fabs( hs.getElements()[j] - (float(counters[j]) / float(ii)) );
				isChoicesOk &= ::fabs( hs.getElements()[j] - (float(counters[j]) / float(ii)) ) < 0.00001f;
			}

//hxa7241::utility::Logger() << PREFIX << "isRescalingOk " << isRescalingOk << "  isChoicesOk " << isChoicesOk;

			isGetChoiceOk = (isRescalingOk & isChoicesOk);
			if( isLogging & !isGetChoiceOk ) hxa7241::utility::Logger() << PREFIX << "*** getChoice test failed  (" << i << ") seed: " << seed;
		}

		isSucceeded &= isGetChoiceOk;
	}


	if( isLogging ) hxa7241::utility::Logger() << PREFIX  << (isSucceeded ? "--- general test succeeded" : "*** general test failed");

	return isSucceeded;
}
#endif
