/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Sheet_h
#define Sheet_h


#include "Primitives.h"

#include "Array.h"




#include "hxa7241_utility.h"
namespace hxa7241 { namespace utility {
	using namespace hxa7241::primitives;


/**
 * a simple two-dimensional dynamic array, suitable for images.<br/><br/>
 *
 * size is explicit - there is no hidden reserve.<br/><br/>
 *
 * getRow, getElement and operator[] wrap around out of bound indexs.
 */
template<class TYPE>
class Sheet
{
/// standard object services -----------------------------------------------------------------------
public:
	               Sheet();
	               Sheet( dword width, dword height );                     /// throws

	virtual       ~Sheet();
	               Sheet( const Sheet& );                                  /// throws
	        Sheet& operator=( const Sheet& );                              /// throws


/// commands ---------------------------------------------------------------------------------------
	virtual void   setSize( dword width, dword height );                   /// throws


/// queries ----------------------------------------------------------------------------------------
	/// get size
	virtual dword  getLength()                                    const;
	virtual dword  getWidth()                                     const;
	virtual dword  getHeight()                                    const;

	/// check size
	static  bool   isSizeWithinRange( dword width, dword height );
	static  dword  getMaxSize();

	/// access
	virtual TYPE*  getMemory()                                    const;
	virtual TYPE*  getRow( dword y )                              const;   /// wraps out of bounds
	virtual TYPE&  getElement( dword x, dword y )                 const;   /// wraps out of bounds
	virtual TYPE&  operator[]( int i )                            const;   /// wraps out of bounds

	virtual void   zeroMemory()                                   const;

	virtual Sheet* clone()                                        const;   /// throws


/// implementation ---------------------------------------------------------------------------------
protected:
	virtual void   assign( const Sheet<TYPE>& );


/// fields -----------------------------------------------------------------------------------------
protected:
	Array<TYPE> storage_m;
	dword       width_m;
	dword       height_m;
};








//bool isSameSize( const Sheet<TYPE1>&, const Sheet<TYPE2>& );








/// standard object services -----------------------------------------------------------------------
template<class TYPE>
Sheet<TYPE>::Sheet()
 :	storage_m(),
	width_m( 0 ),
	height_m( 0 )
{
}


template<class TYPE>
Sheet<TYPE>::Sheet
(
	const dword width,
	const dword height
)
 :	storage_m(),
	width_m( 0 ),
	height_m( 0 )
{
	Sheet<TYPE>::setSize( width, height );
}


template<class TYPE>
Sheet<TYPE>::~Sheet()
{
}


template<class TYPE>
Sheet<TYPE>::Sheet
(
	const Sheet<TYPE>& other
)
 :	storage_m(),
	width_m( 0 ),
	height_m( 0 )
{
	Sheet<TYPE>::assign( other );
}


template<class TYPE>
Sheet<TYPE>& Sheet<TYPE>::operator=
(
	const Sheet<TYPE>& other
)
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
template<class TYPE>
void Sheet<TYPE>::setSize
(
	const dword width,
	const dword height
)
{
	if( !isSizeWithinRange( width, height ) )
	{
		throw "Sheet<>::setSize()\nsize not within range";
	}

	storage_m.setLength( width * height );
	width_m  = width;
	height_m = height;
}




/// queries ----------------------------------------------------------------------------------------
template<class TYPE>
inline
dword Sheet<TYPE>::getLength() const
{
	return storage_m.getLength();
}


template<class TYPE>
inline
dword Sheet<TYPE>::getWidth() const
{
	return width_m;
}


template<class TYPE>
inline
dword Sheet<TYPE>::getHeight() const
{
	return height_m;
}




template<class TYPE>
bool Sheet<TYPE>::isSizeWithinRange
(
	const dword width,
	const dword height
)
{
	bool isWithin = true;

	/// is negative ?
	if( (width < 0) | (height < 0) )
	{
		isWithin = false;
	}
	else if( height != 0 )   /// prevent divide by zero
	{
		/// is the total number of elements greater than max length ?
		if( width > (getMaxSize() / height) )
		{
			isWithin = false;
		}
	}

	return isWithin;
}


template<class TYPE>
inline
dword Sheet<TYPE>::getMaxSize()
{
	return Array<TYPE>::getMaxLength();
}




template<class TYPE>
inline
TYPE* Sheet<TYPE>::getMemory() const
{
	return storage_m.getMemory();
}


template<class TYPE>
TYPE* Sheet<TYPE>::getRow
(
	dword y
) const
{
	y = (y >= 0) ? y : -y;
	return storage_m.getMemory() + ((y % height_m) * width_m);
}


template<class TYPE>
TYPE& Sheet<TYPE>::getElement
(
	const dword x,
	const dword y
) const
{
	return storage_m[ x + (y * width_m) ];
}


template<class TYPE>
inline
TYPE& Sheet<TYPE>::operator[]
(
	const int i
) const
{
	return storage_m[ i ];
}




template<class TYPE>
inline
void Sheet<TYPE>::zeroMemory() const
{
	storage_m.zeroMemory();
}


template<class TYPE1, class TYPE2>
bool isSameSize
(
	const Sheet<TYPE1>& first,
	const Sheet<TYPE2>& second
)
{
	return (first.getWidth()  == second.getWidth()) &
	       (first.getHeight() == second.getHeight());
}




template<class TYPE>
Sheet<TYPE>* Sheet<TYPE>::clone() const
{
	Sheet<TYPE>* clone = new Sheet<TYPE>( *this );
	if( clone == 0 )
	{
		throw "Sheet<>::clone()\ncouldnt get memory";
	}

	return  clone;
}




/// implementation ---------------------------------------------------------------------------------
template<class TYPE>
void Sheet<TYPE>::assign
(
	const Sheet<TYPE>& other
)
{
	storage_m = other.storage_m;
	width_m   = other.width_m;
	height_m  = other.height_m;
}


}}




#endif//Sheet_h
