/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Thread_h
#define Thread_h


#include "Primitives.h"




#include "hxa7241_utility.h"
namespace hxa7241 { namespace utility {
	using namespace hxa7241::primitives;
	using hxa7241::utility::Runnable;
	class ThreadLock;


/**
 * abstract base of a java style thread class.<br/><br/>
 *
 * derivatives must implement thread starting and joining commands, and live state
 * and lock creation queries. forced stop, prioritys and suspension are optional.
 * default implementations do nothing, getPriorityRanges returns zero.<br/><br/>
 *
 * the code the thread will run is separated into a Runnable derivative. this is
 * supplied to the thread at construction, and is called at start (run) and
 * destruction (setPermission(Runnable::STOP)).<br/><br/>
 *
 * a thread is stopped by calling its Runnables setPermission to STOP, which the
 * Runnable implementation should check regularly.<br/>
 * forced stop is likely to cause some resource leakage, and usage is allowed but
 * not recommended.
 *
 * @see Runnable
 */
class Thread
{
/// standard object services -----------------------------------------------------------------------
protected:
	         Thread( Runnable* r );
public:
	virtual ~Thread();
private:
	         Thread( const Thread& );
	Thread&  operator=( const Thread& );


/// commands ---------------------------------------------------------------------------------------
public:
	virtual void  start()                     =0;
	virtual void  stopForcibly();
	virtual void  join()                      =0;
	virtual void  setPriority( dword );
	virtual void  suspend();
	virtual void  resume();


/// queries ----------------------------------------------------------------------------------------
public:
	virtual bool  isAlive()             const =0;
	virtual dword getPriorityRange()    const;

	virtual ThreadLock* createLock()    const =0;

	static const char* CLASS_NAME();


/// fields -----------------------------------------------------------------------------------------
protected:
	Runnable* runnable_m;
};




/**
 * abstract base of a java style thread executable class.<br/><br/>
 *
 * derivatives must implement run, which the thread will call when its started.
 * the implementation of run must regularly check the permission, and return from
 * run if permission is STOP.
 *
 * @see Thread
 */
class Runnable
{
/// standard object services -----------------------------------------------------------------------
protected:
	          Runnable() : permission_m( GO ) {};

	virtual  ~Runnable() {};
private:
	          Runnable( const Runnable& );
	Runnable& operator=( const Runnable& );


/// commands ---------------------------------------------------------------------------------------
public:
	virtual void  run()                            =0;
	virtual void  setPermission( const dword p )   { permission_m = p; }


/// queries ----------------------------------------------------------------------------------------
public:
	virtual dword getPermission() const            { return permission_m; }


/// fields -----------------------------------------------------------------------------------------
public:
	enum { GO = 0, STOP = 1 };

private:
	volatile dword permission_m;
};




/// must be non-member
unsigned __stdcall threadFunction( void* );




/**
 * ties thread suspension/resumption to an object lifetime.<br/><br/>
 *
 * allows simple one code line temporary thread suspension, since resumption is
 * automatically induced by destruction.
 *
 * @see Thread
 */
class ThreadSuspension
{
/// standard object services -----------------------------------------------------------------------
public:
	        ThreadSuspension( Thread* pThread )
	         :	pThread_m( pThread )
	        {
	        	pThread_m->suspend();
	        }

	       ~ThreadSuspension()
	        {
	        	pThread_m->resume();
	        }
private:
	        ThreadSuspension( const ThreadSuspension& );
	ThreadSuspension& operator=( const ThreadSuspension& );


/// fields -----------------------------------------------------------------------------------------
private:
	Thread* pThread_m;
};




/**
 * abstract base of a thread creator.<br/><br/>
 *
 * derivatives must implement create to call new for a particular Thread
 * derivative.
 *
 * @see Thread
 */
class ThreadFactory
{
/// standard object services -----------------------------------------------------------------------
protected:
	         ThreadFactory() {}
public:
	virtual ~ThreadFactory() {}
private:
	         ThreadFactory( const ThreadFactory& );
	ThreadFactory& operator=( const ThreadFactory& );


/// queries ----------------------------------------------------------------------------------------
public:
	virtual Thread* createThread( Runnable* ) const =0;
};


}}




#endif//Thread_h
