/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Win32ThreadLock.h"
#include "Win32Exception.h"
#include "Logger.h"
#include <process>

#include "Win32Thread.h"


using namespace hxa7241::win32;




/// statics
const int  Win32Thread::PRIORITIES_VALUES[]  = { THREAD_PRIORITY_IDLE, THREAD_PRIORITY_LOWEST, THREAD_PRIORITY_BELOW_NORMAL,
                                                 THREAD_PRIORITY_NORMAL,
                                                 THREAD_PRIORITY_ABOVE_NORMAL, THREAD_PRIORITY_HIGHEST, THREAD_PRIORITY_TIME_CRITICAL };
const int  Win32Thread::PRIORITIES_RANGE     = (sizeof(PRIORITIES_VALUES) / sizeof(int)) / 2;
const int* Win32Thread::PRIORITIES           = PRIORITIES_VALUES + PRIORITIES_RANGE;
const char Win32Thread::CLASS_NAME[]         = "hxa7241::win32::Win32Thread";
const char Win32Thread::EXCEPTION_MESSAGE1[] = "couldnt spawn thread";
const char Win32Thread::EXCEPTION_MESSAGE2[] = "join failed";




/// standard object services ----------------------------------------------------------------------
Win32Thread::Win32Thread( Runnable* runnable )
 :	Thread( runnable ),
	threadHandle_m( 0 ),
	threadId_m( 0 )
{
}


Win32Thread::~Win32Thread()
{
	runnable_m->setPermission( Runnable::STOP );
	join();
}




/// commands ---------------------------------------------------------------------------------------
void Win32Thread::start()
{
	const char METHOD_NAME[] = "start";

	if( threadHandle_m == 0 )
	{
		unsigned threadId;
		const unsigned long bt = _beginthreadex( 0,
		                                         unsigned(4096),
		                                         hxa7241::utility::threadFunction,
		                                         static_cast<void*>(runnable_m),
		                                         0,
		                                         &threadId );

		if( bt == static_cast<unsigned long>(-1) )
		{
			throw hxa7241::win32::Win32Exception( EXCEPTION_MESSAGE1, ::GetLastError(), CLASS_NAME, METHOD_NAME );
		}

		threadHandle_m = reinterpret_cast<HANDLE>( bt );
		threadId_m     = DWORD(threadId);
	}
	else
	{
		using hxa7241::utility::Logger;
		Logger() << Logger::EXC1 << "start called while thread " << (void*)threadId_m << "-" << threadHandle_m << " still valid";
	}
}


void Win32Thread::stopForcibly()
{
	if( isAlive() )
	{
		::TerminateThread( threadHandle_m, 0 );

		using hxa7241::utility::Logger;
		Logger() << Logger::EXC1 << "thread " << (void*)threadId_m << "-" << threadHandle_m << " stopped forcibly";

		::CloseHandle( threadHandle_m );
		threadHandle_m = 0;
		threadId_m     = 0;
	}
}


void Win32Thread::join()
{
	if( threadHandle_m != 0 )
	{
		::WaitForSingleObject( threadHandle_m, INFINITE );

		::CloseHandle( threadHandle_m );
		threadHandle_m = 0;
		threadId_m     = 0;
	}
}


void Win32Thread::setPriority( const dword minus3toPlus3 )
{
	if( threadHandle_m != 0 )
	{
		if( minus3toPlus3 < -PRIORITIES_RANGE )
		{
			minus3toPlus3 = -PRIORITIES_RANGE;
		}
		else if( minus3toPlus3 > +PRIORITIES_RANGE )
		{
			minus3toPlus3 = +PRIORITIES_RANGE;
		}

		::SetThreadPriority( threadHandle_m, PRIORITIES[minus3toPlus3] );
	}
}


void Win32Thread::suspend()
{
	if( threadHandle_m != 0 )
	{
		::SuspendThread( threadHandle_m );
		//hxa7241::utility::Logger() << "thread " << (void*)threadId_m << "-" << threadHandle_m << " suspended";
	}
}


void Win32Thread::resume()
{
	if( threadHandle_m != 0 )
	{
		//hxa7241::utility::Logger() << "thread " << (void*)threadId_m << "-" << threadHandle_m << " resumed";
		::ResumeThread( threadHandle_m );
	}
}




/// queries ----------------------------------------------------------------------------------------
bool Win32Thread::isAlive() const
{
	bool is = false;
	if( threadHandle_m != 0 )
	{
		is = bool(::WaitForSingleObject( threadHandle_m, 0 ) == WAIT_TIMEOUT);
	}

	return is;
}


dword Win32Thread::getPriorityRange() const
{
	return PRIORITIES_RANGE;
}


ThreadLock* Win32Thread::createLock() const
{
	return new Win32ThreadLock();
}
